using System;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Threading;
using System.Windows.Forms;

namespace serial {

  /// <summary>
  /// Primarily for showing maps with the actial position.
  /// </summary>
  /// <remarks>
  /// A form for showing a bitmap with defined bounds and a point
  /// with the current location within these bounds.
  /// The contructor of this class expects the name of the maps file
  /// with definition of the bitmap and bounds of the map. The map file
  /// consists of lines, one line for one map defintion. Format of each
  /// line according to the constructor of Maps.
  /// </remarks>
  /// <example>
  /// The following example maps file <c>example.maps</c>:
  /// <code>
  /// 11d59m05.60sE 49d33m16.05sN 12d00m36.03sE 49d32m21.97sN sbach.bmp
  /// 6d37m44.00sE 43d16m26.95sN 6d39m10.66sE 43d15m29.04sN sttropez.jpg
  /// </code>
  /// defines two map entries, one in the bitmap file <c>sbach.bmp</c>
  /// with the boundaries:
  /// <list type="bullet">
  ///  <item>
  ///    Left upper corner of the file <c>sbach.bmp</c>: 11 degreee 59 minutes
  ///    5.6 seconds East, 49 degree 33 minutes 16.05 seconds North.
  ///  </item>
  ///  <item>
  ///    Right bottom corner of the bitmap file: 12 degree 0 minutes
  ///    36.03 seconds East, 49 degree 32 minutes 21.97 seconds North.
  ///  </item>
  /// </list>
  /// and the second map in the file <c>sttropez.jpg</c> with the boundaries
  /// <list type="bullet">
  ///  <item>
  ///    Left upper cordner of the file <c>sttropez.jpg</c>:
  ///    6 degree 37 minuzes 44 seconds East, 43 degree 16 minutes
  ///    26.95 seconds North.
  ///  </item>
  ///  <item>
  ///    Right bottom corner of the file <c>sttropez.jpg</c>:
  ///    6 degree 39 minutes 10.66 seconds East, 43 degree 15 minutes
  ///    29.04 seconds North.
  ///  </item>
  /// </list>
  /// The coordinates may be defined according to Utils.coord2dec().
  /// Any bitmap format accepted by System.Drawing.Graphics.DrawBitmap
  /// is allowed. The name of the bitmap file may be either an absolute or
  /// relative path. Relative paths are automatically prepended by the path
  /// of the executing assembly.
  /// Example code:
  /// <code lang="C#">
  /// using serial;
  /// public void showLocation() {
  ///   BndMapForm f;
  ///   MapPoint   pt;
  ///   pt = new MapPoint("6d37m00.00sE", "43d16m00.00sN");
  ///   f = new BndMapForm("example.maps");
  ///   f.setPoint(pt);
  ///   f.Show();
  /// }
  /// </code>
  /// will start the form with the bitmap <c>sttropez.jpg</c> and with 
  /// thick black point drawn somewhere in the middle of this bitmap.
  /// You'll have probably to scroll the bitmap a little bid to the bottom
  /// and right in order to see the point.
  /// </example>
  /// <seealso cref="Maps">Maps</seealso>
  /// <seealso cref="MapPoint">MapPoint</seealso>
  /// <seealso cref="Utils.coord2dec">Utils.coord2dec()</seealso>
  public class BndBmpForm : Form {

    /// <summary>Scroll bar thickness</summary>
    private const int SBTHICK = 10;

    /// <summary>Where do the scroll bar start, from left upper</summary>
    private const int SBSTART = 5;

    /// <summary>Shifted start of the scroll bar</summary>
    private const int THICKSTART = BndBmpForm.SBTHICK + BndBmpForm.SBSTART;

    /// <summary>Horizontal scroll bar</summary>
    private HScrollBar m_hsb;

    /// <summary>Vertical scroll bar</summary>
    private VScrollBar m_vsb;

    /// <summary>Zoom scrollbar</summary>
    private HScrollBar m_zsb;

    /// <summary>The bitmap to presently show.</summary>
    private Bitmap m_bmp;

    /// <summary>Brush for drawing various tuff</summary>
    private Brush m_brush;

    /// <summary>Current zoom factor</summary>
    private double m_zoom;

    /// <summary>Bounds of the picture in m_bmp</summary>
    private MapBound m_bounds;

    /// <summary>The point within the bounds to show</summary>
    private MapPoint m_point;

    /// <summary>The maps collection</summary>
    private Maps m_maps;

    /// <summary>Initializes an empty form</summary>
    /// <param name="mapsfilename">file with the maps collection</param>
    public BndBmpForm(string mapsfilename) {
      Stream           str;
      Icon             ico;

      this.m_zoom = 1.0;
      this.m_bounds = MapBound.ZeroBound;
      this.m_point = MapPoint.ZeroPoint;
      this.m_bmp = null;
      this.m_hsb = new HScrollBar();
      this.m_hsb.Height = BndBmpForm.SBTHICK;
      this.m_hsb.Location = new Point(BndBmpForm.THICKSTART, BndBmpForm.SBSTART);
      this.m_hsb.Maximum = 100;
      this.m_hsb.Minimum = 0;
      this.m_hsb.Value = 0;
      this.m_hsb.LargeChange = 2;
      this.m_hsb.SmallChange = 1;
      this.m_hsb.ValueChanged += new EventHandler(this.hsb_changed);
      this.m_vsb = new VScrollBar();
      this.m_vsb.Width = BndBmpForm.SBTHICK;
      this.m_vsb.Location = new Point(BndBmpForm.SBSTART, BndBmpForm.THICKSTART);
      this.m_vsb.Maximum = 100;
      this.m_vsb.Minimum = 0;
      this.m_vsb.Value = 0;
      this.m_vsb.LargeChange = 2;
      this.m_vsb.SmallChange = 1;
      this.m_vsb.ValueChanged += new EventHandler(this.vsb_changed);
      this.m_zsb = new HScrollBar(); // location estimated in configScrollBars()
      this.m_zsb.Height = BndBmpForm.SBTHICK;
      this.m_zsb.Maximum = 10;
      this.m_zsb.Minimum = 1;
      this.m_zsb.LargeChange = 2;
      this.m_zsb.SmallChange = 1;
      this.m_zsb.Value = 5;
      this.m_zsb.ValueChanged += new EventHandler(this.zsb_changed);
      this.Controls.Add(this.m_hsb);
      this.Controls.Add(this.m_vsb);
      this.Controls.Add(this.m_zsb);
      this.m_brush = new SolidBrush(Color.Black);
      this.Text = "Map";
      this.Name = "BndBmpForm";
      this.configScrollBars();
      try {
        str = Assembly.GetExecutingAssembly().GetManifestResourceStream("Marta2.ico");
        ico = new Icon(str);
        str.Close();
        this.Icon = ico;
      }
      catch(Exception) { }
      this.m_maps = null;
      this.m_maps = new Maps(mapsfilename);
    } // BndBmpForm()

    /// <summary>The name of the bitmap file to show</summary>
    public string bmpfile {
      set {
        try {
          this.m_zoom = 1.0;
          this.m_zsb.Value = 5;
          this.m_hsb.Value = 0;
          this.m_vsb.Value = 0;
          this.m_bmp = new Bitmap(value);
          this.configScrollBars();
          this.Invalidate();
        }
        catch(Exception) {
          this.m_bmp = null;
        }
      } 
    } // bmpfile

    /// <summary>The bounds for the picture in bmpfile</summary>
    /// <remarks>
    /// May be also used in order to show current bounds resulting
    /// from the current bitmap and point loaded by setPoint() and
    /// basing on the map descriptions in the maps file passed to
    /// the constructor.
    /// </remarks>
    public MapBound mapbounds {
      set {
        this.m_bounds = value;
      }
      get {
        return this.m_bounds;
      }
    } // bounds

    /// <summary>Sets the point to show in the map</summary>
    /// <remarks>
    /// If the point is within the current bound, then the method shows the
    /// point in the current map bitmap.
    /// Else the method searches for a suitable map in the map collection.
    /// If the map has been found, then it loads the map bitmap and shows the
    /// point in the new bitmap.
    /// </remarks>
    /// <param name="point">the point to show</param>
    public void setPoint(MapPoint point) {
      Map  m;

      if (this.m_bounds.pointInside(point)) {
        if (!point.equals(this.m_point)) {
          this.m_point = point;
          this.Invalidate();
        }
      }
      else {
        if (this.m_maps != null) {
          m = this.m_maps.get(point);
          if (m != null) {
            this.m_point = point;
            this.mapbounds = m.bound;
            this.bmpfile = m.bmpfile;
          }
        }
      }
    } // setPoint()

    /// <summary>Configures the vertical and horizontal scrollbar</summary>
    /// <remarks>
    /// Sets the size and the range of the scrollbars depending on the current
    /// form and the bitmap size.
    /// </remarks>
    private void configScrollBars() {
      const int SBWIDTHSHIFT = BndBmpForm.THICKSTART + BndBmpForm.THICKSTART;
      const int ZSBWIDTHSHIFT = SBWIDTHSHIFT + SBWIDTHSHIFT;
      int       frmw, frmh, bmpw, bmph, w2;

      frmw = this.ClientSize.Width;
      frmh = this.ClientSize.Height;
      // horizontal scroll bar
      if (frmw > SBWIDTHSHIFT) {
        if (this.m_bmp != null) {
          bmpw = (int) (this.m_zoom * this.m_bmp.Width);
        }
        else {
          bmpw = 0;
        }
        if (bmpw > frmw) {
          this.m_hsb.Width = frmw - SBWIDTHSHIFT;
          this.m_hsb.Visible = true;
        }
        else {
          this.m_hsb.Visible = false;
          this.m_hsb.Value = 0;
        }
      }
      else {
        this.m_hsb.Visible = false;
        this.m_hsb.Value = 0;
      }
      // vertical scroll bar
      if (frmh > SBWIDTHSHIFT) {
        if (this.m_bmp != null) {
          bmph = (int) (this.m_zoom * this.m_bmp.Height);
        }
        else {
          bmph = 0;
        }
        if (bmph > frmh) {
          this.m_vsb.Height = frmh - SBWIDTHSHIFT;
          this.m_vsb.Visible = true;
        }
        else {
          this.m_vsb.Visible = false;
          this.m_vsb.Value = 0;
        }
      }
      else {
        this.m_vsb.Visible = false;
        this.m_vsb.Value = 0;
      }
      // zoom scroll bar
      if (frmw > ZSBWIDTHSHIFT) {
        w2 = frmw / 2;
        this.m_zsb.Location = new Point(w2, frmh - BndBmpForm.THICKSTART);
        this.m_zsb.Width = w2 - BndBmpForm.THICKSTART;
        this.m_zsb.Visible = true;
      }
      else {
        this.m_zsb.Visible = false;
      }
    } // configScrollBars()

    /// <summary>Paints the bitmap set with bmpfile</summary>
    /// <param name="e">used for the graphics context</param>
    protected override void OnPaint(PaintEventArgs e) {
      const int circle_rad = 10, circle_rad2 = circle_rad + circle_rad;
      Rectangle r;
      int       scroll_x, scroll_y, point_x, point_y,
                bmpw, bmph, circle_x, circle_y;
      double    bndx, bndy, bndw, bndh;

      if (this.m_bmp != null) {
        bmpw = this.m_bmp.Width;
        bmph = this.m_bmp.Height;
        scroll_x = this.m_hsb.Value * bmpw / (this.m_hsb.Maximum - 1);
        scroll_y = this.m_vsb.Value * bmph / (this.m_vsb.Maximum - 1);
        r = new Rectangle(e.ClipRectangle.Left + scroll_x,
                          e.ClipRectangle.Top + scroll_y,
                          (int) (e.ClipRectangle.Width / this.m_zoom),
                          (int) (e.ClipRectangle.Height / this.m_zoom));
        e.Graphics.DrawImage(this.m_bmp,
                             e.ClipRectangle,
                             r,
                             GraphicsUnit.Pixel);
        bndw = this.m_bounds.width;
        bndh = this.m_bounds.height;
        if (bndw != 0 && bndh != 0) {
          bndx = this.m_bounds.left;
          bndy = this.m_bounds.top;
          point_x = (int) (this.m_zoom * ((this.m_point.x - bndx) * bmpw / bndw -
                                           scroll_x));
          point_y = (int) (this.m_zoom * ((this.m_point.y - bndy) * bmph / bndh -
                                           scroll_y));
          if (point_x >= circle_rad && point_y >= circle_rad) {
            circle_x = point_x - circle_rad;
            circle_y = point_y - circle_rad;
            e.Graphics.FillEllipse(this.m_brush,
                                   circle_x, circle_y,
                                   circle_rad2, circle_rad2);
          }
        }
      } // if (this.m_bmp != null)
    } // OnPaint()

    /// <summary>Called every time the size of the for changes</summary>
    /// <param name="e">ignored</param>
    /// <seealso href="configScrollBars">configScrollBars()</seealso>
    protected override void OnResize(EventArgs e) {
      this.configScrollBars();
    } // OnSizeChanged()

    /// <summary>Called if the horizontal scrollbar changes</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void hsb_changed(object sender, EventArgs e) {
      this.Invalidate();
    } // hsbChanged()

    /// <summary>Called if the vertical scrollbar changes</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void vsb_changed(object sender, EventArgs e) {
      this.Invalidate();
    } // hsbChanged()

    /// <summary>Called if the vertical scrollbar changes</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void zsb_changed(object sender, EventArgs e) {
      switch(this.m_zsb.Value) {
        case  1 : this.m_zoom =  0.10; break;
        case  2 : this.m_zoom =  0.30; break;
        case  3 : this.m_zoom =  0.50; break;
        case  4 : this.m_zoom =  0.75; break;
        case  5 : this.m_zoom =  1.00; break;
        case  6 : this.m_zoom =  3.00; break;
        case  7 : this.m_zoom =  5.00; break;
        case  8 : this.m_zoom =  7.50; break;
        case  9 : this.m_zoom = 10.00; break;
        default : this.m_zoom =  1.00; break;
      } // switch
      this.configScrollBars();
      this.Invalidate();
    } // hsbChanged()

  } // class BndBmpForm

} // namespace test

