using System;
using System.Globalization;
using System.IO;
using System.Xml;

namespace serial {

  /// <summary>Creates a Gpx file with a path description</summary>
  /// <remarks>
  /// Used for creating a gpx file describing a path from coordinates
  /// passed to the subsequent calls to the method addLocation().
  /// </remarks>
  /// <example>
  /// <code lang="C#">
  /// using serial;
  ///
  /// public static void createGpx() {
  ///   Gpsx gpsx;
  ///   gpsx = new Gpsx();
  ///   gpsx.open("file.gpx");
  ///   gpsx.addLocation(12.1234, 49.2345, 401.5);
  ///   gpsx.addLocation(12.1244, 49.2355, 402.5);
  ///   gpsx.addLocation(12.1254, 49.2365, 403.5);
  ///   gpsx.close();
  /// }
  /// </code>
  /// The call to this method would create the following gpx file in file.gpx:
  /// <code>
  /// &lt;?xml version="1.0" encoding="utf-8"?&gt;
  /// &lt;gpx xmlns="http://www.topografix.com/GPX/1/0"&gt;
  ///   &lt;time&gt;2008-12-12T11:00:35Z&lt;/time&gt;
  ///   &lt;wpt lat="12.1234" lon="49.2345"&gt;
  ///     &lt;ele&gt;401.5&lt;/ele&gt;
  ///     &lt;time&gt;2008-12-12T11:00:35Z&lt;/time&gt;
  ///     &lt;name&gt;0&lt;/name&gt;
  ///     &lt;desc&gt;0&lt;/desc&gt;
  ///     &lt;sym&gt;Dot&lt;/sym&gt;
  ///     &lt;type&gt;Dot&lt;/type&gt;
  ///   &lt;/wpt&gt;
  ///   &lt;wpt lat="12.1244" lon="49.2355"&gt;
  ///     &lt;ele&gt;402.5&lt;/ele&gt;
  ///     &lt;time&gt;2008-12-12T11:00:35Z&lt;/time&gt;
  ///     &lt;name&gt;1&lt;/name&gt;
  ///     &lt;desc&gt;1&lt;/desc&gt;
  ///     &lt;sym&gt;Dot&lt;/sym&gt;
  ///     &lt;type&gt;Dot&lt;/type&gt;
  ///   &lt;/wpt&gt;
  ///   &lt;wpt lat="12.1254" lon="49.2365"&gt;
  ///     &lt;ele&gt;403.5&lt;/ele&gt;
  ///     &lt;time&gt;2008-12-12T11:00:35Z&lt;/time&gt;
  ///     &lt;name&gt;2&lt;/name&gt;
  ///     &lt;desc&gt;2&lt;/desc&gt;
  ///     &lt;sym&gt;Dot&lt;/sym&gt;
  ///     &lt;type&gt;Dot&lt;/type&gt;
  ///   &lt;/wpt&gt;
  ///   &lt;bounds minlat="12.1234" minlon="49.2345"
  ///              maxlat="12.1254" maxlon="49.2365" &gt;
  /// &lt;/gpx&gt;
  /// </code>
  /// </example>
  public class Gpsx : IDisposable {

    /// <summary>Time format used in gpx files</summary>
    private const string TIMEFORMAT = "yyyy-MM-ddTHH:mm:ssZ";

    /// <summary>Dummy string put into sym and type elements</summary>
    private const string SYM = "Dot";

    /// <summary>The Xml writer</summary>
    private XmlWriter m_writer;

    /// <summary>Minimum latitude between open() and close()</summary>
    private double m_minlat;

    /// <summary>Minimum longitude between open() and close()</summary>
    private double m_minlon;

    /// <summary>Maximum latitude between open() and close()</summary>
    private double m_maxlat;

    /// <summary>Maximum longitude between open() and close()</summary>
    private double m_maxlon;

    /// <summary>Internal counter of locations</summary>
    private int m_counter;

    /// <summary>Initializes an empty object</summary>
    public Gpsx() {
      this.m_writer = null;
      this.m_minlat = 180.0;
      this.m_minlon = 180.0;
      this.m_maxlat = -180.0;
      this.m_maxlon = -180.0;
      this.m_counter = 0;
    } // Gpsx()

    /// <summary>Closes the xml document</summary>
    ~Gpsx() {
      this.close();
    } // ~Gpsx

    /// <summary>Creates the given XML file</summary>
    /// <param name="filename">the xml file to create</param>
    public void open(string filename) {
      XmlWriterSettings settings;

      this.close();
      settings = new XmlWriterSettings();
      settings.Indent = true;
      this.m_writer = XmlWriter.Create(filename, settings);
      this.m_writer.WriteStartElement("gpx",
                                      "http://www.topografix.com/GPX/1/0");
      this.m_writer.WriteElementString("time",
                                       DateTime.Now.ToString(TIMEFORMAT));
    } // open()

    /// <summary>Closes the xml document</summary>
    public void close() {
      if (this.m_writer != null) {
        this.m_writer.WriteStartElement("bounds");
        this.m_writer.WriteAttributeString("minlat",
                                           this.m_minlat.ToString(CultureInfo.InvariantCulture));
        this.m_writer.WriteAttributeString("minlon",
                                           this.m_minlon.ToString(CultureInfo.InvariantCulture));
        this.m_writer.WriteAttributeString("maxlat",
                                           this.m_maxlat.ToString(CultureInfo.InvariantCulture));
        this.m_writer.WriteAttributeString("maxlon",
                                           this.m_maxlon.ToString(CultureInfo.InvariantCulture));
        this.m_writer.WriteEndElement();
        this.m_writer.WriteEndElement();
        this.m_writer.Close();
        this.m_writer = null;
      }
      this.m_counter = 0;
      this.m_minlat = 180.0;
      this.m_minlon = 180.0;
      this.m_maxlat = -180.0;
      this.m_maxlon = -180.0;
    } // close()

    /// <summary>Adds the given location to the Gpx file</summary>
    /// <param name="nlat">latitude north, negative for south</param>
    /// <param name="elon">longitute east, negative for west</param>
    /// <param name="ele">elevation</param>
    public void addLocation(double nlat, double elon, double ele) {
      string name;

      if (this.m_writer != null) {
        name = this.m_counter.ToString();
        this.m_counter++;
        this.m_writer.WriteStartElement("wpt");
        this.m_writer.WriteAttributeString("lat",
                                           nlat.ToString(CultureInfo.InvariantCulture));
        this.m_writer.WriteAttributeString("lon",
                                           elon.ToString(CultureInfo.InvariantCulture));
        this.m_writer.WriteElementString("ele",
                                         ele.ToString(CultureInfo.InvariantCulture));
        this.m_writer.WriteElementString("time",
                                         DateTime.Now.ToString(TIMEFORMAT));
        this.m_writer.WriteElementString("name", name);
        this.m_writer.WriteElementString("desc", name);
        this.m_writer.WriteElementString("sym", Gpsx.SYM);
        this.m_writer.WriteElementString("type", Gpsx.SYM);
        this.m_writer.WriteEndElement();
      }
      if (nlat < this.m_minlat) this.m_minlat = nlat;
      if (nlat > this.m_maxlat) this.m_maxlat = nlat;
      if (elon < this.m_minlon) this.m_minlon = elon;
      if (elon > this.m_maxlon) this.m_maxlon = elon;
    } // addLocation()

    /// <summary>Adds a comment entry to the gpx file</summary>
    /// <remarks>
    /// Use it in order to store any not positioning related informations
    /// into the file, like exceptions, additional logs etc.
    /// </remarks>
    /// <param name="comment">The comment to add</param>
    public void addComment(string comment) {
      if (this.m_writer != null) {
        this.m_writer.WriteComment(comment);
      }
    } // addComment()

    /// <summary>Closes the xml document</summary>
    public void Dispose() {
      this.close();
    } // Dispose()

    /// <summary>If the gpx file has already been opened</summary>
    public bool isOpen {
      get {
        if (this.m_writer == null) return false;
        return true;
      }
    } // isOpen
  } // class Gpsx

} // namespace serial

