using System;
using System.Collections;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;

namespace serial {

  /// <summary>Client implementation for the server in Serial2IP</summary>
  /// <remarks>
  /// Reads information from the server and calls callbacks for lines read.
  /// </remarks>
  /// <example>
  /// Create a simple TCP/IP client implementation
  /// to <c>localhost</c> on port 5000 and printing all incoming text lines
  /// to the standard output:
  /// <code lang="C#">
  /// using System;
  /// using System.Collections;
  /// using serial;
  /// 
  /// public class IPClientTest {
  ///   private static void readline(ArrayList arr) {
  ///     foreach(object o in arr) {
  ///       Console.WriteLine(o);
  ///     }
  ///   }
  /// 
  ///   public static void Main() {
  ///     IPClient clnt;
  ///     clnt = new IPClient();
  ///     clnt.RdLineCb = new ReadLineDelegate(IPClientTest.readline);
  ///     clnt.oper(ServerType.TCP, "localhost", 5000);
  ///   }
  /// }
  /// </code>
  /// Compile the code using for example
  /// <code>
  /// csc -out:test.exe IPClient.cs IPClientTest.cs Utils.cs
  /// gmcs -out:test.exe IPClient.cs IPClientTest.cs Utils.cs
  /// </code>
  /// Then obtain a server, like 
  /// <A HREF="http://netcat.sourceforge.net">netcat</A> or
  /// <A HREF="http://rbytes.org/software/netcat-review/">netcat for Windows</A>
  /// and bring it to act as a TCP server listening on port 5000 and slowly
  /// sending a content of a large text file:
  /// <code>
  ///  netcat -i 1 -l -p 5000 &lt; largefile.txt
  /// </code>
  /// then start the client using <c>test.exe</c> or <c>mono test.exe</c>
  /// and watch the contents of largefile.txt coming
  /// through TCP. You can also configure the client for example for listening
  /// to UDP broadcast messages using <c>ServerType.BCAST</c> by the call to
  /// <c>IPClient.oper()</c> and feed the client from netcast using this time
  /// <code>
  /// netcat -i 1 -u 255.255.255.255 5000 &lt; largefile.txt
  /// </code>
  /// The client may also communicate with UDP servers, s2ip.exe may also act
  /// as an UDP server. In this case the client sends first anything (byte 0)
  /// to the server in order to let the server discover the UDP port of the
  /// client, on which the client listenes for answers from the server.
  /// </example>
  public class IPClient {

    /// <summary>TCP client for type ServerType.TCP</summary>
    private TcpClient m_tcpclnt;

    /// <summary>UDP client for type ServerType.UDB and BCAST</summary>
    private UdpClient m_udpclnt;

    /// <summary>Reading thread</summary>
    private Thread m_lthr;

    /// <summary>Used for translating bytes into line strings</summary>
    private StringBuilder m_sbline;

    /// <summary>Callback used for lines extracted from the input</summary>
    private ReadLineDelegate m_rdlinecb;

    /// <summary>Current IP client type</summary>
    private ServerType m_type;

    /// <summary>Current server name</summary>
    private string m_server;

    /// <summary>Current port</summary>
    private int m_port;

    /// <summary>For filtering lines from serial port.</summary>
    /// <remarks>
    /// Used, if if the read line delegate RdLineCb is set.
    /// In both the filter and the delegate are set, then the
    /// delegate gets only lines matching starting with the filter
    /// string. Unfortunally not all CF implementations support
    /// regular expressions und we must be happy with this rudimentary
    /// solution.
    /// </remarks>
    private string m_filter;

    /// <summary>If the client is active</summary>
    /// <remarks>
    /// True if the client is active in any of the modes TCP, UDP or BCAST,
    /// false if it is switched off.
    /// </remarks>
    private bool m_isactive;

    /// <summary>Initializes an empty client</summary>
    /// <remarks>Use oper() in order to start reading</remarks>
    public IPClient() {
      this.m_type = ServerType.OFF;
      this.m_server = null;
      this.m_port = 0;
      this.m_tcpclnt = null;
      this.m_udpclnt = null;
      this.m_lthr = null;
      this.m_rdlinecb = null;
      this.m_filter = null;
      this.m_sbline = new StringBuilder(4096);
    } // IPClient()

    /// <summary>Destructor switches the client off</summary>
    ~IPClient() {
      this.oper(ServerType.OFF, null, 0);
    } // ~IPClient()

    /// <summary>Sets the callback for lines retreived from server</summary>
    public ReadLineDelegate RdLineCb {
      set {
        this.m_rdlinecb = value;
      }
    } // RdLineCb

    /// <summary>For filtering lines coming from the serial port</summary>
    /// <remarks>
    /// Used, if if the read line delegate RdLineCb is set.
    /// In both the filter and the delegate are set, then the
    /// delegate gets only lines matching starting with the filter
    /// string. Unfortunally not all CF implementations support
    /// regular expressions und we must be happy with this rudimentary
    /// solution.
    /// Use for example $GPGGA for interesting GPS strings.
    /// </remarks>
    public string Filter {
      get {
        return this.m_filter;
      }
      set {
        this.m_filter = value;
      }
    } // Filter

    /// <summary>Switches the client off or on in the specified mode</summary>
    /// <param name="typ">
    /// stops the client for OFF or starts for the protocol type
    /// </param>
    /// <param name="server">
    /// name or ip of the server to connect to, ignored for the type BCAST
    ///  and OFF
    /// </param>
    /// <param name="port">
    /// the server port or the port to listen on for type BCAST, ignored for OFF
    /// </param>
    public virtual void oper(ServerType typ, string server, int port) {
      switch(typ) {
        case ServerType.OFF :
          if (this.m_tcpclnt != null && this.m_isactive) {
            this.m_isactive = false;
            this.m_tcpclnt.Close();
          }
          if (this.m_udpclnt != null && this.m_isactive) {
            this.m_isactive = false;
            this.m_udpclnt.Close();
          }
          if (this.m_lthr != null) this.m_lthr.Join();
          this.m_lthr = null;
          this.m_tcpclnt = null;
          this.m_udpclnt = null;
          this.m_sbline.Length = 0;
          this.m_type = ServerType.OFF;
          this.m_server = null;
          this.m_port = 0;
          break;
        case ServerType.TCP :
          this.oper(ServerType.OFF, server, port);
          this.m_tcpclnt = new TcpClient(server, port);
          this.m_isactive = true;
          this.m_lthr = new Thread(this.readFktTCP);
          this.m_lthr.Start();
          this.m_type = ServerType.TCP;
          this.m_server = server;
          this.m_port = port;
          break;
        case ServerType.UDP :
          this.oper(ServerType.OFF, server, port);
          this.m_udpclnt = new UdpClient();
          this.m_isactive = true;
          this.m_udpclnt.Connect(server, port);
          this.m_lthr = new Thread(this.readFktUDP);
          this.m_lthr.Start();
          // send anything to the server in order to initialize the connection
          this.m_udpclnt.Send(new byte[] { 0 }, 1);
          this.m_type = ServerType.UDP;
          this.m_server = server;
          this.m_port = port;
          break;
        case ServerType.BCAST :
          this.oper(ServerType.OFF, server, port);
          this.m_udpclnt = new UdpClient(port);
          this.m_isactive = true;
          this.m_lthr = new Thread(this.readFktUDP);
          this.m_lthr.Start();
          this.m_type = ServerType.BCAST;
          this.m_server = null;
          this.m_port = port;
          break;
        default :
          this.m_type = ServerType.OFF;
          this.m_server = null;
          this.m_port = 0;
          throw new ArgumentException(string.Format("Type {0} not implemented yet",
                                                    typ));
      } // switch
    } // oper()

    /// <summary>Converts the byte array into an array of line strings</summary>
    /// <param name="b">the byte array, for example from the server</param>
    /// <param name="len">number of valid bytes in b</param>
    private void extractLines(byte[] b, int len) {
      ArrayList arr;
      string    s;
      int       i;
      char      c;

      if (this.m_rdlinecb != null && b != null && len > 0 && this.m_isactive) {
        arr = new ArrayList();
        for(i = 0; i < len; i++) {
          c = Convert.ToChar(b[i]);
          if (c == '\r' || c == '\n') {
            if (this.m_sbline.Length > 0) {
              s = this.m_sbline.ToString();
              if (this.m_filter == null || s.StartsWith(this.m_filter))
                arr.Add(s);
              this.m_sbline.Length = 0;
            }
          }
          else {
            this.m_sbline.Append(c);
          }
        } // for i
        if (this.m_isactive && arr.Count > 0 && this.m_rdlinecb != null) {
          this.m_rdlinecb(arr);
        }
      } // if
    } // extractLines()

    /// <summary>Reads TCP packets</summary>
    /// <remarks>Reads TCP packets and calls callback for lines read</remarks>
    /// <seealso href="extractLines">extractLines()</seealso>
    private void readFktTCP() {
      NetworkStream ns;
      int           len;
      byte[]        b;

      b = new byte[4096];
      ns = this.m_tcpclnt.GetStream();
      while(this.m_tcpclnt != null &&
            this.m_tcpclnt.Client.Connected &&
            this.m_isactive) {
        try {
          len = ns.Read(b, 0, b.Length);
          this.extractLines(b, len);
        }
        catch(Exception) { }
      } // while
    } // readFktTCP()

    /// <summary>Reads UDP packets.</summary>
    /// <remarks>Reads UDP packets and calls callback for lines read</remarks>
    /// <seealso href="extractLines">extractLines()</seealso>
    private void readFktUDP() {
      IPEndPoint srv;
      byte[]     b;

      srv = new IPEndPoint(IPAddress.Any, 0);
      try {
        while(this.m_udpclnt != null && this.m_isactive) {
          srv.Address = IPAddress.Any;
          srv.Port = 0;
          b = this.m_udpclnt.Receive(ref srv);
          this.extractLines(b, b.Length);
        } // while
      }
      catch(Exception) {}
    } // readFktUDP()

    /// <summary>String representation of the object</summary>
    /// <returns>The description of the current configuration</returns>
    public override string ToString() {
      StringBuilder sb;

      sb = new StringBuilder();
      sb.Append(this.m_type).Append(' ');
      if (this.m_type != ServerType.BCAST && 
          ! string.IsNullOrEmpty(this.m_server))
      {
        sb.Append(this.m_server);
      }
      sb.Append(':').Append(this.m_port);
      return sb.ToString();
    } // ToString()
  } // class IPClient

} // namespace serial

