using System;
using System.Collections;

namespace serial {

  /// <summary>Reads GPS data from the server</summary>
  /// <remarks>
  /// This class is a replacement of the class SerialGPS, which uses
  /// IP (TCP, UDP, UDP broadcast) rather than the serial port for reading
  /// NMEA sequences. It supperts currently GGA sequences, from which it
  /// extracts the latitude, longitude, elevation and the UTC time, recomputes
  /// latitude and longitude into UTM using Utils.toUTM() and calls the
  /// callback method specified in the property CoordCb with the values
  /// extracted from the NMEA GGA sequence.
  /// </remarks>
  /// <example>
  /// Create a simple client test program connecting to the port 5000 of the
  /// server (here localhost) and printing the retrieved coordinates to the
  /// standard output. Use default system (Datum.WGS84) in order to convert
  /// geographical coordinates into UTM:
  /// <code lang="C#">
  /// using System;
  /// using serial;
  /// 
  /// public class IPClientGPSTest {
  ///   private static void coordinates(double tim,
  ///                                   double latitude,
  ///                                   double longitude,
  ///                                   double utmnorth,
  ///                                   double utmeast,
  ///                                   double altitude) {
  ///     Console.WriteLine("{0} {1} {2} {3} {4} {5}",
  ///                       tim, latitude, longitude,
  ///                       utmnorth, utmeast, altitude);
  ///   }
  /// 
  ///   public static void Main() {
  ///     IPClientGPS clnt;
  /// 
  ///     clnt = new IPClientGPS();
  ///     clnt.CoordCb = new CoordDelegate(IPClientGPSTest.coordinates);
  ///     clnt.oper(ServerType.TCP, "localhost", 5000);
  ///   }
  /// }
  /// </code>
  /// Compile the program to <c>test.exe</c> using either the Microsoft or
  /// mono compiler:
  /// <code>
  /// csc -out:test.exe IPClientGPSTest.cs IPClientGPS.cs IPClient.cs Utils.cs
  /// gmcs -out:test.exe IPClientGPSTest.cs IPClientGPS.cs IPClient.cs Utils.cs
  /// </code>
  /// Obtain a source for valid NMEA GGA sequences. It could be of course
  /// a working GPS device, from which the data can be served using the
  /// application s2ip.exe. Alternatively use a file with recorded NMEA
  /// sequences, like the example file <c>nmea.txt</c> below:
  /// <code>
  /// $GPGGA,061245.000,4932.8663,N,01200.0203,E,1,04,2.2,424.3,M,47.4,M,,0000*5F
  /// $GPGGA,061246.000,4932.8665,N,01200.0220,E,1,03,2.5,424.4,M,47.4,M,,0000*5C
  /// $GPGGA,061247.000,4932.8667,N,01200.0240,E,1,04,2.2,424.5,M,47.4,M,,0000*58
  /// $GPGGA,061248.000,4932.8668,N,01200.0262,E,1,04,2.2,424.9,M,47.4,M,,0000*54
  /// $GPGGA,061249.000,4932.8670,N,01200.0275,E,1,04,2.2,425.0,M,47.4,M,,0000*52
  /// $GPGGA,061250.000,4932.8673,N,01200.0282,E,1,04,2.2,424.6,M,47.4,M,,0000*56
  /// $GPGGA,061251.000,4932.8676,N,01200.0287,E,1,04,2.2,424.5,M,47.4,M,,0000*54
  /// $GPGGA,061252.000,4932.8680,N,01200.0288,E,1,04,2.2,424.4,M,47.4,M,,0000*50
  /// $GPGGA,061253.000,4932.8683,N,01200.0293,E,1,04,2.2,424.8,M,47.4,M,,0000*54
  /// $GPGGA,061254.000,4932.8682,N,01200.0293,E,1,03,4.1,424.4,M,47.4,M,,0000*5C
  /// </code>
  /// and create a TCP server for example using
  /// <A HREF="http://netcat.sourceforge.net">netcat</A> listening for incoming
  /// connections on TCP port 5000 and serving slowly this file:
  /// <code>
  /// netcat -i 1 -l -p 5000 &lt; nmea.txt
  /// </code>
  /// Finally start the compiled test program using <c>test.exe</c> or
  /// <c>mono test.exe</c> and wathc the coordinates coming from the network.
  /// See also the documentation on the class IPClient for alternative
  /// communication methods, like UDP or UDB broadcast.
  /// </example>
  /// <seealso cref="Utils.toUTM">Utils.toUTM()</seealso>
  /// <seealso cref="IPClient">IPClient</seealso>
  /// <seealso cref="SerialGPS">SerialGPS</seealso>
  public class IPClientGPS : IPClient {

    /// <summary>Default system for recomputing into UTM</summary>
    public const Datum DEFAULT_SYSTEM = Datum.WGS84;

    /// <summary>Last coordinates retrieved in parseNMEA</summary>
    private double[] m_lastcoord;

    /// <summary>System used for recomputing into UTM</summary>
    private Datum m_system;

    /// <summary>The delegate for handling lines retreived from server</summary>
    /// <remarks>Set to parseNMEA()</remarks>
    private ReadLineDelegate m_gpscb;

    /// <summary>Callback for handling coordinates from GPS</summary>
    private CoordDelegate m_coordcb;

    /// <summary>Initializes an empty object</summary>
    public IPClientGPS() : base() {
      int i;

      this.m_coordcb = null;
      this.m_system = IPClientGPS.DEFAULT_SYSTEM;
      this.m_lastcoord = new double[5];
      for(i = 0; i < this.m_lastcoord.Length; i++) this.m_lastcoord[i] = 0.0;
      this.m_gpscb = new ReadLineDelegate(this.parseNMEA);
    } // IPClientGPS()

    /// <summary>System used for recomputing into UTM</summary>
    public Datum System {
      get { return this.m_system; }
      set { this.m_system = value; }
    } // System

    /// <summary>Sets callback method for coordinates from GPS</summary>
    public CoordDelegate CoordCb {
      set {
        this.m_coordcb = value;
      }
    } // CoordCb

    /// <summary>Callback method for handling incoming NMEA strings</summary>
    /// <remarks>
    /// If the coordinate callback CoordCb has not been set, then the method
    /// does absolutely nothing. Also it does nothing if the incoming gps
    /// string is not a valid NMEA GGA string according to Utils.fromGGA().
    /// Else it parses the string using Utils.fromGGA() and calls the callback
    /// method set in CoordCb, if the coordinates (but not the UTC time)
    /// changed since last call (m_lastcoord).
    /// The method calls Utils.dontSuspend() in order to prevent suspending
    /// windows ce/mobile based device as long as valid data arrives.
    /// </remarks>
    /// <param name="arr">Array of NMEA line strings</param>
    private void parseNMEA(ArrayList arr) {
      double tim1, lat1, lon1, utmn1, utme1, alt1;
      double tim2, lat2, lon2, utmn2, utme2, alt2;
      bool   isValid;

      if (this.m_coordcb != null) {
        tim2 = 0.0; lat2 = 0.0; lon2 = 0.0; utmn2 = 0.0; utme2 = 0.0; alt2 = 0.0;
        isValid = false;
        foreach(object o in arr) {
          if (Utils.fromGGA(out tim1,
                            out lat1,
                            out lon1,
                            out utmn1,
                            out utme1,
                            out alt1,
                            o.ToString(),
                            this.m_system)) {
             tim2 = tim1;
             lat2 = lat1;
             lon2 = lon1;
             utmn2 = utmn1;
             utme2 = utme1;
             alt2 = alt1;
             isValid = true;
           }
        } // foreach
        if (isValid &&
            (Math.Abs(lat2 - this.m_lastcoord[0]) > 0.0002 ||
             Math.Abs(lon2 - this.m_lastcoord[1]) > 0.0002 ||
             Math.Abs(utmn2 - this.m_lastcoord[2]) > 0.01 ||
             Math.Abs(utme2 - this.m_lastcoord[3]) > 0.01 ||
             Math.Abs(alt2 - this.m_lastcoord[4]) > 0.5)) {
          this.m_lastcoord[0] = lat2;
          this.m_lastcoord[1] = lon2;
          this.m_lastcoord[2] = utmn2;
          this.m_lastcoord[3] = utme2;
          this.m_lastcoord[4] = alt2;
          this.m_coordcb(tim2, lat2, lon2, utmn2, utme2, alt2);
        }
        // prevent suspending device as long as we get any data
        Utils.dontSuspend();
      } // if (this.m_coordcb != null)
    } // parseNMEA()

    /// <summary>Switches the client off or on in the specified mode</summary>
    /// <param name="typ">
    /// stops the client for OFF or starts for the protocol type
    /// </param>
    /// <param name="server">
    /// name or ip of the server to connect to, ignored for the type BCAST
    ///  and OFF
    /// </param>
    /// <param name="port">
    /// the server port or the port to listen on for type BCAST, ignored for OFF
    /// </param>
    public override void oper(ServerType typ, string server, int port) {
      base.oper(ServerType.OFF, null, 0);
      this.RdLineCb = this.m_gpscb;
      this.Filter = null;
      base.oper(typ, server, port);
    }
  } // class IPClientGPS

} // namespace serial

