using System;
using System.Drawing;
using System.IO;
using System.Text;

namespace serial {

  /// <summary>Contains one map definition</summary>
  /// <remarks>
  /// A map definition consists of a name of a bitmap file, containing for
  /// example the scan or screenshot of the map itself, and of the coordinates
  /// of the boundaries of the map: left upper and right bottom corner.
  /// An instance of this class can bi initialized from a string, consisting
  /// of 5 fields separated by whitespaces, commas or semicilons:
  /// <list type="number">
  ///  <item>
  ///   The longitude of the left boundary.
  ///   The coordinate strings should be defined according to
  ///   Utils.coord2dec().
  ///  </item>
  ///  <item>The latitude of the top boundary.</item>
  ///  <item>The longitude of the right boundary.</item>
  ///  <item>The latitude of the bottom boundary.</item>
  ///  <item>
  ///   The name of the bitmap file. If the name is an absolute path,
  ///   then it is used as it is. Else the name is completed by the
  ///   path of the exexuting assembly. This class doesn't validate the
  ///   type of the bitmap, but generally bitmap types should be used,
  ///   which can be presented in the form BndBmpForm. The form uses
  ///   System.Drawing.Graphics.DrawImage() which is able to handle the
  ///   most popular formats like bmp, jpg or png.
  ///  </item>
  /// </list>
  /// </remarks>
  /// <example>
  /// <code lang="C#">
  /// using System;
  /// using serial;
  /// public void maptest() {
  ///   Map      m;
  ///   MapPoint p;
  ///   m = new Map("11d59m05.60sE 49d33m16.05sN 12d00m36.03sE 49d32m21.97sN sbach.bmp"));
  ///   p = new MapPoint("12d00m00.00sE 49d00m00.00sN");
  ///   if (m.pointInside(p))
  ///     Console.WriteLine("Point {0} is inside of the map {1}", m, p);
  ///   else
  ///     Console.WriteLine("Point {0} is not inside of the map {1}", m, p);
  /// }
  /// </code>
  /// </example>
  /// <seealso cref="Utils.coord2dec">Utils.coord2dec()</seealso>
  /// <seealso cref="Maps">Maps</seealso>
  /// <seealso cref="MapBound">MapBound</seealso>
  public class Map {

    /// <summary>Separators used for the init string</summary>
    private static readonly char[] SEPS = new char[] { ' ', ',', ';' };

    /// <summary>Bound of a map</summary>
    private MapBound m_bound;

    /// <summary>Name of the file with the bitmap of the file</summary>
    private string m_bmpfile;

    /// <summary>Initializes an empty invalid map</summary>
    public Map() {
      this.m_bound = MapBound.ZeroBound;
      this.m_bmpfile = null;
    } // Map()

    /// <summary>Initializes the map from the given string</summary>
    /// <remarks>
    /// The input string is assumed to consist of 5 fields separated
    /// by whitespaces, commas or semicolons.
    /// The 1. field is the left of the bound, East or West.
    /// The 2. field is the top of the bound, North or South.
    /// The 3. field is the right of the bound, East or West.
    /// The 4. field is the bottom of the bound, North or South.
    /// The 5. field is the name of the bitmap file.
    /// The coordinate strings must be valid according to Utils.coord2dec().
    /// The bitmap file name may be either relative or absolute. Relative file
    /// names are searched in the directory of the exe file, see also property
    /// bmpfile. The bitmap file doesn't need to exist during the creation
    /// of the object, though the property bitmap will return null until the
    /// file is not present. Any bitmap file format recognized by
    /// System.Drawing.Bitmap is allowed.
    /// </remarks>
    /// <param name="str">the init string</param>
    /// <exception cref="ArgumentException">for invalid init strings</exception>
    public Map(string str) {
      string[] tokens;

      this.m_bound = MapBound.ZeroBound;
      this.m_bmpfile = null;
      if (str == null) throw new ArgumentException("null init string");
      tokens = str.Split(Map.SEPS);
      if (tokens.Length < 5) throw new ArgumentException("invalid init string");
      this.m_bound = new MapBound(tokens[0], tokens[1], tokens[2], tokens[3]);
      this.bmpfile = tokens[4];
    }

    /// <summary>Sets or gets the name of the bitmap file</summary>
    /// <remarks>
    /// If the name of the bitmap file is an absolute path, the the absolute
    /// path is used. Else the full path of the bitmap file is constructed
    /// from the directory of the exe file and the file name.
    /// </remarks>
    public string bmpfile {
      get {
        return this.m_bmpfile;
      }
      set {
        if (value == null || Path.IsPathRooted(value))
          this.m_bmpfile = value;
        else {
          this.m_bmpfile = Path.Combine(Utils.Exedir, value);
        }
      }
    }

    /// <summary>Returns the bounds of the map</summary>
    public MapBound bound {
      get {
        return this.m_bound;
      }
    } // bound

    /// <summary>Checks if the given point is inside of bound</summary>
    /// <param name="point">the point to check</param>
    /// <returns>true, if the point is withing the bound</returns>
    public bool pointInside(MapPoint point) {
      return this.m_bound.pointInside(point);
    }

    /// <summary>String representation of the object</summary>
    /// <returns>bounds and name of the bitmap file as string</returns>
    public override string ToString() {
      StringBuilder sb;

      sb = new StringBuilder();
      sb.Append(this.m_bound).Append(' ').Append(this.bmpfile);
      return sb.ToString();
    } // ToString()

  } // class Map

} // namespace serial

