using System;
using System.Text;

namespace serial {

  /// <summary>Bounds of a map picture</summary>
  /// <remarks>
  /// Approx. the same like System.Drawing.Rectangle but for doubles.
  /// Bound units don't matter, but must be the same as the units of the
  /// location used in MapPoint.
  /// Usually decimal degrees are used here, positive for East and North,
  /// negative for West and South.
  /// </remarks>
  /// <seealso cref="MapPoint">MapPoint</seealso>
  public class MapBound {

    /// <summary>Dummy zero bound</summary>
    public static readonly MapBound ZeroBound = new MapBound();

    /// <summary>Left coordinate.</summary>
    private double m_left;

    /// <summary>Top coordinate.</summary>
    private double m_top;

    /// <summary>Width of the bound from left to right.</summary>
    /// <remarks>May be also negative like for the eastern earth half</remarks>
    private double m_width;

    /// <summary>Height of the bound.</summary>
    /// <remarks>May be also negative like for the southern earth half</remarks>
    private double m_height;

    /// <summary>Initializes a zero bound</summary>
    public MapBound() {
      this.m_left = 0.0; this.m_top = 0.0; this.m_width = 0.0; this.m_height = 0.0;
    } // MapBound()

    /// <summary>Initializes the bounds</summary>
    /// <param name="aLeft">the left coordinate</param>
    /// <param name="aTop">the top coordinate</param>
    /// <param name="aRight">the coordinate of the right corner</param>
    /// <param name="aBottom">the coordinate of the bottom</param>
    public MapBound(double aLeft,
                    double aTop,
                    double aRight,
                    double aBottom) {
      this.m_left = 0.0; this.m_top = 0.0; this.m_width = 0.0; this.m_height = 0.0;
      this.m_left = aLeft;
      this.m_top = aTop;
      this.m_width = aRight - aLeft;
      this.m_height = aTop - aBottom;
    } // MapBound()

    /// <summary>From the given coordinates</summary>
    /// <remarks>The coordinate strings according to Utils.coord2dec()</remarks>
    /// <param name="aLeft">the left coordinate</param>
    /// <param name="aTop">the top coordinate</param>
    /// <param name="aRight">the coordinate of the right corner</param>
    /// <param name="aBottom">the coordinate of the bottom</param>
    /// <seealso cref="Utils.coord2dec">Utils.coord2dec()</seealso>
    public MapBound(string aLeft,
                    string aTop,
                    string aRight,
                    string aBottom) {
      this.m_left = 0.0; this.m_top = 0.0; this.m_width = 0.0; this.m_height = 0.0;
      this.m_left = Utils.coord2dec(aLeft, false);
      this.m_top = Utils.coord2dec(aTop, true);
      this.m_width = Utils.coord2dec(aRight, false) - this.m_left;
      this.m_height = Utils.coord2dec(aBottom, true) - this.m_top;
    } // MapBound()

    /// <summary>Left coordinate.</summary>
    public double left {
      get {
        return this.m_left;
      }
    } // left

    /// <summary>Top coordinate.</summary>
    public double top {
      get {
        return this.m_top;
      }
    } // top

    /// <summary>Width of the bound from left to right.</summary>
    /// <remarks>May be also negative like for the eastern earth half</remarks>
    public double width {
      get {
        return this.m_width;
      }
    } // width

    /// <summary>Height of the bound.</summary>
    /// <remarks>May be also negative like for the southern earth half</remarks>
    public double height {
      get {
        return this.m_height;
      }
    } // height

    /// <summary>Checks, if the given point is inside of the bound</summary>
    /// <param name="point">the point to check</param>
    /// <returns>true, if the point is within the bound</returns>
    public bool pointInside(MapPoint point) {
      double hlp, min, max;
      bool   retval;

      retval = false;
      if (point != null) {
        hlp = this.m_left + this.m_width;
        if (hlp > this.m_left) { min = this.m_left; max = hlp; }
        else { min = hlp; max = this.m_left; }
        if (point.x > min && point.x < max) {
          hlp = this.m_top + this.m_height;
          if (hlp > this.m_top) { min = this.m_top; max = hlp; }
          else { min = hlp; max = this.m_top; }
          if (point.y > min && point.y < max) retval = true;
        }
      }
      return retval;
    } // pointInside()

    /// <summary>String representation of the object</summary>
    /// <returns>left, top, width, height, decimal in a string</returns>
    public override string ToString() {
      StringBuilder sb;

      sb = new StringBuilder();
      sb.Append(this.left).Append(' ')
        .Append(this.top).Append(' ')
        .Append(this.width).Append(' ')
        .Append(this.height);
      return sb.ToString();
    } // ToString()

  } // class MapBound

} // namespace serial

