using System.Text;

namespace serial {

  /// <summary>Single point on the map</summary>
  /// <remarks>
  /// Similar to System.Drawing.Point but with doubles, mostly for the current
  /// location to show on the map in BndBmpForm.
  /// </remarks>
  /// <seealso cref="MapBound">MapBound</seealso>
  /// <seealso cref="BndBmpForm.setPoint">BndBmpForm.setPoint()</seealso>
  public class MapPoint {

    /// <summary>Dummy zero point</summary>
    public static readonly MapPoint ZeroPoint = new MapPoint();

    /// <summary>X coordinate of the point.</summary>
    private double m_x;

    /// <summary>Y coordinate of the point.</summary>
    private double m_y;

    /// <summary>Initializes a zero point</summary>
    public MapPoint() {
      this.m_x = 0.0; this.m_y = 0.0;
    } // MapPoint()

    /// <summary>Initializes the point with the given coordinates</summary>
    /// <remarks>
    /// The corrdinates may be theoretically any double numbers, no validation
    /// is performed here. Usually the coordinates should follow the same
    /// standards as in the class MapBounds, for example they should be
    /// valid geographical coordinates in decimal degrees, positive for
    /// North and East, negative for South and West.
    /// </remarks>
    /// <param name="anX">
    /// the horizontal X coordinate.
    /// It is usually the longitude in decimal degrees, positive for East
    /// and negative for West.
    /// </param>
    /// <param name="aY">
    /// the vertical Y coordinate.
    /// It is usually the latitude in decimal degrees, positive for North,
    /// negative for South.
    /// </param>
    public MapPoint(double anX, double aY) {
      this.m_x = 0.0; this.m_y = 0.0;
      this.m_x = anX;
      this.m_y = aY;
    } // MapPoint()

    /// <summary>Initializes the point from coordinate strings</summary>
    /// <remarks>
    /// Format of coordinate strings according to Utils.coord2dec()
    /// </remarks>
    /// <seealso cref="Utils.coord2dec">Utils.coord2dec()</seealso>
    /// <param name="anX">the longitude string</param>
    /// <param name="aY">the latitude string</param>
    public MapPoint(string anX, string aY) {
      this.m_x = 0.0; this.m_y = 0.0;
      this.m_x = Utils.coord2dec(anX, false);
      this.m_y = Utils.coord2dec(aY, true);
    } // MapPoint()

    /// <summary>X coordinate of the point.</summary>
    /// <remarks>
    /// Theoretically any double number, the class makes no validations.
    /// Usually valid geographical coordinates in decimal degrees, positive
    /// for East and negative for West should be used here, similar to the
    /// standards used in MapBounds.
    /// </remarks>
    public double x {
      get { return this.m_x; }
    } // x

    /// <summary>Y coordinate of the point.</summary>
    /// <remarks>
    /// Theoretically any double number, the class makes no validations.
    /// Usually valid geographical coordinates in decimal degrees, positive
    /// for North and negative for South should be used here, similar to the
    /// standards used in MapBounds.
    /// </remarks>
    public double y {
      get { return this.m_y; }
    } // y

    /// <summary>Compares point with the given one</summary>
    /// <param name="point">the point to compare</param>
    /// <returns>true, if both points exactly equal, false else</returns>
    public bool equals(MapPoint point) {
      if (point != null && point.x == this.x && point.y == this.y)
        return true;
      return false;
    } // equals()

    /// <summary>The string representation of the object</summary>
    /// <returns>x and y as decimal coordinates in a string</returns>
    public override string ToString() {
      StringBuilder sb;

      sb = new StringBuilder();
      sb.Append(this.x).Append(' ').Append(this.y);
      return sb.ToString();
    } // ToString()
  } // class MapPint
} // namespace serial

