using System;
using System.Collections;
using System.IO;

namespace serial {

  /// <summary>Collection of Map objects</summary>
  /// <remarks>
  /// Contains a collection of map definitions read from a configuration
  /// file. Each map definition consists of a bitmap file with the map
  /// itself and the boundaries of the file.
  /// </remarks>
  /// <example>
  /// The followingfile <c>example.maps</c>:
  /// <code>
  /// 11d59m05.60sE 49d33m16.05sN 12d00m36.03sE 49d32m21.97sN sbach.bmp
  /// 6d37m44.00sE 43d16m26.95sN 6d39m10.66sE 43d15m29.04sN sttropez.jpg
  /// </code>
  /// defines two map entries, one in the bitmap file <c>sbach.bmp</c>
  /// with the boundaries:
  /// <list type="bullet">
  ///  <item>
  ///    Left upper corner of the file <c>sbach.bmp</c>: 11 degreee 59 minutes
  ///    5.6 seconds East, 49 degree 33 minutes 16.05 seconds North.
  ///  </item>
  ///  <item>
  ///    Right bottom corner of the bitmap file: 12 degree 0 minutes
  ///    36.03 seconds East, 49 degree 32 minutes 21.97 seconds North.
  ///  </item>
  /// </list>
  /// and the second map in the file <c>sttropez.jpg</c> with the boundaries
  /// <list type="bullet">
  ///  <item>
  ///    Left upper cordner of the file <c>sttropez.jpg</c>:
  ///    6 degree 37 minuzes 44 seconds East, 43 degree 16 minutes
  ///    26.95 seconds North.
  ///  </item>
  ///  <item>
  ///    Right bottom corner of the file <c>sttropez.jpg</c>:
  ///    6 degree 39 minutes 10.66 seconds East, 43 degree 15 minutes
  ///    29.04 seconds North.
  ///  </item>
  /// </list>
  /// <code lang="C#">
  /// using System;
  /// using serial;
  /// public void checkMap() {
  ///   Maps     maps;
  ///   Map      map;
  ///   MapPoint pt;
  ///
  ///   maps = new Maps("example.maps");
  ///   pt = new Point(12.5, 49.3);
  ///   map = maps.get(pt);
  ///   if (map != null) {
  ///     Console.WriteLine("Found map {0} for point {1}", map);
  ///   }
  /// }
  /// </code>
  /// </example>
  /// <seealso cref="Map">Map</seealso>
  public class Maps : ArrayList {

    /// <summary>Empty maps collection</summary>
    public Maps() : base() {
    } // Maps()

    /// <summary>Reads map definitions from the given file</summary>
    /// <remarks>
    /// The file is assumed to contain lines with a valid map definition,
    /// according to the constructor of the class Map. Each invalid line is
    /// silently ignored.
    /// </remarks>
    /// <param name="filename">the name of the file to read</param>
    public Maps(string filename) {
      Map          m;
      StreamReader str;
      Exception    e;
      string       line;

      str = null;
      e = null;
      try {
        str = new StreamReader(filename);
        while((line = str.ReadLine()) != null) {
          try {
            m = new Map(line);
            this.Add(m);
          }
          catch(ArgumentException) {}
        }
      }
      catch(Exception excp) { e = excp; }
      if (str != null) str.Close();
      if (e != null) throw e;
    } // Maps()

    /// <summary>Returns the first map containing the given point</summary>
    /// <param name="point">the point</param>
    /// <returns>the first map for the point or null, if not found</returns>
    public Map get(MapPoint point) {
      Map m, retval;

      retval = null;
      if (point != null) {
        foreach(object o in this) {
          try {
            m = (Map) o;
            if (m != null && m.pointInside(point)) {
              retval = m;
              break;
            }
          }
          catch(Exception) {}
        } // foreach
      }
      return retval;
    } // get()
  } // class Maps

} // namespace serial

