using System;
using System.Collections;
using System.IO.Ports;
using System.Text;
using System.Threading;

namespace serial {

  /// <summary>Encapsulates serial port operations</summary>
  /// <remarks>
  /// Open, close, read and write operations on the serial port.
  /// No event driven, continous reading of serial ports in a
  /// thread with predefined timeout. This makes this class available
  /// to mono and to some broken implementations on Windows CE.
  /// </remarks>
  /// <example>
  /// The following example defines method readCallback() executed from
  /// the serial port reading thread, which prints the lines from the
  /// serial port to the standard output. Parallelly to the reading
  /// operations the method Main() also reads lines from the standard
  /// input and sends them directly to the serial port. The example
  /// program opens the serial port COM4, this is usually the GPS
  /// intermediate driver (GPSID) on HTC3300. Please be aware, that the
  /// port are differently named on different platforms, for example
  /// COM1: on desktop windows systems, /dev/ttyS0 on linux or COM1
  /// on many PDAs. The list of valid port names can be retrieved from
  /// static property Serial.COMPORTS.
  /// <code lang="C#">
  /// using System;
  /// using System.Collections;
  /// using serial;
  /// public class Test {
  ///   private static void readCallback(ArrayList arr) {
  ///     foreach(object o in arr) Console.WriteLine(o);
  ///   }
  ///   public static void Main() {
  ///     Serial s;
  ///     string str;
  ///
  ///     s = new Serial();
  ///     s.PortName = "COM4";
  ///     s.RdLineCb = new ReadLineDelegate(Test.readCallback);
  ///     s.Open();
  ///     s.Read = true;
  ///     while ((str = Console.ReadLine()) != null &amp;&amp; str != "quit") {
  ///       s.Write(str);
  ///     }
  ///     s.Close();
  ///   }
  /// }
  /// </code>
  /// Additionally to the property RdLineCb the property RdByteCb defines
  /// defines a callback for direct handling of byte arrays read from the
  /// port.
  /// </example>
  public class Serial {

    /// <summary>Baudrates supported by the module</summary>
    public static readonly int[] BAUDRATES = { 
      256000, 128000, 115200, 57600, 38400, 19200,  14400,
        9600,   4800,   2400,  1200,   600,   300,    110
    };

    /// <summary>Number of databits supported</summary>
    public static readonly int[] DATABITS = { 8, 7 };

    /// <summary>Default baudrate</summary>
    public static readonly int DEFAULT_BAUDRATE;

    /// <summary>Default parity</summary>
    public static readonly Parity DEFAULT_PARITY;

    /// <summary>Default number of stop bits</summary>
    public static readonly StopBits DEFAULT_STOPBITS;

    /// <summary>Default number of databits</summary>
    public static readonly int DEFAULT_DATABITS;

    /// <summary>Default timeout for read and write operations</summary>
    /// <remarks>
    /// The default timeout is set internally in this class to 2000ms.
    /// </remarks>
    public static readonly int DEFAULT_TIMEOUT;

    /// <summary>Default serial port name</summary>
    /// <remarks>
    /// The default port name is the name of the port in the instance
    /// of System.IO.Ports.SerialPort initialized without any
    /// configuration effort. Usually it is COM1, COM1: or /dev/ttyS0,
    /// depending on the platform, on which the program is running.
    /// </remarks>
    public static readonly string DEFAULT_COMPORT;

    /// <summary>All available serial port names</summary>
    /// <remarks>
    /// Depending on, if the code has been compiled with CHECK_AVAILABLE_PORTS
    /// defined or not this array contains a sorted output of 
    /// System.IO.Ports.SerialPort.GetPortNames() or only these ports, which
    /// could be opened without an exception. The check, if the port can be
    /// opened, takes a remarkable amount of time on small and slow mobile
    /// devices, therefore it is better to compile the code without this
    /// define.
    /// </remarks>
    public static readonly string[] COMPORTS;

    /// <summary>Buffer len for receving bytes from the serial port</summary>
    private const int BUFLEN = 4096;

    /// <summary>Sets default values</summary>
    /// <remarks>
    /// The constructor reads default setting from a SerialPort object.
    /// Each potentially available port name returned by
    /// SerialPort.GetPortNames() is opened and added to Serial.COMPORTS only,
    /// if the opening succeeds. Therefore serial ports opened by another
    /// applications, not accessible due to security reasons or currently not
    /// existing in the system are not added to Serial.COMPORTS.
    /// </remarks>
    static Serial() {
      SerialPort port;
      string[]   portnames;
#if CHECK_AVAILABLE_PORTS
      string[]   eff_portnames;
      int        i, len;
#endif

      port = new SerialPort();
      Serial.DEFAULT_COMPORT = port.PortName;
      Serial.DEFAULT_BAUDRATE = port.BaudRate;
      Serial.DEFAULT_PARITY = port.Parity;
      Serial.DEFAULT_STOPBITS = port.StopBits;
      Serial.DEFAULT_DATABITS = port.DataBits;
      Serial.DEFAULT_TIMEOUT = 2000; // default is usually no timeout
      portnames =  SerialPort.GetPortNames();
      Array.Sort(portnames);
#if CHECK_AVAILABLE_PORTS
      try {
        len = portnames.Length;
        eff_portnames = new string[len];
        port = new SerialPort();
        port.ReadTimeout = Serial.DEFAULT_TIMEOUT;
        port.WriteTimeout = Serial.DEFAULT_TIMEOUT;
        len = 0;
        foreach(string portname in portnames) {
          try {
            port.PortName = portname;
            port.Open();
            port.Close();
            eff_portnames[len] = portname;
            len++;
          }
          catch(Exception) { }
        }
        Serial.COMPORTS = new string[len];
        for(i = 0; i < len; i++) Serial.COMPORTS[i] = eff_portnames[i];
      }
      catch(Exception) {
        // really bad, try to return something reasonable anyway
        Serial.COMPORTS = portnames;
      }
#else
      Serial.COMPORTS = portnames;
#endif
    } // Serial()

    /// <summary>The serial port</summary>
    private SerialPort m_serial;

    /// <summary>The buffer to use for reading from the serial port</summary>
    private byte[] m_buf;

    /// <summary>Parsed line strings from the buffer</summary>
    /// <remarks>Used, if m_rdlinecb is set</remarks>
    private StringBuilder m_sbline;

    /// <summary>For filtering lines from serial port.</summary>
    /// <remarks>
    /// Used, if if the read line delegate RdLineCb is set. 
    /// In both the filter and the delegate are set, then the
    /// delegate gets only lines matching starting with the filter
    /// string. Unfortunally not all CF implementations support
    /// regular expressions und we must be happy with this rudimentary
    /// solution.
    /// </remarks>
    private string m_filter;

    /// <summary>Callback handling strings read from port</summary>
    private ReadLineDelegate m_rdlinecb;

    /// <summary>Callback handling single bytes read from port</summary>
    private ReadByteDelegate m_rdbytecb;

    /// <summary>Serial port reading thread</summary>
    private Thread m_thr;

    /// <summary>Flag for continuing reading thread</summary>
    private bool m_doread;

    /// <summary>Initializes the serial port with default values</summary>
    public Serial() {
      this.m_buf = new byte[Serial.BUFLEN];
      this.m_sbline = new StringBuilder(Serial.BUFLEN);
      this.m_serial = new SerialPort();
      this.m_serial.ReadTimeout = Serial.DEFAULT_TIMEOUT;
      this.m_serial.WriteTimeout = Serial.DEFAULT_TIMEOUT;
      this.m_thr = null;
      this.m_rdlinecb = null;
      this.m_rdbytecb = null;
      this.m_filter = null;
      this.m_doread = false;
    } // Serial()

    /// <summary>Closes serial port</summary>
    ~Serial() {
       this.Close();
    } // ~Serial()

    /// <summary>Returns the string representation of the object</summary>
    public override string ToString() {
      StringBuilder sb;

      sb = new StringBuilder(128);
      sb.Append(this.PortName)
        .Append(' ')
        .Append(this.BaudRate)
        .Append(' ')
        .Append(this.DataBits)
        .Append(' ')
        .Append(this.Parity)
        .Append(' ')
        .Append(this.StopBits)
        .Append(" timeout ")
        .Append(this.Timeout)
        .Append(" opened ")
        .Append(this.IsOpen);
      return sb.ToString();
    } // ToString()

    /// <summary>Baud rate</summary>
    public int BaudRate {
      set {
        if (this.m_serial != null) this.m_serial.BaudRate = value;
      }
      get {
        if (this.m_serial != null) return this.m_serial.BaudRate;
        return Serial.DEFAULT_BAUDRATE;
      }
    } // BaudRate

    /// <summary>Baud rate</summary>
    public int DataBits {
      set {
        if (this.m_serial != null) this.m_serial.DataBits = value;
      }
      get {
        if (this.m_serial != null) return this.m_serial.DataBits;
        return Serial.DEFAULT_DATABITS;
      }
    } // DataBits

    /// <summary>Parity</summary>
    public Parity Parity {
      set {
        if (this.m_serial != null) this.m_serial.Parity = value;
      }
      get {
        if (this.m_serial != null) return this.m_serial.Parity;
        return Serial.DEFAULT_PARITY;
      }
    } // Parity

    /// <summary>Serial port name</summary>
    public string PortName {
      set {
        if (this.m_serial != null) this.m_serial.PortName = value;
      }
      get {
        if (this.m_serial != null) return this.m_serial.PortName;
        return Serial.DEFAULT_COMPORT;
      }
    } // PortName

    /// <summary>Number of stop bits</summary>
    public StopBits StopBits {
      set {
        if (this.m_serial != null) this.m_serial.StopBits = value;
      }
      get {
        if (this.m_serial != null) return this.m_serial.StopBits;
        return Serial.DEFAULT_STOPBITS;
      }
    } // StopBits

    /// <summary>Timeout in milliseconds for both read and write</summary>
    public int Timeout {
      set {
        if (this.m_serial != null) {
          this.m_serial.ReadTimeout = value;
          this.m_serial.WriteTimeout = value;
        }
      }
      get {
        if (this.m_serial != null) return this.m_serial.ReadTimeout;
        return Serial.DEFAULT_TIMEOUT;
      }
    } // Timeout

    /// <summary>
    /// Calback function for handling strings read from the serial port.
    /// </summary>
    public ReadLineDelegate RdLineCb {
      set {
        this.m_rdlinecb = value;
      }
    } // RdLineCb

    /// <summary>
    /// Calback function for handling bytes read from the serial port.
    /// </summary>
    public ReadByteDelegate RdByteCb {
      set {
        this.m_rdbytecb = value;
      }
    } // RdByteCb

    /// <summary>Tells if the serial port is opened</summary>
    public bool IsOpen {
      get {
        if (this.m_serial != null) return this.m_serial.IsOpen;
        return false;
      }
    } // IsOpen

    /// <summary>If the reading thread have to be/is started</summary>
    public bool Read {
      get {
        if (this.m_thr == null) return false;
        return true;
      }
      set {
        if (value) {
          if (this.IsOpen) {
            // start or restart reading thread
            this.Read = false;
            this.m_doread = true;
            this.m_thr = new Thread(this.readFkt);
            this.m_thr.IsBackground = true;
            this.m_thr.Start();
          }
        }
        else {
          // stop reading thread, if started
          this.m_doread = false;
          if (this.m_thr != null) {
            this.m_thr.Join();
            this.m_thr = null;
          }
        }
      }
    } // Read

    /// <summary>For filtering lines coming from the serial port</summary>
    /// <remarks>
    /// Used, if if the read line delegate RdLineCb is set. 
    /// In both the filter and the delegate are set, then the
    /// delegate gets only lines matching starting with the filter
    /// string. Unfortunally not all CF implementations support
    /// regular expressions und we must be happy with this rudimentary
    /// solution.
    /// Use for example $GPGGA for interesting GPS strings.
    /// </remarks>
    public string Filter {
      get {
        return this.m_filter;
      }
      set {
        this.m_filter = value;
      }
    } // Rexpr

    /// <summary>Reading thread routine</summary>
    /// <remarks>
    /// Calls the callback function defined in RdLineCb (m_rdlinecb) after
    /// each line string read from the serial port.
    /// If the filter has been set using Filter, then the 
    /// callback function is called only for strings read from the serial
    /// port and starting with the defined filter string.
    /// </remarks>
    private void readFkt() {
      ArrayList     arr;
      string        s;
      int           len, i;
      char          c;

      if (this.m_serial != null && this.IsOpen) {
        while (this.m_doread) {
          try {
            len = this.m_serial.Read(this.m_buf, 0, Serial.BUFLEN);
            if (len > 0) {
              if (this.m_doread && this.m_rdbytecb != null)
                this.m_rdbytecb(this.m_buf, len);
              if (this.m_rdlinecb != null) {
                arr = new ArrayList();
                for(i = 0; i < len; i++) {
                  c = Convert.ToChar(this.m_buf[i]);
                  if (c == '\r' || c == '\n') {
                    if (this.m_sbline.Length > 0) {
                      s = this.m_sbline.ToString();
                      if (this.m_filter == null || s.StartsWith(this.m_filter))
                        arr.Add(s);
                      this.m_sbline.Length = 0;
                    }
                  }
                  else {
                    this.m_sbline.Append(c);
                  }
                } // for
                if (this.m_doread && arr.Count > 0) this.m_rdlinecb(arr);
              } // if (this.m_rdlinecb != null)
            } // if (len > 0)
          }
          catch(TimeoutException) { }
        } // while
      } // if
    } // readFkt()

    /// <summary>Closes serial port</summary>
    public void Close() {
      this.Read = false;
      if (this.m_serial != null && this.m_serial.IsOpen) {
        this.m_serial.Close();
      }
    } // Close()

    /// <summary>Open serial port</summary>
    public virtual void Open() {
      this.Close();
      if (this.m_serial != null) { 
        this.m_serial.Open();
        this.Read = false;
      }
    } // Open()

    /// <summary>Writes the given string to the serial port</summary>
    /// <param name="s">The string to write</param>
    public void Write(string s) {
      if (this.m_serial != null) {
        this.m_serial.Write(s);
      }
    } // Write()

    /// <summary>Writes number of bytes from the byte array</summary>
    /// <param name="buffer">the buffer of bytes</param>
    /// <param name="count">the number of bytes to write</param>
    public void WriteBytes(byte[] buffer, int count) {
      if (this.m_serial != null) {
        this.m_serial.Write(buffer, 0, count);
      }
    } // WriteBytes()

  } // class Serial

} // namespace serial

