using System;
using System.IO;
using System.IO.Ports;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;

namespace serial {

  /// <summary>
  /// Copies information from the serial port to IP.
  /// </summary>
  /// <remarks>
  /// Copies information arriving at the serial port to IP, not vice versa.
  /// The class can act as a TCP or UDP server or it can broadcast the
  /// information to a specified UDP port.
  /// If the server is running as UDP, then the UDP client must initiate the
  /// connection by sending anything to the server. The class IPClient sends
  /// foe example a single byte 0.
  /// </remarks>
  /// <example>
  /// TCP server, copying stream from the serial port <c>COM4</c>.
  /// Please be aware, that the serial port names are platform dependent,
  /// for example the 4th serial port on a desktop windows ma be called
  /// <c>COM4:</c> and on linux <c>/dev/ttyS3</c>. Use <c>Serial.COMPORTS</c>
  /// in order to list valid port names.
  /// <code lang="C#">
  /// using Serial.IO.Ports;
  /// using System;
  /// using serial;
  /// public static void serial2ipcopy() {
  ///   Serial2IP s;
  ///   s = new Serial2IP();
  ///   s.PortName = "COM4";
  ///   s.BaudRate = 9600; s.DataBits = 8;
  ///   s.Parity = Parity.None; s.StopBits = StopBits.One;
  ///   s.oper(ServerType.TCP, 5000);
  ///   Console.Write("Hit ENTER to close"); Console.ReadLine();
  ///   s.oper(ServerType.OFF, 0);
  /// }
  /// </code>
  /// The stream generated by this server method can be retrieved for
  /// example using
  /// <c>netcat localhost 5000</c>
  /// or <c>telnet localhost 5000</c>.
  /// The UDP server is started using parameter <c>ServerType.UDP</c> rather
  /// than <c>ServerType.TCP</c> for the call to <c>s.oper()</c>.
  /// In this case the stream from the server using for example
  /// <c>echo anything | netcat -u localhost 5000</c>.
  /// The broadcasting server is started using the parameter
  /// <c>ServerType.BCAST</c> and the stream from the server can be tested
  /// using <c>netcat -l -p 5000</c>.
  /// Netcat for unix systems can be downloaded from
  /// <A HREF="http://netcat.sourceforge.net">netcat.sourceforge.net</A>,
  /// netcat for windows - from
  /// <A HREF="http://rbytes.org/software/netcat-review/">rbytes.org/software/netcat-review</A>.
  /// The netcat executable for windows is named nc.exe rather than netcat.
  /// </example>
  public class Serial2IP : Serial {

    /// <summary>Default UDP or TCP port to use</summary>
    /// <remarks>
    /// Only informational meaning. There is no functional default port,
    /// the port number must be always explicitely set using oper().
    /// </remarks>
    public const int DEFAULTPORT = 5000;

    /// <summary>Default IP server type</summary>
    /// <remarks>
    /// Only informational meaning. There is no functional default server
    /// operation type and it must be set using oper().
    /// </remarks>
    public const ServerType DEFAULTTYP = ServerType.BCAST;

    /// <summary>Maximal number of connections for the TCP/UDP server</summary>
    private const int MAXCONNECTIONS = 10;

    /// <summary>Array of connected TCP clients</summary>
    /// <remarks>
    /// Used if working as TCP server. Placeholders for not connected clients
    /// are null. Size MAXCONNECTIONS,
    /// </remarks>
    private TcpClient[] m_tcpclnts;

    /// <summary>Array of connected UDP clients</summary>
    /// <remarks>
    /// Used if working as UDP server. Placeholders for not connected clients
    /// are null. Size MAXCONNECTIONS, If working as broadcast server, the
    /// array contains available broadcast addresses (usually 255.255.255.255).
    /// </remarks>
    private IPEndPoint[] m_udpclnts;

    /// <summary>The TCP server</summary>
    /// <remarks>
    /// Used if working as a TCP server, for UDP or broadcast this variable
    /// remains null.
    /// </remarks>
    private TcpListener m_tcpserver;

    /// <summary>The UDP server</summary>
    /// <remarks>
    /// Used if working as an UDP server or broadcast.
    /// </remarks>
    private UdpClient m_udpserver;

    /// <summary>Thread listening for incoming connections</summary>
    /// <remarks>
    /// The thread listens for incoming TCP connections or incoming
    /// initialization in the UDP mode. For UDP the client must
    /// send something (anything, one byte is enough) in order to be
    /// recognized.
    /// </remarks>
    private Thread m_lthr;

    /// <summary>Stream for logging, if logging switched on</summary>
    private StreamWriter m_logstr;

    /// <summary>If the server is active</summary>
    private bool m_isactive;

    /// <summary>Initializes an empty and switched off server</summary>
    public Serial2IP() : base() {
      int i;

      this.m_logstr = null;
      this.m_tcpserver = null;
      this.m_udpserver = null;
      this.m_lthr = null;
      this.m_isactive = false;
      this.m_tcpclnts = new TcpClient[Serial2IP.MAXCONNECTIONS];
      this.m_udpclnts = new IPEndPoint[Serial2IP.MAXCONNECTIONS];
      for(i = 0; i < Serial2IP.MAXCONNECTIONS; i++) {
        this.m_tcpclnts[i] = null;
        this.m_udpclnts[i] = null;
      }
      this.RdLineCb = null;
      this.RdByteCb = new ReadByteDelegate(this.copyFkt);
    } // Serial2IP()

    /// <summary>Destructor switches the server off</summary>
    ~Serial2IP() {
      this.oper(ServerType.OFF, 0);
    } // ~Serial2IP()

    /// <summary>If logging is switched on or off</summary>
    public bool Log {
      get {
        if (this.m_logstr != null) return true;
        return false;
      }
      set {
        if (this.m_logstr != null) {
          try { this.m_logstr.Close(); } catch(Exception) {}
        }
        if (value) {
          try {
            this.m_logstr = new StreamWriter(Utils.getExtFileName(".log"));
          }
          catch(Exception) {}
        }
        else
          this.m_logstr = null;
      }
    } // Log

    /// <summary>Logs the given information if logging switched on</summary>
    /// <param name="title">the title of the log</param>
    /// <param name="msg">the log message</param>
    private void dolog(string title, object msg) {
      if (this.m_logstr != null && title != null && msg != null) {
        try {
          this.m_logstr.WriteLine("{0:yyyyMMdd HHmmss} {1}: {2}",
                                  DateTime.Now, title, msg.ToString());
          this.m_logstr.Flush();
        }
        catch(Exception) {}
      }
    } // dolog()

    /// <summary>Starts or stops copying serial port data</summary>
    /// <remarks>
    /// If switching on, the method opens the serial port and it starts
    /// copying data from it to TCP or UDP (also broadcast) using the
    /// callback method copyFkt(). For the type TCP the thread method
    /// listeningFktTCP() is started in order to receive incoming
    /// connections. For the type UDP the thread method listeningFktUDP()
    /// is started instead. In this case the client must send something
    /// in order to be recognized by this server. For broadcast (type BCAST)
    /// there is no listening method, the method builds an UDP client with
    /// a broadcast address instead and the callback method copyFkt() writes
    /// serial data to this client only. Windows broadcasts 255.255.255.255
    /// are presently used rather than the real broadcast addresses of
    /// the networks currently defined in the system.
    /// </remarks>
    /// <param name="typ">
    /// The type of the server or OFF for switching the server off.
    /// If whitching off, the serial port will be also closed.
    /// </param>
    /// <param name="port">
    /// The TCP or UDP port name. The server listenes on this port for
    /// types TCP and UDP or sends to this port for type BCAST.
    /// This variable has no meaning, if the the server is switched off.
    /// </param>
    public void oper(ServerType typ, int port) {
      int i;

      if (this.Log) this.dolog("oper", string.Format("{0} {1}", typ, port));
      switch(typ) {
        case ServerType.OFF :
          if (this.m_tcpserver != null && this.m_isactive) {
            lock(this) {
              for(i = 0; i < this.m_tcpclnts.Length; i++) {
                if (this.m_tcpclnts[i] != null) {
                  if (this.m_tcpclnts[i].Client.Connected) {
                    try { this.m_tcpclnts[i].Close(); } catch(Exception) {}
                  }
                  this.m_tcpclnts[i] = null;
                }
              } // for i
              this.m_isactive = false;
              this.m_tcpserver.Stop();
            } // lock
          } // if
          if (this.m_udpserver != null && this.m_isactive) {
            lock(this) {
              for(i = 0; i < this.m_tcpclnts.Length; i++)
                this.m_udpclnts[i] = null;
              this.m_isactive = false;
              this.m_udpserver.Close();
            } // lock
          }
          if (this.m_lthr != null) this.m_lthr.Join();
          this.m_lthr = null;
          this.m_tcpserver = null;
          this.m_udpserver = null;
          this.Close();
          break;
        case ServerType.TCP :
          this.oper(ServerType.OFF, port);
          this.m_tcpserver = new TcpListener(IPAddress.Any, port);
          this.m_tcpserver.Start();
          this.m_isactive = true;
          this.m_lthr = new Thread(this.listeningFktTCP);
          this.m_lthr.Start();
          this.Open();
          this.Read = true;
          break;
        case ServerType.UDP :
          this.oper(ServerType.OFF, port);
          this.m_udpserver = new UdpClient(port);
          this.m_isactive = true;
          this.m_lthr = new Thread(this.listeningFktUDP);
          this.m_lthr.Start();
          this.Open();
          this.Read = true;
          break;
        case ServerType.BCAST :
          this.oper(ServerType.OFF, port);
          this.m_udpserver = new UdpClient();
          this.m_udpclnts[0] = new IPEndPoint(IPAddress.Broadcast, port);
          this.m_isactive = true;
          this.Open();
          this.Read = true;
          break;
        default :
          throw new ArgumentException(string.Format("Type {0} not implemented yet",
                                                    typ));
      } // switch
    } // oper()

    /// <summary>Thread method listening on incoming TCP connections</summary>
    private void listeningFktTCP() {
      NetworkStream ns;
      TcpClient     clnt;
      byte[]        b;
      int           i;
      bool          fnd;

      while(this.m_tcpserver != null && this.m_isactive) {
        try {
          clnt = this.m_tcpserver.AcceptTcpClient();
          this.dolog("listeningFktTCP", clnt.Client.RemoteEndPoint);
          fnd = false;
          lock(this) {
            for(i = 0; i < this.m_tcpclnts.Length; i++) {
              if (this.m_tcpclnts[i] == null) {
                if (!fnd) {
                  this.m_tcpclnts[i] = clnt;
                  fnd = true;
                }
              }
              else {
                if (!this.m_tcpclnts[i].Client.Connected) {
                  if (fnd) {
                    this.m_tcpclnts[i] = null;
                  }
                  else {
                    this.m_tcpclnts[i] = clnt;
                    fnd = true;
                  }
                }
              }
            } // for i
          } // lock
          if (!fnd) {
            try {
              b = Encoding.ASCII.GetBytes("Maximum number of connections exceeded\r\n");
              this.dolog("listeningFktTCP", "Maximum number of connections exceeded");
              ns = clnt.GetStream();
              ns.Write(b, 0, b.Length);
              clnt.Close();
            }
            catch(Exception) {}
          }
        }
        catch(Exception excp) {
          if (this.m_tcpserver != null && this.m_isactive) {
            this.dolog("listeningFktTCP", excp);
          }
        }
      } // while
    } // listeningFktTCP()

    /// <summary>Thread method for listening on incoming UDP connections</summary>
    /// <remarks>
    /// The UDP client must send something in order to be recognized.
    /// </remarks>
    private void listeningFktUDP() {
      IPEndPoint clnt;
      byte[]     b;
      int        i;
      bool       fnd;

      while(this.m_udpserver != null && this.m_isactive) {
        try {
          clnt = new IPEndPoint(IPAddress.Any, 0);
          this.m_udpserver.Receive(ref clnt);
          this.dolog("listeningFktUDP", clnt);
          fnd = false;
          lock(this) {
            for(i = 0; i < this.m_udpclnts.Length; i++) {
              if (this.m_udpclnts[i] != null && this.m_udpclnts[i].Equals(clnt)) {
                fnd = true;
                break;
              }
            } // for
            if (!fnd) {
              for(i = 0; i < this.m_udpclnts.Length; i++) {
                if (this.m_udpclnts[i] == null) {
                  this.m_udpclnts[i] = clnt;
                  fnd = true;
                  break;
                }
              } // for
            } // if (!fnd)
          } // lock
          if (!fnd) {
            try {
              b = Encoding.ASCII.GetBytes("Maximum number of connections exceeded\r\n");
              this.dolog("listeningFktUDP", "Maximum number of connections exceeded");
              this.m_udpserver.Send(b, b.Length, clnt);
            }
            catch(Exception) {}
          }
        }
        catch(Exception excp) {
          if (this.m_udpserver != null && this.m_isactive) {
            this.dolog("listeningFktUDP", excp);
          }
        }
      } // while
    } // listeningFktUDP()

    /// <summary>Sends data from the serial port</summary>
    /// <remarks>
    /// Sends the given data from the serial port to any connected client.
    /// This method is called as a callback for bytes retrieved from the
    /// serial port (RdByteCb).
    /// </remarks>
    /// <param name="b">array of bytes read from the serial port</param>
    /// <param name="len">number of valid bytes in b</param>
    private void copyFkt(byte[] b, int len) {
      TcpClient     clnttcp;
      IPEndPoint    clntudp;
      NetworkStream ns;
      int           i;

      try {
        if (this.m_isactive) {
          if (this.m_tcpserver != null) {
            lock(this) {
              for(i = 0; i < this.m_tcpclnts.Length; i++) {
                clnttcp = this.m_tcpclnts[i];
                if (clnttcp != null) {
                  if (clnttcp.Client.Connected) {
                    try {
                      ns = clnttcp.GetStream();
                      ns.Write(b, 0, len);
                    }
                    catch(Exception excp) {
                      if (!clnttcp.Client.Connected) {
                        this.m_tcpclnts[i] = null;
                      }
                      else {
                        this.dolog("copyFkt write TCP", excp);
                      }
                    }
                  }
                  else {
                    this.m_tcpclnts[i] = null;
                  }
                }
              } // for i
            } // lock
          } // if (this.m_tcpserver != null)
          if (this.m_udpserver != null) {
            lock(this) {
              for(i = 0; i < this.m_tcpclnts.Length; i++) {
                clntudp = this.m_udpclnts[i];
                if (clntudp != null) {
                  try {
                    this.m_udpserver.Send(b, len, clntudp);
                  }
                  catch(Exception) {
                    this.m_udpclnts[i] = null;
                  }
                }
              } // for i
            } // lock
          } // if (this.m_udpserver != null)
        } // if (this.m_isactive)
      }
      catch(Exception excp) {
        this.dolog("copyFkt", excp);
      }
    } // copyFkt()
  } // class Serial2IP

} // namespace serial

