using System;
using System.Drawing;
using System.IO;
using System.IO.Ports;
using System.Reflection;
using System.Text;
using System.Windows.Forms;

namespace serial {

  /// <summary>The main form for the serial 2 IP copy application</summary>
  /// <remarks>
  /// <IMG SRC="s2ip.png"/>
  /// <list type="table">
  /// <item>
  ///  <term>m_ports</term>
  ///  <description>
  ///   The name of the serial port to use.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_baud</term>
  ///  <description>
  ///   The baud rate for the serial port communications.
  ///   9600 or 4800 could be plasible choices.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_databits</term>
  ///  <description>
  ///   Number of databits for the serial port. For example 8.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_parities</term>
  ///  <description>
  ///   The parity for the serial port, for example None.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_stopbits</term>
  ///  <description>
  ///   Number of stopbits for the serial port. One can be a plausible choice.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_ipmodes</term>
  ///  <description>
  ///   IP server mode, in which the serial port data is served through
  ///   IP. TCP or UDP - act as TCP or UDP server on listen for connections
  ///   on the port specified in m_ipport. BCAST - broadcast data to
  ///   to the UDP port specified in m_ipport. See also documentation on
  ///   the class Serial2IP for more details on the IP modes.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_ipport</term>
  ///  <description>
  ///   The TCP or UDP port to listen for incoming connections, if
  ///   TCP or UDP has been selected in m_ipmodes. The UDP port to which
  ///   to broadcast the serial port data, if BCAST has been selected in
  ///   m_ipmodes.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_status</term>
  ///  <description>
  ///   on, if copying data from serial port to IP has been started, off,
  ///   if it is inactive.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Port/Open, m_menuopen</term>
  ///  <description>
  ///   Starts copying data from the serial port selected in m_ports
  ///   to IP, using protocol selected in m_ipmodes and listening on
  ///   port or broadcasting to port defined in m_ipport.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Port/Close</term>
  ///  <description>
  ///   Closes the serial port and the IP server.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Port/Log, m_menulog</term>
  ///  <description>
  ///   Starts or stops logging, which may be used for investigating failures.
  ///   Especially exceptions and user interactions are logged.
  ///   Application logs to a file s2ip&lt;yyyyMMddHHmmss&gt;.log located
  ///   in the same directory as the execcutable. yyyyMMddHHmmss is the
  ///   timestamp at which logging has been switched on.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Port/Exit</term>
  ///  <description>
  ///   Exits the application.
  ///  </description>
  /// </item>
  /// </list>
  /// Warning! s2ip.exe only copies data from the serial port to IP.
  /// The program is inable to copy stream coming from IP into the serial
  /// port. Thus you can't for example enter PSRF103 sequences using IP.
  /// </remarks>
  /// <example>
  /// <IMG SRC="s2ip-example.png"/>
  /// <list type="bullet">
  /// <item>
  /// Estimate an IP connection between the desktop computer and the PDA.
  /// Refer to the Active Sync documentation on how to do it with MS Windows
  /// and Active Sync. Refer to
  /// <A HREF="http://www.pocketpcfaq.com/wce/linux-serial.htm">http://www.pocketpcfaq.com/wce/linux-serial.htm</A>
  /// on how to estimate a PPP connection between the PDA and Unix.
  /// </item>
  /// <item>
  /// Select valid serial port parameters in m_ports, m_baud, m_databits,
  /// m_parities and m_stopbits according to your GPS device. On HTC3300
  /// (aka. T-Mobile MDA Compact III) it could be for example COM4,
  /// baud 9600, 8 databits, no parity, one stop bit.
  /// </item>
  /// <item>
  /// Select IP mode TCP, port 5000.
  /// </item>
  /// <item>
  /// Select menu item Port/Open. "on" should appear in the label m_status.
  /// </item>
  /// <item>
  /// If using Active Sync with network enabled, the IP address of the 
  /// PDA device is usually 169.254.2.1, check ipconfig on your Windows PC
  /// in doubt. The IP address is surely known to you, if you estimated a
  /// PPP connection to a unix server. Connect to the PDA using for example:
  /// <code>telnet 169.254.2.1 5000</code>
  /// and watch incoming NMEA sequences.
  /// </item>
  /// <item>
  /// You can also use sg.exe or sgps.exe, configured for TCP, port 5000
  /// and the IP address above as the name or address of the server, started
  /// on your desktop computer.
  /// </item>
  /// </list>
  /// </example>
  /// <seealso cref="Serial2IP">Serial2IP</seealso>
  /// <seealso cref="Serial2IPGUI">Serial2IPGUI</seealso>
  public class Serial2IPForm : Form {

    /// <summary>Serial 2 IP copy engine</summary>
    private Serial2IP m_serial;

    /// <summary>Names of COM ports available in the system</summary>
    private ComboBox m_ports;

    /// <summary>Control for the baud rate of the port</summary>
    private ComboBox m_baud;

    /// <summary>Control for the number of data bits.</summary>
    private ComboBox m_databits;

    /// <summary>Available COM port parities</summary>
    private ComboBox m_parities;

    /// <summary>Available COM port parities</summary>
    private ComboBox m_stopbits;

    /// <summary>Available IP server modes</summary>
    private ComboBox m_ipmodes;

    /// <summary>TCP/UDP port</summary>
    private TextBox m_ipport;

    /// <summary>Status information: active, inactive</summary>
    private Label m_status;

    /// <summary>Menu item for starting the operation</summary>
    private MenuItem m_menuopen;

    /// <summary>Menu item for switching logging on or off</summary>
    private MenuItem m_menulog;

    /// <summary>Name of the config file</summary>
    private static string CONFIGFILE {
      get {
        return Utils.getExtFileName(".cfg");
      }
    } // CONFIGFILE

    /// <summary>
    /// Common method for showing exception errors occured in this form.
    /// </summary>
    /// <param name="aTitle">The title of the exception</param>
    /// <param name="anExcp">The excpetion itself</param>
    private static void printExcp(string aTitle, Exception anExcp) {
      StringBuilder sb;

      sb = new StringBuilder(aTitle);
      sb.Append('\n');
      sb.Append(anExcp.ToString());
      MessageBox.Show(sb.ToString(), "Error");
    } // printExcp()

    /// <summary>Reads configuration into combos</summary>
    /// <remarks>
    /// If the configuration can'b read for some reason, then defaults
    /// are assumed.
    /// </remarks>
    private void readCfg() {
      StreamReader stream;
      string       comport, baudrate, databits, parity, stopbits,
                   ipmode, ipport;

      stream = null;
      comport = null;
      baudrate = null;
      databits = null;
      parity = null;
      stopbits = null;
      ipmode = null;
      ipport = null;
      try {
        stream = new StreamReader(Serial2IPForm.CONFIGFILE);
        comport = stream.ReadLine();
        baudrate = stream.ReadLine();
        databits = stream.ReadLine();
        parity = stream.ReadLine();
        stopbits = stream.ReadLine();
        ipmode = stream.ReadLine();
        ipport = stream.ReadLine();
      }
      catch(Exception) { }
      if (stream != null) stream.Close();
      if (comport == null) comport = Serial.DEFAULT_COMPORT;
      if (baudrate == null) baudrate = Serial.DEFAULT_BAUDRATE.ToString();
      if (databits == null) databits = Serial.DEFAULT_DATABITS.ToString();
      if (parity == null) parity = Serial.DEFAULT_PARITY.ToString();
      if (stopbits == null) stopbits = Serial.DEFAULT_STOPBITS.ToString();
      if (ipmode == null) ipmode = Serial2IP.DEFAULTTYP.ToString();
      if (ipport == null) ipport = Serial2IP.DEFAULTPORT.ToString();
      this.m_ports.Text = comport;
      this.m_baud.Text = baudrate;
      this.m_databits.Text = databits;
      this.m_parities.Text = parity;
      this.m_stopbits.Text = stopbits;
      this.m_ipmodes.Text = ipmode;
      this.m_ipport.Text = ipport;
    } // readCfg()

    /// <summary>Writes setting to the configuration file</summary>
    private void writeCfg() {
      StreamWriter stream;

      stream = null;
      try {
        stream = new StreamWriter(Serial2IPForm.CONFIGFILE);
        stream.WriteLine(this.m_ports.Text);
        stream.WriteLine(this.m_baud.Text);
        stream.WriteLine(this.m_databits.Text);
        stream.WriteLine(this.m_parities.Text);
        stream.WriteLine(this.m_stopbits.Text);
        stream.WriteLine(this.m_ipmodes.Text);
        stream.WriteLine(this.m_ipport.Text);
      }
      catch(Exception) { }
      if (stream != null) stream.Close();
    } // writeCfg()

    /// <summary>Event method called if the form is closed</summary>
    /// <remarks>
    /// Writes current config to the file and switches to copying
    /// of serial port data off.
    /// </remarks>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void onClose(object sender, EventArgs e) {
      Exception ee;

      ee = null;
      try { 
        this.m_serial.oper(ServerType.OFF, 0);
      }
      catch(Exception excp) {
        ee = excp;
      }
      try { this.writeCfg(); } catch(Exception excp) { ee = excp; }
      if (ee != null) Serial2IPForm.printExcp("exit", ee);
    } // onClose()

    /// <summary>Opens the opened serial port and IP server</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuOpen_Click(object sender, EventArgs e) {
      ServerType typ;
      int        ipport;

      try {
        this.m_serial.oper(ServerType.OFF, 0);
        this.m_serial.PortName = this.m_ports.Text;
        this.m_serial.BaudRate = int.Parse(this.m_baud.Text);
        this.m_serial.DataBits = int.Parse(this.m_databits.Text);
        this.m_serial.Parity = (Parity) Enum.Parse(typeof(Parity),
                                                   this.m_parities.Text,
                                                   false);
        this.m_serial.StopBits = (StopBits) Enum.Parse(typeof(StopBits),
                                                       this.m_stopbits.Text,
                                                       false);
        typ = (ServerType) Enum.Parse(typeof(ServerType),
                                      this.m_ipmodes.Text,
                                      false);
        ipport = int.Parse(this.m_ipport.Text);
        this.m_serial.oper(typ, ipport);
        this.m_status.Text = "on";
        this.m_menuopen.Checked = true;
      }
      catch(Exception excp) {
        Serial2IPForm.printExcp("close", excp);
      }
    } // menuOpenClick()

    /// <summary>Closes the opened serial port and IP server</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuClose_Click(object sender, EventArgs e) {
      try {
        this.m_serial.oper(ServerType.OFF, 0);
        this.m_status.Text = "off";
        this.m_menuopen.Checked = false;
      }
      catch(Exception excp) {
        Serial2IPForm.printExcp("close", excp);
      }
    } // menuOpenClick()

    /// <summary>Switches logging on or off</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuLog_Click(object sender, EventArgs e) {
      try {
        if (this.m_serial.Log) {
          this.m_serial.Log = false;
          this.m_menulog.Checked = false;
        }
        else {
          this.m_serial.Log = true;
          this.m_menulog.Checked = true;
        }
      }
      catch(Exception excp) {
        Serial2IPForm.printExcp("log", excp);
      }
    } // menuLog_Click()

    /// <summary>Exits the applications</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuExit_Click(object sender, EventArgs e) {
      this.Close();
    } // menuOpenClick()

    /// <summary>Initializes form controls</summary>
    private void InitializeComponent() {
      MainMenu menu;
      MenuItem menuPort, mitem;

      this.SuspendLayout();
      menuPort = new MenuItem();
      menuPort.Text = "Port";
      menuPort.MenuItems.Add(this.m_menuopen);
      mitem = new MenuItem();
      mitem.Text = "Close";
      mitem.Click += this.menuClose_Click;
      menuPort.MenuItems.Add(mitem);
      menuPort.MenuItems.Add(this.m_menulog);
      mitem = new MenuItem();
      mitem.Text = "Exit";
      mitem.Click += this.menuExit_Click;
      menuPort.MenuItems.Add(mitem);
      menu = new MainMenu();
      menu.MenuItems.Add(menuPort);
      this.m_ports.Location = new Point(5, 5);
      this.m_ports.Size = new Size(60, 30);
      this.m_baud.Location = new Point(65, 5);
      this.m_baud.Size = new Size(50, 30);
      this.m_databits.Location = new Point(115, 5);
      this.m_databits.Size = new Size(40, 30);
      this.m_parities.Location = new Point(155, 5);
      this.m_parities.Size = new Size(50, 30);
      this.m_stopbits.Location = new Point(205, 5);
      this.m_stopbits.Size = new Size(50, 30);
      this.m_ipmodes.Location = new Point(5, 35);
      this.m_ipmodes.Size = new Size(60, 30);
      this.m_ipport.Location = new Point(65, 35);
      this.m_ipport.Size = new Size(50, 30);
      this.m_status.Location = new Point(120, 35);
      this.m_status.Size = new Size(50, 30);
      this.m_status.Text = "off";
      this.AutoScroll = true;
      this.Controls.Clear();
      this.Controls.Add(this.m_ports);
      this.Controls.Add(this.m_baud);
      this.Controls.Add(this.m_databits);
      this.Controls.Add(this.m_parities);
      this.Controls.Add(this.m_stopbits);
      this.Controls.Add(this.m_ipmodes);
      this.Controls.Add(this.m_ipport);
      this.Controls.Add(this.m_status);
      this.Name = "Serial2IPForm";
      this.Text = "Serial 2 IP copy";
      this.Menu = menu;
      this.ResumeLayout(false);
      this.Closing += this.onClose;
    } // InitializeComponent()

    /// <summary>Initializes the form</summary>
    public Serial2IPForm() {
      Icon     ico;
      Stream   str;

      this.m_serial = new Serial2IP();
      this.m_ports = new ComboBox();
      this.m_baud = new ComboBox();
      this.m_databits = new ComboBox();
      this.m_parities = new ComboBox();
      this.m_stopbits = new ComboBox();
      this.m_ipmodes = new ComboBox();
      this.m_ipport = new TextBox();
      this.m_status = new Label();
      this.m_menuopen = new MenuItem();
      this.m_menuopen.Text = "Open";
      this.m_menuopen.Click += this.menuOpen_Click;
      this.m_menulog = new MenuItem();
      this.m_menulog.Text = "Log";
      this.m_menulog.Click += this.menuLog_Click;
      this.InitializeComponent();
      // fill combos with all available values
      foreach(string s in Serial.COMPORTS) this.m_ports.Items.Add(s);
      foreach(int i in Serial.BAUDRATES) this.m_baud.Items.Add(i);
      foreach(int i in Serial.DATABITS) this.m_databits.Items.Add(i);
      this.m_parities.Items.Add(Parity.Even.ToString());
      this.m_parities.Items.Add(Parity.Mark.ToString());
      this.m_parities.Items.Add(Parity.None.ToString());
      this.m_parities.Items.Add(Parity.Odd.ToString());
      this.m_parities.Items.Add(Parity.Space.ToString());
      this.m_stopbits.Items.Add(StopBits.None.ToString());
      this.m_stopbits.Items.Add(StopBits.One.ToString());
      this.m_stopbits.Items.Add(StopBits.OnePointFive.ToString());
      this.m_stopbits.Items.Add(StopBits.Two.ToString());
      this.m_ipmodes.Items.Add(ServerType.TCP.ToString());
      this.m_ipmodes.Items.Add(ServerType.UDP.ToString());
      this.m_ipmodes.Items.Add(ServerType.BCAST.ToString());
      this.readCfg();
      try {
        str = Assembly.GetExecutingAssembly().GetManifestResourceStream("Marta3.ico");
        ico = new Icon(str);
        str.Close();
        this.Icon = ico;
      }
      catch(Exception) { }
    } // Serial2IPForm

  } // class Serial2IPForm

} // namespace serial

