using System;
using System.IO.Ports;
using System.Text;

namespace serial {

  /// <summary>Copies data from one port to another</summary>
  /// <remarks>
  /// Trnasparently copies data from one serial port to another in both
  /// directions.
  /// </remarks>
  /// <example>
  /// The following example copies data between serial port <c>COM1</c>
  /// and <c>COM2</c>, both ports using current or default baud, number
  /// of databits, stopbits and parity.
  /// <code lang="C#">
  /// using System;
  /// using serial;
  /// public static void portCopy() {
  ///   SerialCpy s;
  ///   s = new SerialCpy();
  ///   s.PortName1 = "COM1"; s.PortName2 = "COM2";
  ///   s.Open();
  ///   Console.Write("Hit ENTER to close"); Console.ReadLine();
  ///   s.Close();
  /// }
  /// </code>
  /// </example>
  public class SerialCpy {

    /// <summary>2 serial ports</summary>
    private Serial[] m_serial;

    /// <summary>Creates instances of two serial ports</summary>
    public SerialCpy() {
      this.m_serial = new Serial[2];
      this.m_serial[0] = new Serial();
      this.m_serial[1] = new Serial();     
      this.m_serial[0].RdByteCb = new ReadByteDelegate(this.read1);
      this.m_serial[1].RdByteCb = new ReadByteDelegate(this.read2);
    } // SerialCpy()

    /// <summary>Closes the serial ports if needed</summary>
    ~SerialCpy() {
      this.Close();
    } // ~SerialCpy

    /// <summary>Returns the string representation of the objects</summary>
    /// <returns>The string representation</returns>
    public override string ToString() {
      StringBuilder sb;

      sb = new StringBuilder();
      sb.Append(this.m_serial[0].ToString());
      sb.Append('\n');
      sb.Append(this.m_serial[1].ToString());
      return sb.ToString();
    } // ToString()

    /// <summary>Baudrate of the 1st port</summary>
    public int BaudRate1 {
      set {
        this.m_serial[0].BaudRate = value;
      }
      get {
        return this.m_serial[0].BaudRate;
      }
    } // BaudRate1

    /// <summary>Baudrate of the 2nd port</summary>
    public int BaudRate2 {
      set {
        this.m_serial[1].BaudRate = value;
      }
      get {
        return this.m_serial[1].BaudRate;
      }
    } // BaudRate2

    /// <summary>Databits of the 1st port</summary>
    public int DataBits1 {
      set {
        this.m_serial[0].DataBits = value;
      }
      get {
        return this.m_serial[0].DataBits;
      }
    } // DataBits1

    /// <summary>Databits of the 2nd port</summary>
    public int DataBits2 {
      set {
        this.m_serial[1].DataBits = value;
      }
      get {
        return this.m_serial[1].DataBits;
      }
    } // DataBits2

    /// <summary>Parity of the 1st port</summary>
    public Parity Parity1 {
      set {
        this.m_serial[0].Parity = value;
      }
      get {
        return this.m_serial[0].Parity;
      }
    } // Parity1

    /// <summary>Parity of the 2nd port</summary>
    public Parity Parity2 {
      set {
        this.m_serial[1].Parity = value;
      }
      get {
        return this.m_serial[1].Parity;
      }
    } // Parity2

    /// <summary>Name of the 1st port</summary>
    public string PortName1 {
      set {
        this.m_serial[0].PortName = value;
      }
      get {
        return this.m_serial[0].PortName;
      }
    } // PortName1

    /// <summary>Name of the 2nd port</summary>
    public string PortName2 {
      set {
        this.m_serial[1].PortName = value;
      }
      get {
        return this.m_serial[1].PortName;
      }
    } // PortName2

    /// <summary>Stop bits of the 1st port</summary>
    public StopBits StopBits1 {
      set {
        this.m_serial[0].StopBits = value;
      }
      get {
        return this.m_serial[0].StopBits;
      }
    } // StopBits1

    /// <summary>Stop bits of the 2nd port</summary>
    public StopBits StopBits2 {
      set {
        this.m_serial[1].StopBits = value;
      }
      get {
        return this.m_serial[1].StopBits;
      }
    } // StopBits2

    /// <summary>Timeout of the 1st port</summary>
    public int Timeout1 {
      set {
        this.m_serial[0].Timeout = value;
      }
      get {
        return this.m_serial[0].Timeout;
      }
    } // Timeout1

    /// <summary>Timeout of the 2nd port</summary>
    public int Timeout2 {
      set {
        this.m_serial[1].Timeout = value;
      }
      get {
        return this.m_serial[1].Timeout;
      }
    } // Timeout2

    /// <summary>Closes serial ports, if opened</summary>
    public void Close() {
      foreach(Serial s in this.m_serial) {
        s.Close();
      }
    } // Close()

    /// <summary>Opens both ports and starts copying streams</summary>
    public void Open() {
      this.Close();
      try {
        foreach(Serial s in this.m_serial) s.Open();
        foreach(Serial s in this.m_serial) s.Read = true;
      }
      catch(Exception excp) {
        try { this.Close(); } catch(Exception) {}
        throw excp;
      }
    } // Open()

    /// <summary>Checks if both ports are opened</summary>
    public bool IsOpen {
      get {
        foreach(Serial s in this.m_serial) {
          if (!s.IsOpen) return false;
        }
        return true;
      }
    } // IsOpen

    /// <summary>Handling of byte read from serial port int1</summary>
    /// <remarks>
    /// Writes the received bytes to m_serial[portidx], if not empty or null.
    /// </remarks>
    /// <param name="portidx">Index of the port receiving bytes read</param>
    /// <param name="b">the array of bytes to handle</param>
    /// <param name="len">number of valid bytes in b</param>
    private void commonRead(int portidx, byte[] b, int len) {
      if (b != null && len > 0)
        this.m_serial[portidx].WriteBytes(b, len);
    } // commonRead()

    /// <summary>Read callback for port 1</summary>
    /// <remarks>Call commonRead() with index 0</remarks>
    /// <param name="b">Array of bytes read on port 1</param>
    /// <param name="len">Number of valid bytes in b</param>
    private void read1(byte[] b, int len) {
      this.commonRead(1, b, len);
    } // read1()

    /// <summary>Read callback for port 2</summary>
    /// <remarks>Call commonRead() with index 0</remarks>
    /// <param name="b">Array of bytes read on port 2</param>
    /// <param name="len">Number of valid bytes in b</param>
    private void read2(byte[] b, int len) {
      this.commonRead(0, b, len);
    } // read2()
      
  } // class SerialCpy

} // namespace serial

