using System;
using System.Drawing;
using System.IO;
using System.IO.Ports;
using System.Reflection;
using System.Text;
using System.Windows.Forms;

namespace serial {

  /// <summary>
  /// Form for the application scpy.exe copying one serial port to another.
  /// </summary>
  /// <seealso cref="SerialCpy">SerialCpy</seealso>
  /// <seealso cref="SerialCpyGUI">SerialCpyGUI</seealso>
  public class SerialCpyForm : Form {

    /// <summary>Name of the config file</summary>
    private static string CONFIGFILE {
      get {
        return Utils.getExtFileName(".cfg");
      }
    } // CONFIGFILE

    /// <summary>Serial port names of both ports</summary>
    private ComboBox[] m_ports;

    /// <summary>Baud rates for both ports</summary>
    private ComboBox[] m_baud;

    /// <summary>Number of databits for both ports</summary>
    private ComboBox[] m_databits;

    /// <summary>Parities for both ports</summary>
    private ComboBox[] m_parities;

    /// <summary>Number of stop bits for both ports</summary>
    private ComboBox[] m_stopbits;

    /// <summary>Status of the serial ports, if opened</summary>
    private Label      m_status;

    /// <summary>The serial copy engine</summary>
    private SerialCpy  m_serialcpy;

    /// <summary>
    /// Common method for showing exception errors occured in this form.
    /// </summary>
    /// <param name="aTitle">The title of the exception</param>
    /// <param name="anExcp">The excpetion itself</param>
    private void printExcp(string aTitle, Exception anExcp) {
      StringBuilder sb;

      sb = new StringBuilder(aTitle);
      sb.Append('\n');
      sb.Append(anExcp.ToString());
      MessageBox.Show(sb.ToString(), "Error");
    } // printExcp()

    /// <summary>Called on closing the form, closes serial ports</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void onClose(object sender, EventArgs e) {
      Exception ee;

      ee = null;
      try { this.m_serialcpy.Close(); } catch(Exception excp) { ee = excp; }
      try { this.writeCfg(); } catch(Exception excp) { ee = excp; }
      if (ee != null) this.printExcp("exit", ee);
    } // onClose()

    /// <summary>Opens the serial ports and starts copying</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuOpen_Click(object sender, EventArgs e) {
      try {
        this.m_serialcpy.PortName1 = this.m_ports[0].Text;
        this.m_serialcpy.PortName2 = this.m_ports[1].Text;
        this.m_serialcpy.BaudRate1 = int.Parse(this.m_baud[0].Text);
        this.m_serialcpy.BaudRate2 = int.Parse(this.m_baud[1].Text);
        this.m_serialcpy.DataBits1 = int.Parse(this.m_databits[0].Text);
        this.m_serialcpy.DataBits2 = int.Parse(this.m_databits[1].Text);
        this.m_serialcpy.Parity1 = (Parity) Enum.Parse(typeof(Parity),
                                                       this.m_parities[0].Text,
                                                       false);
        this.m_serialcpy.Parity2 = (Parity) Enum.Parse(typeof(Parity),
                                                       this.m_parities[1].Text,
                                                       false);
        this.m_serialcpy.StopBits1 = (StopBits) Enum.Parse(typeof(StopBits),
                                                           this.m_stopbits[0].Text,
                                                           false);
        this.m_serialcpy.StopBits2 = (StopBits) Enum.Parse(typeof(StopBits),
                                                           this.m_stopbits[1].Text,
                                                           false);
        this.m_serialcpy.Open();
        this.m_status.Text = this.m_serialcpy.ToString();
      }
      catch(Exception excp) {
        this.printExcp("open", excp);
      }
    } // menuOpen_Click()

    /// <summary>Closes the serial ports</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuClose_Click(object sender, EventArgs e) {
      try {
        this.m_serialcpy.Close();
        this.m_status.Text = string.Empty;
      }
      catch(Exception excp) {
        this.printExcp("close", excp);
      }
    } // menuClose_Click()

    /// <summary>Exits the application</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuExit_Click(object sender, EventArgs e) {
      this.Close();
    } // menuExit_Click()

    /// <summary>
    /// Reads initial contents of the controls from the config file
    /// </summary>
    private void readCfg() {
      StreamReader stream;
      string[]     s;
      int          i;

      s = new string[10]; // comport, baudrate, databits, parity, stopbits * 2
      for(i = 0; i < s.Length; i++) s[i] = null;
      stream = null;
      try {
        stream = new StreamReader(SerialCpyForm.CONFIGFILE);
        for(i = 0; i < s.Length; i++) s[i] = stream.ReadLine();
      }
      catch(Exception) {}
      if (stream != null) stream.Close();
      if (s[0] == null) s[0] = Serial.DEFAULT_COMPORT;
      if (s[1] == null) s[1] = Serial.DEFAULT_COMPORT;
      if (s[2] == null) s[2] = Serial.DEFAULT_BAUDRATE.ToString();
      if (s[3] == null) s[3] = Serial.DEFAULT_BAUDRATE.ToString();
      if (s[4] == null) s[4] = Serial.DEFAULT_DATABITS.ToString();
      if (s[5] == null) s[5] = Serial.DEFAULT_DATABITS.ToString();
      if (s[6] == null) s[6] = Serial.DEFAULT_PARITY.ToString();
      if (s[7] == null) s[7] = Serial.DEFAULT_PARITY.ToString();
      if (s[8] == null) s[8] = Serial.DEFAULT_STOPBITS.ToString();
      if (s[9] == null) s[9] = Serial.DEFAULT_STOPBITS.ToString();
      this.m_ports[0].Text = s[0];
      this.m_ports[1].Text = s[1];
      this.m_baud[0].Text = s[2];
      this.m_baud[1].Text = s[3];
      this.m_databits[0].Text = s[4];
      this.m_databits[1].Text = s[5];
      this.m_parities[0].Text = s[6];
      this.m_parities[1].Text = s[7];
      this.m_stopbits[0].Text = s[8];
      this.m_stopbits[1].Text = s[9];
    } // readCfg()

    /// <summary>
    /// Writes current contents of the controls into the config file
    /// </summary>
    private void writeCfg() {
      StreamWriter stream;

      stream = null;
      try {
        stream = new StreamWriter(SerialCpyForm.CONFIGFILE);
        stream.WriteLine(this.m_ports[0].Text);
        stream.WriteLine(this.m_ports[1].Text);
        stream.WriteLine(this.m_baud[0].Text);
        stream.WriteLine(this.m_baud[1].Text);
        stream.WriteLine(this.m_databits[0].Text);
        stream.WriteLine(this.m_databits[1].Text);
        stream.WriteLine(this.m_parities[0].Text);
        stream.WriteLine(this.m_parities[1].Text);
        stream.WriteLine(this.m_stopbits[0].Text);
        stream.WriteLine(this.m_stopbits[1].Text);
      }
      catch(Exception) {}
      if (stream != null) stream.Close();
    } // writeCfg()

    /// <summary>Layout of the form</summary>
    private void InitializeComponent() {
      MainMenu   menu;
      MenuItem[] mitem;
      MenuItem   menuPort;

      this.SuspendLayout();
      mitem = new MenuItem[3];
      mitem[0] = new MenuItem();
      mitem[0].Text = "Open";
      mitem[0].Click += menuOpen_Click;
      mitem[1] = new MenuItem();
      mitem[1].Text = "Close";
      mitem[1].Click += menuClose_Click;
      mitem[2] = new MenuItem();
      mitem[2].Text = "Exit";
      mitem[2].Click += menuExit_Click;
      menuPort = new MenuItem();
      menuPort.Text = "Port";
      foreach(MenuItem mit in mitem) menuPort.MenuItems.Add(mit);
      menu = new MainMenu();
      menu.MenuItems.Add(menuPort);
      this.Name = "SerialCpyForm";
      this.Text = "Serial Copy";
      this.Menu = menu;
      this.m_status.Location = new Point(5,2);
      this.m_status.Size = new Size(260, 60);
      this.m_ports[0].Location = new Point(5, 65);
      this.m_ports[0].Size = new Size(60, 30);
      this.m_baud[0].Location = new Point(65, 65);
      this.m_baud[0].Size = new Size(50, 30);
      this.m_databits[0].Location = new Point(115, 65);
      this.m_databits[0].Size = new Size(40, 30);
      this.m_parities[0].Location = new Point(155, 65);
      this.m_parities[0].Size = new Size(50, 30);
      this.m_stopbits[0].Location = new Point(205, 65);
      this.m_stopbits[0].Size = new Size(50, 30);
      this.m_ports[1].Location = new Point(5, 100);
      this.m_ports[1].Size = new Size(60, 30);
      this.m_baud[1].Location = new Point(65, 100);
      this.m_baud[1].Size = new Size(50, 30);
      this.m_databits[1].Location = new Point(115, 100);
      this.m_databits[1].Size = new Size(40, 30);
      this.m_parities[1].Location = new Point(155, 100);
      this.m_parities[1].Size = new Size(50, 30);
      this.m_stopbits[1].Location = new Point(205, 100);
      this.m_stopbits[1].Size = new Size(50, 30);
      this.ClientSize = new Size(270, 140);
      this.AutoScroll = true;
      this.Controls.Add(this.m_status);
      foreach(Control c in this.m_ports) this.Controls.Add(c);
      foreach(Control c in this.m_baud) this.Controls.Add(c);
      foreach(Control c in this.m_databits) this.Controls.Add(c);
      foreach(Control c in this.m_parities) this.Controls.Add(c);
      foreach(Control c in this.m_stopbits) this.Controls.Add(c);
      this.ResumeLayout(false);
    } // InitializeComponent()

    /// <summary>
    /// Creates the form
    /// </summary>
    public SerialCpyForm() {
      Stream str;
      Icon   ico;

      this.m_ports = new ComboBox[2];
      this.m_baud = new ComboBox[2];
      this.m_databits = new ComboBox[2];
      this.m_parities = new ComboBox[2];
      this.m_stopbits = new ComboBox[2];
      this.m_ports[0] = new ComboBox();
      this.m_ports[1] = new ComboBox();
      this.m_baud[0] = new ComboBox();
      this.m_baud[1] = new ComboBox();
      this.m_databits[0] = new ComboBox();
      this.m_databits[1] = new ComboBox();
      this.m_parities[0] = new ComboBox();
      this.m_parities[1] = new ComboBox();
      this.m_stopbits[0] = new ComboBox();
      this.m_stopbits[1] = new ComboBox();
      this.m_status = new Label();
      this.InitializeComponent();
      foreach(string s in Serial.COMPORTS) {
        foreach(ComboBox c in this.m_ports) c.Items.Add(s);
      }
      foreach(int i in Serial.BAUDRATES) {
        foreach(ComboBox c in this.m_baud) c.Items.Add(i);
      }
      foreach(int i in Serial.DATABITS) {
        foreach(ComboBox c in this.m_databits) c.Items.Add(i);
      }
      foreach(ComboBox c in this.m_parities) {
        c.Items.Add(Parity.Even.ToString());
        c.Items.Add(Parity.Mark.ToString());
        c.Items.Add(Parity.None.ToString());
        c.Items.Add(Parity.Odd.ToString());
        c.Items.Add(Parity.Space.ToString());
      }
      foreach(ComboBox c in this.m_stopbits) {
        c.Items.Add(StopBits.None.ToString());
        c.Items.Add(StopBits.One.ToString());
        c.Items.Add(StopBits.OnePointFive.ToString());
        c.Items.Add(StopBits.Two.ToString());
      }
      this.Closing += onClose;
      try {
        this.readCfg();
        str = Assembly.GetExecutingAssembly().GetManifestResourceStream("Marta2.ico");
        ico = new Icon(str);
        str.Close();
        this.Icon = ico;
        this.m_serialcpy = new SerialCpy();
      }
      catch(Exception excp) {
        this.printExcp("ctor", excp);
      }
    } // SerialCpyForm()

  } // ckass SerialCpyForm
} // namespace serial

