using System;
using System.Collections;
using System.IO;
using System.IO.Ports;
using System.Drawing;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;

namespace serial {

  /// <summary>The main form of the application sg.exe</summary>
  /// <remarks>
  /// <IMG SRC="sg.png"/>
  /// <list type="table">
  /// <item>
  ///  <term>m_ports</term>
  ///  <description>
  ///   Combo box for selecting the GPS serial port. Contains all serial
  ///   ports available on the system, see also Serial.COMPORTS.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_baud</term>
  ///  <description>
  ///   Combo box for selecting a baud rate of the GPS serial port.
  ///   Usually 9600 or 4800 are a good choice.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_databits</term>
  ///  <description>
  ///   Combo box for selecting the number of databits for communicating
  ///   through the serial port. Usually 8 is good choice.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_parities</term>
  ///  <description>
  ///   Combo box for selecting the parity for the serial port.
  ///   No parity (None) is the mostly used default.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_stopbits</term>
  ///  <description>
  ///   Combo box for selecting the number of stopbits for the serial port.
  ///   1 stop bit (One) could be a reasonable default.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_type</term>
  ///  <description>
  ///   IP client type, may be either OFF, TCP, UDP or BCAST.
  ///   For OFF the serial port is used.
  ///   Refer to the documentation of the class IPClient for more details
  ///   on using IP rather than the serial port.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_server</term>
  ///  <description>
  ///   Name or IP of the server, if m_type is TCP or UDP. For m_type BCAST
  ///   the settings in this control are ignored.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_ipport</term>
  ///  <description>
  ///   The TCP or UDP port on which the server is listening for connections
  ///   for m_type TCP or UDP. The UDP port, on which this client is listening
  ///   for broadcasts for m_type BCAST.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_status</term>
  ///  <description>
  ///   Shows some informations about the type of the current connection,
  ///   serial or IP, if opened using menu item Port/Open.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_input</term>
  ///  <description>
  ///   Only if using serial port, no meaning for IP based connections.
  ///   String to send to the serial port, escape sequences according
  ///   to <c>System.Text.RegularExpressions.Regex.Escape()</c> can be used
  ///   for non alphanumeric characters. On desktop computers the strings
  ///   may be entered directly from the keyboard or read from a file using
  ///   menu item Config/Read file. On PDAs there is no oportunity to enter
  ///   the strings from keyboard, only the file option is possible.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_answer</term>
  ///  <description>
  ///   Data sent by the serial port or the IP server.
  ///   This control showns all line strings sent to the callback
  ///   Serial.RdLineCb, which in this class is implemented in the method
  ///   SerialPortCb.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Port/Open, m_menuopen</term>
  ///  <description>
  ///   Opens the serial port or a connection to an IP server specified in
  ///   the controls m_ports, m_type, m_server, m_ipport.
  ///   Opening the serial port or connection doesn't automatically cause
  ///   reading the stream from there. You have additionally to activate
  ///   reading using the menu item Port/Read.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Port/Close</term>
  ///  <description>
  ///   Closes connection to the serial port or IP server.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Port/Read, m_menurdport</term>
  ///  <description>
  ///   Switches reading from the serial port or IP server on or off.
  ///   Reading must be explicitely switched on after Port/Open, else the
  ///   application ignores any information coming from there.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Port/Send, m_menusend</term>
  ///  <description>
  ///   Sends string from m_input to the serial port.
  ///   Sending to IP servers is not possible.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Port/Filter, m_menufilter</term>
  ///  <description>
  ///   Switch filtering on or off. If switched on, then only NMEA 0183 GGA
  ///   sequences are recognized by the application, any other strings from
  ///   the serial port or IP are ignored.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Port/Exit</term>
  ///  <description>
  ///   Exits the application.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Config/Read file</term>
  ///  <description>
  ///   Reads lines from a text file into m_input.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Config/Log, m_menulog</term>
  ///  <description>
  ///   Switches logging on or off. May be used in order to trace application
  ///   errors. Each time the logging is switched on, a new log file named
  ///   sg.log is opened or created in the append mode.
  ///  </description>
  /// </item>
  /// </list>
  /// </remarks>
  /// <seealso cref="IPClient">IPClient</seealso>
  /// <seealso cref="Serial">Serial</seealso>
  /// <seealso cref="SerialGUI">SerialGUI</seealso>
  public class SerialForm : Form {

    /// <summary>Maximal number of inputs kept in the history</summary>
    /// <remarks>
    /// This is the maximal number of inputs entered by the user kept
    /// in the combo box m_input.
    /// </remarks>
    private static readonly int MAXHISTORY = 50;

    /// <summary>Filter string</summary>
    /// <remarks>The for allows only $GPGGA NMEA strings</remarks>
    private static readonly string FILTER = "$GPGGA";

    /// <summary>Name of the config file</summary>
    private static string CONFIGFILE {
      get {
        return Utils.getExtFileName(".cfg");
      }
    } // CONFIGFILE 

    /// <summary>Name of the log file</summary>
    private static string LOGFILE {
      get {
        return Utils.getExtFileName(".log");
      }
    } // LOGFILE

    /// <summary>
    /// Info about the current status of the connection
    /// </summary>
    private Label m_status;

    /// <summary>
    /// Control for the input string sent to the serial port
    /// </summary>
    private ComboBox m_input;

    /// <summary>
    /// Control for the answer string retrieved from the serial port.
    /// </summary>
    private TextBox m_answer;

    /// <summary>
    /// Names of COM ports available in the system.
    /// </summary>
    private ComboBox m_ports;

    /// <summary>
    /// Control for the baud rate of the port.
    /// </summary>
    private ComboBox m_baud;

    /// <summary>
    /// Control for the number of data bits.
    /// </summary>
    private ComboBox m_databits;

    /// <summary>
    /// Available COM port parities
    /// </summary>
    private ComboBox m_parities;

    /// <summary>
    /// Available COM port parities
    /// </summary>
    private ComboBox m_stopbits;

    /// <summary>Type of the input signal</summary>
    /// <remarks>
    /// OFF for the serial port defined in m_ports, m_baud,
    /// m_databits, m_parities and m_stopbits. TCP, UDB or BCAST
    /// for an IP based sorce like this one produced by s2ip.exe.
    /// </remarks>
    private ComboBox m_type;

    /// <summary>Name or IP of the server, if IP used</summary>
    private TextBox m_server;

    /// <summary>TCP or UDP port</summary>
    /// <remarks>
    /// Used only, if the m_type contains TCP, UDP or BCAST.
    /// </remarks>
    private TextBox m_ipport;

    /// <summary>Serial port</summary>
    private Serial m_serial;

    /// <summary>IP client used for IP communication</summary>
    private IPClient m_ipclient;

    /// <summary>Menu item port/open</summary>
    private MenuItem m_menuopen;

    /// <summary>Menu item config/log</summary>
    private MenuItem m_menulog;

    /// <summary>Menu item port/read</summary>
    private MenuItem m_menurdport;

    /// <summary>Menu item port/filter</summary>
    private MenuItem m_menufilter;

    /// <summary>Menu item port/send</summary>
    private MenuItem m_menusend;

    /// <summary>Stream used for logging</summary>
    private StreamWriter m_logstr;

    /// <summary>Delegate for showing serial port answers</summary>
    /// <remarks>
    /// The serial port thread or event procedure cannot write driectly
    /// into the forms controls, therefore we define a delegate, which is
    /// executed using Form.Invoke().
    /// See also
    /// http://www.ppc-welt.info/community/archive/index.php?t-118493.html
    /// </remarks>
    private EventHandler m_answercb;

    /// <summary>
    /// The answer string processed in the delegate method showAnswer().
    /// </summary>
    private string m_answerstr;

    /// <summary>The initial OpenFileDialog directory</summary>
    /// <remarks>
    /// Initially set to the directory of the exe. The estimated by the
    /// last directory of the file selected in menuRead_Click()
    /// </remarks>
    private string m_initofddir;

    /// <summary>
    /// Common method for showing exception errors occured in this form.
    /// </summary>
    /// <param name="aTitle">The title of the exception</param>
    /// <param name="anExcp">The excpetion itself</param>
    private void printExcp(string aTitle, Exception anExcp) {
      StringBuilder sb;

      sb = new StringBuilder(aTitle);
      sb.Append('\n');
      sb.Append(anExcp.ToString());
      MessageBox.Show(sb.ToString(), "Error");
      this.dolog(aTitle, anExcp);
    } // printExcp()

    /// <summary>Switches logging on or off</summary>
    private bool Log {
      set {
        if (this.m_logstr != null) {
          try { this.m_logstr.Close(); } catch(Exception) {}
          this.m_logstr = null;
        }
        if (value) {
          try { this.m_logstr = new StreamWriter(SerialForm.LOGFILE, true); }
          catch(Exception) {}
        }
      } // set
      get {
        if (this.m_logstr == null) return false;
        return true;
      } // get
    } // Log

    /// <summary>Log method</summary>
    /// <param name="descr">Log entry description</param>
    /// <param name="o">The log content</param>
    private void dolog(string descr, object o) {
      if (this.Log && o != null) {
        try {
          this.m_logstr.WriteLine("{0:yyyyMMdd HHmmss} {1}: {2}",
                                  DateTime.Now, descr, o.ToString());
          this.m_logstr.Flush();
        }
        catch(Exception) {}
      }
    } // dolog()

    /// <summary>Event method called if the form is closed</summary>
    /// <remarks>
    /// Writes current config to the file and closes serial port.
    /// </remarks>
    private void onClose(object sender, EventArgs e) {
      Exception ee;

      ee = null;
      try { this.m_serial.Close(); } catch(Exception excp) { ee = excp; }
      try { 
        this.m_ipclient.oper(ServerType.OFF, null, 0);
      }
      catch(Exception excp) {
        ee = excp;
      }
      if (this.m_logstr != null)
        try { this.m_logstr.Close(); } catch(Exception excp) { ee = excp; }
      try { this.writeCfg(); } catch(Exception excp) { ee = excp; }
      if (ee != null) this.printExcp("exit", ee);
    } // onClose()

    /// <summary>
    /// Event handler for the menu item close.
    /// </summary>
    /// <remarks>
    /// Closes the serial port.
    /// </remarks>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuClose_Click(object sender, EventArgs e) {
      try {
        this.m_serial.Close();
        this.m_ipclient.oper(ServerType.OFF, null, 0);
        this.m_status.Text = string.Empty;
        this.m_menuopen.Checked = false;
        this.m_menurdport.Checked = this.m_serial.Read;
      }
      catch(Exception excp) {
        this.printExcp("close", excp);
      }
    }

    /// <summary>Switches logging on or off</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuLog_Click(object sender, EventArgs e) {
      try {
        if (this.Log) {
          this.Log = false;
          this.m_menulog.Checked = false;
        }
        else {
          this.Log = true;
          this.m_menulog.Checked = true;
        }
      }
      catch(Exception excp) {
        this.printExcp("log", excp);
      }
    } // menuLog_Click()

    /// <summary>
    /// Event handler for the menu item exit.
    /// </summary>
    /// <remarks>
    /// Closes the application.
    /// </remarks>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuExit_Click(object sender, EventArgs e) {
      try {
        this.m_serial.Close();
        this.m_ipclient.oper(ServerType.OFF, null, 0);
      }
      catch(Exception excp) {
        this.printExcp("exit", excp);
      }
      this.Close();
    } // menuExit_Click()

    /// <summary>
    /// Event handler for the menu item open
    /// </summary>
    /// <remarks>
    /// Opens the com port. Must be called before send.
    /// </remarks>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuOpen_Click(object sender, EventArgs e) {
      ServerType typ;
      int        port;

      try {
        this.m_menuopen.Checked = false;
        this.m_serial.Close();
        this.m_ipclient.oper(ServerType.OFF, null, 0);
        typ = (ServerType) Enum.Parse(typeof(ServerType),
                                      this.m_type.Text,
                                      false);
        if (typ == ServerType.OFF) {
          this.m_serial.PortName = this.m_ports.Text;
          this.m_serial.BaudRate = int.Parse(this.m_baud.Text);
          this.m_serial.DataBits = int.Parse(this.m_databits.Text);
          this.m_serial.Parity = (Parity) Enum.Parse(typeof(Parity),
                                                     this.m_parities.Text,
                                                     false);
          this.m_serial.StopBits = (StopBits) Enum.Parse(typeof(StopBits),
                                                         this.m_stopbits.Text,
                                                         false);
          this.m_serial.Open();
          this.m_status.Text = this.m_serial.ToString();
          this.m_menurdport.Checked = this.m_serial.Read;
        }
        else {
          port = int.Parse(this.m_ipport.Text);
          this.m_ipclient.oper(typ, this.m_server.Text, port);
          this.m_status.Text = this.m_ipclient.ToString();
        }
        this.m_menuopen.Checked = true;
      }
      catch(Exception excp) {
        this.printExcp("open", excp);
      }
    } // menuOpenClick()

    /// <summary>
    /// Event handler for the menu item send
    /// </summary>
    /// <remarks>
    /// Sends the string from m_input to the serial port.
    /// All escape strings like \r, \n \xNN etc. are unescaped. Thus the 
    /// user may enter almost every character. Espacially usafull for
    /// constructing binary SIRF III sequences.
    /// Don't forget to append \r\n at the end of the string while interacting
    /// with modems, SIRF chipsets etc.
    /// </remarks>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuSend_Click(object sender, EventArgs e) {
      StringBuilder sb;
      string        sunesc, s;
      byte[]        b;
      bool          found;

      try {
        s = this.m_input.Text;
        if (s != string.Empty) {
          found = false;
          foreach(string it in this.m_input.Items) {
            if (s == it) {
              found = true;
              break;
            }
          } // foreach
          if (!found) {
            if (this.m_input.Items.Count > SerialForm.MAXHISTORY) {
              this.m_input.Items.RemoveAt(0);
            }
            this.m_input.Items.Add(s);
          }
          if (this.m_serial.IsOpen) {
            sunesc = Regex.Unescape(s);
            b = Encoding.ASCII.GetBytes(sunesc);
            this.m_serial.WriteBytes(b, b.Length);
            if (this.Log) {
              sb = new StringBuilder(s);
              foreach(byte bt in b) sb.Append(' ').Append(bt);
              this.dolog("sent", sb);
            }
          }
          else {
            MessageBox.Show("Open port first", "Error");
          }
        } // if
      }
      catch(Exception excp) {
        this.printExcp("send", excp);
      }
    } // menuSend_Click()

    /// <summary>
    /// Reads a string from a file and puts it into the input combo.
    /// </summary>
    /// <remarks>
    /// Port reading is stoped for the time of the execution of this
    /// method. The small PDA could be inable to handle the numerous
    /// serial port events and this open file dialog simultaneously.
    /// A we could still have locking problems ...
    /// </remarks>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuRead_Click(object sender, EventArgs e) {
      OpenFileDialog ofd;
      StreamReader   stream;
      string         fn, s;
      int            i;
      bool           oldrdport;

      stream = null;
      try {
        oldrdport = this.m_serial.Read;
        this.m_serial.Read = false;
        ofd = new OpenFileDialog();
        if (!File.Exists(this.m_initofddir)) this.m_initofddir = Utils.Exedir;
        ofd.InitialDirectory = this.m_initofddir;
        ofd.Filter = "text|*.txt";
        ofd.FilterIndex = 1;
        ofd.ShowDialog();
        fn = ofd.FileName;
        if (fn != null && fn != string.Empty) {
          stream = new StreamReader(fn);
          while((s = stream.ReadLine()) != null) {
            s = s.Trim();
            if (s != string.Empty) {
              i = this.m_input.Items.Count;
              if (i > SerialForm.MAXHISTORY) {
                this.m_input.Items.RemoveAt(0);
              }
              this.m_input.Items.Add(s);
            }
          }
          this.m_initofddir = Path.GetDirectoryName(fn);
        }
        this.m_serial.Read = oldrdport;
      }
      catch(Exception excp) {
        this.printExcp("read", excp);
      }
      if (stream != null) stream.Close();
    } // menuRead_Click()

    /// <summary>Starts or stops reading from serial port</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuRdPort_Click(object sender, EventArgs e) {
      try {
        if (this.m_serial.Read)
          this.m_serial.Read = false;
        else
          this.m_serial.Read = true;
        this.m_menurdport.Checked = this.m_serial.Read;
      }
      catch(Exception excp) {
        this.printExcp("read port", excp);
      }
    } // menuRdPort_Click()

    /// <summary>Switch filter on or off</summary>
    /// <remarks>
    /// If switched on, then only lines starting with SerialForm.FILTER
    /// are shown in m_answer. If switched off, then all lines are
    /// passed to m_answer.
    /// </remarks>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuFilter_Click(object sender, EventArgs e) {
      try {
        if (this.m_serial.Filter == null || this.m_ipclient.Filter == null) {
          this.m_serial.Filter = SerialForm.FILTER;
          this.m_ipclient.Filter = SerialForm.FILTER;
          this.m_menufilter.Checked = true;
        }
        else {
          this.m_serial.Filter = null;
          this.m_ipclient.Filter = null;
          this.m_menufilter.Checked = false;
        }
       
      }
      catch(Exception excp) {
        this.printExcp("filter", excp);
      }
    } // menuFilter_Click()

    /// <summary>Called if the type in m_type changes</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void typeChanged(object sender, EventArgs e) {
      this.deactivateControls();
    } // typeChanged()

    /// <summary>Reads configuration into combos</summary>
    /// <remarks>
    /// If the configuration can'b read for some reason, then defaults
    /// are assumed.
    /// </remarks>
    private void readCfg() {
      StreamReader stream;
      string       comport, baudrate, databits, parity, stopbits, type, ipport,
                   server, s, filter, initofddir;
      int          i;

      stream = null;
      comport = null;
      baudrate = null;
      databits = null;
      parity = null;
      stopbits = null;
      type = null;
      server = null;
      ipport = null;
      filter = null;
      initofddir = null;
      this.m_input.Items.Clear();
      try {
        stream = new StreamReader(SerialForm.CONFIGFILE);
        comport = stream.ReadLine();
        baudrate = stream.ReadLine();
        databits = stream.ReadLine();
        parity = stream.ReadLine();
        stopbits = stream.ReadLine();
        type = stream.ReadLine();
        server = stream.ReadLine();
        ipport = stream.ReadLine();
        filter = stream.ReadLine();
        initofddir = stream.ReadLine();
        i = 0;
        while((s = stream.ReadLine()) != null) {
          if (i > SerialForm.MAXHISTORY) break;
          i++;
          this.m_input.Items.Add(s);
        }
      }
      catch(Exception excp) {
        this.dolog("read config", excp);
      }
      if (stream != null) stream.Close();
      if (comport == null) comport = Serial.DEFAULT_COMPORT;
      if (baudrate == null) baudrate = Serial.DEFAULT_BAUDRATE.ToString();
      if (databits == null) databits = Serial.DEFAULT_DATABITS.ToString();
      if (parity == null) parity = Serial.DEFAULT_PARITY.ToString();
      if (stopbits == null) stopbits = Serial.DEFAULT_STOPBITS.ToString();
      if (type == null) type = ServerType.OFF.ToString();
      if (server == null) server = "169.254.2.1";
      if (ipport == null) ipport = "5000";
      if (filter == null || filter != "true") {
        this.m_serial.Filter = null;
        this.m_menufilter.Checked = false;
      }
      else {
        this.m_serial.Filter = SerialForm.FILTER;
        this.m_menufilter.Checked = true;
      }
      if (initofddir == null) initofddir = Utils.Exedir;
      this.m_ports.Text = comport;
      this.m_baud.Text = baudrate;
      this.m_databits.Text = databits;
      this.m_parities.Text = parity;
      this.m_stopbits.Text = stopbits;
      this.m_type.Text = type;
      this.m_server.Text = server;
      this.m_ipport.Text = ipport;
      this.m_initofddir = initofddir;
    } // readCfg()

    /// <summary>Writes setting to the configuration file</summary>
    private void writeCfg() {
      StreamWriter stream;

      stream = null;
      try {
        stream = new StreamWriter(SerialForm.CONFIGFILE);
        stream.WriteLine(this.m_ports.Text);
        stream.WriteLine(this.m_baud.Text);
        stream.WriteLine(this.m_databits.Text);
        stream.WriteLine(this.m_parities.Text);
        stream.WriteLine(this.m_stopbits.Text);
        stream.WriteLine(this.m_type.Text);
        stream.WriteLine(this.m_server.Text);
        stream.WriteLine(this.m_ipport.Text);
        if (this.m_serial.Filter == null)
          stream.WriteLine("false");
        else
          stream.WriteLine("true");
        foreach(string s in this.m_input.Items) {
          stream.WriteLine(s);
        }
        stream.WriteLine(this.m_initofddir);
      }
      catch(Exception excp) {
        this.dolog("write config", excp);
      }
      if (stream != null) stream.Close();
    } // writeCfg()

    /// <summary>Writes string encapsulated into m_answer</summary>
    /// <remarks>
    /// Used because serialPortCb as an thread delegate (similar to event)
    /// can't write directly into controls. See also
    /// http://www.ppc-welt.info/community/archive/index.php?t-118493.html.
    /// The string to show in the control m_answer is expected in 
    /// m_answerstr.
    /// </remarks>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void showAnswer(object sender, EventArgs e) {
      string        s;

      try {
        s = this.m_answerstr;
        if (!string.IsNullOrEmpty(s)) {
          this.m_answer.Text = s;
        }
      }
      catch(Exception excp) {
        this.dolog("showAnswer", excp);
      }
    } // addAnwer()

    /// <summary>Callbck for strings read from the serial port</summary>
    /// <param name="arr">Array of line strings received from port</param>
    private void serialPortCb(ArrayList arr) {
      StringBuilder sb;
      string        sh;

      try {
        sb = new StringBuilder();
        foreach(object o in arr) {
          sh = o.ToString().Replace('\r', ' ')
                           .Replace('\n', ' ')
                           .Replace('\t', ' ').Trim();
          this.dolog("cb", sh);
          sb.Append(sh).Append('\r').Append('\n');
        }
        this.m_answerstr = sb.ToString();
        this.Invoke(this.m_answercb);
      }
      catch(Exception excp) {
        this.dolog("cb", excp);
      }
    } // serialPortCb()

    /// <summary>Activates or deactivates controls depending on type</summary>
    /// <remarks>
    /// A lot of controls must not be used in the IP mode. In the serial
    /// port mode there is no sense for example to use the IP port control.
    /// This method marks controls as grayed, for which there is no sense to
    /// use in the certain mode.
    /// </remarks>
    private void deactivateControls() {
      ServerType typ;

      try {
        typ = (ServerType) Enum.Parse(typeof(ServerType),
                                      this.m_type.Text,
                                      false);
        if (typ == ServerType.OFF) {
          this.m_server.Enabled = false;
          this.m_ipport.Enabled = false;
          this.m_ports.Enabled = true;
          this.m_baud.Enabled = true;
          this.m_databits.Enabled = true;
          this.m_parities.Enabled = true;
          this.m_stopbits.Enabled = true;
          this.m_input.Enabled = true;
          this.m_menurdport.Enabled = true;
          this.m_menusend.Enabled = true;
        }
        else {
          this.m_server.Enabled = true;
          this.m_ipport.Enabled = true;
          this.m_ports.Enabled = false;
          this.m_baud.Enabled = false;
          this.m_databits.Enabled = false;
          this.m_parities.Enabled = false;
          this.m_stopbits.Enabled = false;
          this.m_input.Enabled = false;
          this.m_menurdport.Enabled = false;
          this.m_menusend.Enabled = false;
        }
      }
      catch(Exception excp) {
        this.dolog("deactivateControls", excp);
      }
    } // deactivateControls()

    /// <summary>Layout of the form</summary>
    private void InitializeComponent() {
      MainMenu   menu;
      MenuItem[] mit_port;
      MenuItem[] mit_cfg;
      MenuItem   menuPort, menuCfg;

      this.SuspendLayout();
      mit_port = new MenuItem[6];
      mit_port[0] = this.m_menuopen;
      mit_port[1] = new MenuItem();
      mit_port[1].Text = "Close";
      mit_port[1].Click += menuClose_Click;
      mit_port[2] = this.m_menurdport;
      mit_port[3] = this.m_menusend;
      mit_port[4] = this.m_menufilter;
      mit_port[5] = new MenuItem();
      mit_port[5].Text = "Exit";
      mit_port[5].Click += menuExit_Click;
      mit_cfg = new MenuItem[2];
      mit_cfg[0] = new MenuItem();
      mit_cfg[0].Text = "Read file";
      mit_cfg[0].Click += menuRead_Click;
      mit_cfg[1] = this.m_menulog;
      menuPort = new MenuItem();
      menuPort.Text = "Port";
      menuPort.MenuItems.Add(mit_port[0]);
      menuPort.MenuItems.Add(mit_port[1]);
      menuPort.MenuItems.Add(mit_port[2]);
      menuPort.MenuItems.Add(mit_port[3]);
      menuPort.MenuItems.Add(mit_port[4]);
      menuPort.MenuItems.Add(mit_port[5]);
      menuCfg = new MenuItem();
      menuCfg.Text = "Config";
      menuCfg.MenuItems.Add(mit_cfg[0]);
      menuCfg.MenuItems.Add(mit_cfg[1]);
      menu = new MainMenu();
      menu.MenuItems.Add(menuPort);
      menu.MenuItems.Add(menuCfg);
      this.ClientSize = new Size(270, 180);
      this.m_status.Location = new Point(5, 2);
      this.m_status.Size = new Size(260, 20);
      this.m_ports.Location = new Point(5, 25);
      this.m_ports.Size = new Size(60, 30);
      this.m_baud.Location = new Point(65, 25);
      this.m_baud.Size = new Size(50, 30);
      this.m_databits.Location = new Point(115, 25);
      this.m_databits.Size = new Size(40, 30);
      this.m_parities.Location = new Point(155, 25);
      this.m_parities.Size = new Size(50, 30);
      this.m_stopbits.Location = new Point(205, 25);
      this.m_stopbits.Size = new Size(50, 30);
      this.m_type.Location = new Point(5, 55);
      this.m_type.Size = new Size(60, 30);
      this.m_server.Location = new Point(65, 55);
      this.m_server.Size = new Size(100, 30);
      this.m_ipport.Location = new Point(170, 55);
      this.m_ipport.Size = new Size(50, 30);
      this.m_input.Location = new Point(5,  85);
      this.m_input.Size = new Size(260, 20);
      this.m_answer.Location = new Point(5, 110);
      this.m_answer.Size = new Size(260, 60);
      this.AutoScroll = true;
      this.Controls.Clear();
      this.Controls.Add(this.m_status);
      this.Controls.Add(this.m_ports);
      this.Controls.Add(this.m_baud);
      this.Controls.Add(this.m_databits);
      this.Controls.Add(this.m_parities);
      this.Controls.Add(this.m_stopbits);
      this.Controls.Add(this.m_type);
      this.Controls.Add(this.m_server);
      this.Controls.Add(this.m_ipport);
      this.Controls.Add(this.m_input);
      this.Controls.Add(this.m_answer);
      this.Name = "SerialForm";
      this.Text = "SerialGUI";
      this.Menu = menu;
      this.ResumeLayout(false);
      this.Closing += onClose;
    } // InitializeComponent()

    /// <summary>
    /// Creates the form
    /// </summary>
    public SerialForm() {
      ReadLineDelegate rd;
      Stream           str;
      Icon             ico;

      this.m_logstr = null;
      this.m_status = new Label();
      this.m_ports = new ComboBox();
      this.m_baud = new ComboBox();
      this.m_databits = new ComboBox();
      this.m_parities = new ComboBox();
      this.m_stopbits = new ComboBox();
      this.m_type = new ComboBox();
      this.m_server = new TextBox();
      this.m_ipport = new TextBox();
      this.m_input = new ComboBox();
      this.m_answer = new TextBox();
      this.m_answer.Multiline = true;
      this.m_answer.ReadOnly = true;
      this.m_answer.WordWrap = false;
      this.m_answer.ScrollBars = ScrollBars.Vertical | ScrollBars.Horizontal;
      this.m_serial = new Serial();
      this.m_ipclient = new IPClient();
      this.m_serial.Filter = null;
      rd = new ReadLineDelegate(this.serialPortCb);
      this.m_answercb = new EventHandler(this.showAnswer);
      this.m_serial.RdLineCb = rd;
      this.m_ipclient.RdLineCb = rd;
      this.m_menuopen = new MenuItem();
      this.m_menulog = new MenuItem();
      this.m_menufilter = new MenuItem();
      this.m_menurdport = new MenuItem();
      this.m_menusend = new MenuItem();
      this.m_menuopen.Text = "Open";
      this.m_menuopen.Click += this.menuOpen_Click;
      this.m_menulog.Text = "Log";
      this.m_menulog.Click += this.menuLog_Click;
      this.m_menurdport.Text = "Read";
      this.m_menurdport.Click += this.menuRdPort_Click;
      this.m_menufilter.Text = "Filter";
      this.m_menufilter.Click += this.menuFilter_Click;
      this.m_menusend.Text = "Send";
      this.m_menusend.Click += this.menuSend_Click;
      this.m_type.TextChanged += this.typeChanged;
      this.InitializeComponent();
      // fill combos with all available values
      foreach(string s in Serial.COMPORTS) this.m_ports.Items.Add(s);
      foreach(int i in Serial.BAUDRATES) this.m_baud.Items.Add(i);
      foreach(int i in Serial.DATABITS) this.m_databits.Items.Add(i);
      this.m_parities.Items.Add(Parity.Even.ToString());
      this.m_parities.Items.Add(Parity.Mark.ToString());
      this.m_parities.Items.Add(Parity.None.ToString());
      this.m_parities.Items.Add(Parity.Odd.ToString());
      this.m_parities.Items.Add(Parity.Space.ToString());
      this.m_stopbits.Items.Add(StopBits.None.ToString());
      this.m_stopbits.Items.Add(StopBits.One.ToString());
      this.m_stopbits.Items.Add(StopBits.OnePointFive.ToString());
      this.m_stopbits.Items.Add(StopBits.Two.ToString());
      this.m_type.Items.Add(ServerType.OFF.ToString());
      this.m_type.Items.Add(ServerType.TCP.ToString());
      this.m_type.Items.Add(ServerType.UDP.ToString());
      this.m_type.Items.Add(ServerType.BCAST.ToString());
      this.readCfg();
      this.deactivateControls();
      try {
        str = Assembly.GetExecutingAssembly().GetManifestResourceStream("Marta1.ico");
        ico = new Icon(str);
        str.Close();
        this.Icon = ico;
      }
      catch(Exception excp) {
        this.printExcp("ctor", excp);
      }
    } // SerialForm()

  } // class SerialForm
} // namespace serial

