using System;
using System.Collections;
using System.IO.Ports;

namespace serial {

  /// <summary>Reads coordinates from a GPS serial port</summary>
  /// <remarks>
  /// Reads serial port using the method parseNMEA(), which is passed to
  /// the porperty RdLineCb to the base class. Each line successfully read
  /// from the serial port is parsed using Utils.fromGGA() and for the
  /// coordinates, if any, the delegate set by CoordDelegate is called.
  /// Only NMEA GGA sequences currently supported.
  /// </remarks>
  /// <example>
  /// The following example reads GPS coordinates from COM4, 9600, 8N1,
  /// WGS84 is used for converting geographical coordinates into UTM.
  /// The coordinates are the printed to standard output.
  /// <code lang="C#">
  /// using System;
  /// using serial;
  /// 
  /// public class SerialGPSTest {
  ///   private static void coordinates(double tim,
  ///                                   double latitude,
  ///                                   double longitude,
  ///                                   double utmnorth,
  ///                                   double utmeast,
  ///                                   double altitude) {
  ///     Console.WriteLine("{0} {1} {2} {3} {4} {5}",
  ///                       tim, latitude, longitude,
  ///                       utmnorth, utmeast, altitude);
  ///   }
  /// 
  ///   public static void Main() {
  ///     SerialGPS s;
  /// 
  ///     s = new SerialGPS();
  ///     s.CoordCb = new CoordDelegate(SerialGPSTest.coordinates);
  ///     s.Open("COM4", 9600, 8, Parity.None, StopBits.One, Datum.WGS84);
  ///     Console.WriteLine("Hit ENTER to close"); Console.ReadLine();
  ///     s.Close();
  ///   }
  /// }
  /// </code>
  /// The code can be compiled using:
  /// <code>
  /// csc -out:test.exe SerialGPSTest.cs SeriaGPS.cs Serial.cs Utils.cs
  /// gmcs -out:test.exe SerialGPSTest.cs SeriaGPS.cs Serial.cs Utils.cs
  /// </code>
  /// </example>
  /// <seealso cref="Utils.fromGGA">Utils.fromGGA()</seealso>
  /// <seealso cref="Utils.toUTM">Utils.toUTM()</seealso>
  public class SerialGPS : Serial {

    /// <summary>Default system for recomputing into UTM</summary>
    public const Datum DEFAULT_SYSTEM = Datum.WGS84;

    /// <summary>Last coordinates retrieved in parseNMEA</summary>
    private double[] m_lastcoord;

    /// <summary>System used for recomputing into UTM</summary>
    private Datum m_system;

    /// <summary>Delegate for handling strings coming from GPS</summary>
    /// <remarks>Always set to parseNMEA</remarks>
    private ReadLineDelegate m_gpscb;

    /// <summary>
    /// Callback method for handling coordinates retreived from GPS
    /// </summary>
    private CoordDelegate m_coordcb;

    /// <summary>Initializes the class</summary>
    public SerialGPS() {
      int i;

      this.m_coordcb = null;
      this.m_system = SerialGPS.DEFAULT_SYSTEM;
      this.m_lastcoord = new double[5];
      for(i = 0; i < this.m_lastcoord.Length; i++) this.m_lastcoord[i] = 0.0;
      this.m_gpscb = new ReadLineDelegate(this.parseNMEA);
    } // SerialGPS()

    /// <summary>System to use for recomputing into UTM</summary>
    public Datum System {
      get { return this.m_system; }
      set { this.m_system = value; }
    } // System

    /// <summary>Sets callback method for coordinates from GPS</summary>
    public CoordDelegate CoordCb {
      set {
        this.m_coordcb = value;
      }
    } // CoordCb

    /// <summary>Callback method for handling incoming NMEA strings</summary>
    /// <remarks>
    /// If the coordinate callback CoordCb has not been set, then the method
    /// does absolutely nothing. Also it does nothing if the incoming gps
    /// string is not a valid NMEA GGA string according to Utils.fromGGA().
    /// Else it parses the string using Utils.fromGGA() and calls the callback
    /// method set in CoordCb, if the coordinates (but not the UTC time)
    /// changed since last call (m_lastcoord).
    /// The method calls Utils.dontSuspend() in order to prevent suspending
    /// windows ce/mobile based device as long as valid data arrives.
    /// </remarks>
    /// <param name="arr">Array of NMEA line strings</param>
    private void parseNMEA(ArrayList arr) {
      double tim1, lat1, lon1, utmn1, utme1, alt1;
      double tim2, lat2, lon2, utmn2, utme2, alt2;
      bool   isValid;

      if (this.m_coordcb != null) {
        tim2 = 0.0; lat2 = 0.0; lon2 = 0.0; utmn2 = 0.0; utme2 = 0.0; alt2 = 0.0;
        isValid = false;
        foreach(object o in arr) {
          if (Utils.fromGGA(out tim1,
                            out lat1,
                            out lon1,
                            out utmn1,
                            out utme1,
                            out alt1,
                            o.ToString(),
                            this.m_system)) {
             tim2 = tim1;
             lat2 = lat1;
             lon2 = lon1;
             utmn2 = utmn1;
             utme2 = utme1;
             alt2 = alt1;
             isValid = true;
           }
        } // foreach
        if (isValid &&
            (Math.Abs(lat2 - this.m_lastcoord[0]) > 0.0002 ||
             Math.Abs(lon2 - this.m_lastcoord[1]) > 0.0002 ||
             Math.Abs(utmn2 - this.m_lastcoord[2]) > 0.01 ||
             Math.Abs(utme2 - this.m_lastcoord[3]) > 0.01 ||
             Math.Abs(alt2 - this.m_lastcoord[4]) > 0.5)) {
          this.m_lastcoord[0] = lat2;
          this.m_lastcoord[1] = lon2;
          this.m_lastcoord[2] = utmn2;
          this.m_lastcoord[3] = utme2;
          this.m_lastcoord[4] = alt2;
          this.m_coordcb(tim2, lat2, lon2, utmn2, utme2, alt2);
        }
        // prevent suspending device as long as we get any data
        Utils.dontSuspend();
      } // if (this.m_coordcb != null)
    } // parseNMEA()

    /// <summary>Opens serial port with parameters already set</summary>
    /// <remarks>
    /// Additionaly to the base method this method some defaults
    /// necessary for proper GPS functionality.
    /// </remarks>
    public override void Open() {
      this.Close();
      this.RdLineCb = this.m_gpscb;
      this.RdByteCb = null;
      this.Filter = null;
      base.Open();
      this.Read = true;
    } // Open()

    /// <summary>Opens serial port with the given parameters</summary>
    /// <param name="portname">serial port name</param>
    /// <param name="baudrate">serial port baudrate</param>
    /// <param name="databits">number of data bits</param>
    /// <param name="parity">serial port parity</param>
    /// <param name="stopbits">number of stop bits</param>
    /// <param name="initialSystem">system to use for UTM</param>
    public void Open(string   portname,
                     int      baudrate,
                     int      databits,
                     Parity   parity,
                     StopBits stopbits,
                     Datum    initialSystem) {
      this.Close();
      this.PortName = portname;
      this.BaudRate = baudrate;
      this.DataBits = databits;
      this.Parity = parity;
      this.StopBits = stopbits;
      this.System = initialSystem;
      this.Open();
    } // Open()
  } // class SerialGPS

} // namespace serial

