using System;
using System.IO;
using System.IO.Ports;
using System.Drawing;
using System.Reflection;
using System.Text;
using System.Windows.Forms;

namespace serial {

  /// <summary>
  /// The main form of the program sgps.exe
  /// </summary>
  /// <remarks>
  /// <IMG SRC="sgps.png"/>
  /// <list type="table">
  /// <item>
  ///  <term>m_ports</term>
  ///  <description>
  ///   Combo box for selecting the GPS serial port. Contains all serial
  ///   ports available on the system, see also Serial.COMPORTS.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_baud</term>
  ///  <description>
  ///   Combo box for selecting a baud rate of the GPS serial port.
  ///   Usually 9600 or 4800 are a good choice.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_databits</term>
  ///  <description>
  ///   Combo box for selecting the number of databits for communicating
  ///   through the serial port. Usually 8 is good choice.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_parities</term>
  ///  <description>
  ///   Combo box for selecting the parity for the serial port.
  ///   No parity (None) is the mostly used default.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_stopbits</term>
  ///  <description>
  ///   Combo box for selecting the number of stopbits for the serial port.
  ///   1 stop bit (One) could be a reasonable default.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_system</term>
  ///  <description>
  ///   The type of the ellispoid used for convering into UTM coordinates.
  ///   The mostly used standard is WGS84. See Utils.toUTM() for more
  ///   details.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_type</term>
  ///  <description>
  ///   The kind in which the application gets NMEA's GGA sequences with
  ///   GPS coordinates. It may be OFF for getting them from the serial port
  ///   selected in m_ports or one of TCP, UDP, BCAST for getting them through
  ///   IP. In this case the IP parameters are defined in the text boxes
  ///   m_server and m_ipport. NMEA sequences coming on the serial port can
  ///   be streamed using for example the application s2ip.exe, thus you can
  ///   use the PDA only as the GPS server and this application sgps.exe can
  ///   run on a more powerfull desktop computer. See also the documentation
  ///   of the classes IPClient and IPClientGPS for more details on using
  ///   IP communication.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_server</term>
  ///  <description>
  ///   The name or IP of the IP server streaming NMEA sequences. Used, if
  ///   either TCP or UDP has been selected in m_type. No meaning, if
  ///   broadcast streaming (BCAST) has been selected in m_type.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_ipport</term>
  ///  <description>
  ///   Port on which the TCP or UDP server sends the NMEA sequences, if
  ///   TCP or UDP selected in m_type. Port on which to listen for NMEA
  ///   sequences broadcast, if BCAST selected in m_type.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_lblutctim</term>
  ///  <description>
  ///   The UTC time from the last valid NMEA GGA sequence received from GPS.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_lbllat</term>
  ///  <description>
  ///   The latitude in decimal degrees from the last valid NMEA GGA sequence
  ///   received from GPS.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_lbllon</term>
  ///  <description>
  ///   The longitude in decimal degrees from the last valid NMEA GGA sequence
  ///   received from GPS.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_lblnutm</term>
  ///  <description>
  ///   The UTM northing coordinate computed from the latitude and longitude
  ///   for the ellipsoid type selected in m_system.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_lbleutm</term>
  ///  <description>
  ///   The UTM easting coordinate computed from the latitude and longitude
  ///   for the elliposid type selected in m_system.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>m_lblalt</term>
  ///  <description>
  ///   The altitude (elevation) in meters from the last valid NMEA GGA
  ///   sequence retrieved from GPS.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Port/Open, m_menuopen</term>
  ///  <description>
  ///   Opens the GPS serial port or the IP communication type
  ///   depending on the present selection in m_system and starts reading
  ///   NMEA 0183 GGA sequences. Assuming good outdoor conditions it takes
  ///   usually several minutes until the first valid GGA sequence arrives and
  ///   the coordinates are shown in m_lblutctim, m_lbllat, m_lblon, m_lblnutm,
  ///   m_lbleutm and m_lblalt. The current settings in the combo boxes
  ///   m_ports, m_baud, m_databits, m_parities, m_stopbits, m_system, m_type,
  ///   m_server and m_ipport are used for opening, changes in the selections
  ///   in these combo boxes are ignored until the menu item Port/Open is 
  ///   activated again. Port/Open always closes the presently opened NMEA
  ///   stream, if needed.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Port/Close</term>
  ///  <description>
  ///   Closes the presently opened NMEA stream: serial port or IP client.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Port/Log, m_menulog</term>
  ///  <description>
  ///   Switches writing to a gpx file on or off. gpx files may be used in
  ///   order to store a path for later anylysis using for example Google
  ///   Earth. This application also writes evtl. exceptions and error
  ///   messages as comments to the gpx files so that the file may be
  ///   additionally used for investigating the application errors.
  ///   The name of the gpx file is estimated automatically by the application,
  ///   each time the writing to the gpx file is beeing activated.
  ///   The file name is created as the full path of the executable of
  ///   this applications completed by the current timestamp and the file
  ///   extension gpx. See the documentation on the class Gpsx for more
  ///   details on the gpx format.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Port/Map, m_menumap</term>
  ///  <description>
  ///   Activates or deactivates application showing a map for the
  ///   current position. See the documentation on the class
  ///   BndBmpForm for more details on how to create maps. The map
  ///   definitions should be located in the file having the same path as
  ///   the current executabe with the extension map. For example if
  ///   the current executable is located in <c>\temp\sgps.exe</c>, then the
  ///   map definition file should be located in <c>\temp\sgps.map</c>.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Port/Display off</term>
  ///  <description>
  ///   Switches the battery-hungry display of the PDA off. No meaning
  ///   on desktop computers.
  ///   Running on PDA this application periodically resets the inactivity
  ///   timer prohibiting the PDA from going into suspend mode and prohibiting
  ///   the power management from switching the display off. In order to save
  ///   the battery life time the display may be switched off manually using
  ///   this menu item. The display may be switched on again by pressing any
  ///   of the PDA's buttons. See also the documentation of the method
  ///   Usils.displayOff() for more details on the implementation of this
  ///   function.
  ///  </description>
  /// </item>
  /// <item>
  ///  <term>Port/Exit</term>
  ///  <description>
  ///   Closes ony operations and exits the application.
  ///  </description>
  /// </item>
  /// </list>
  /// </remarks>
  /// <example>
  /// <list type="bullet">
  /// <item>
  /// If you want to use the GPS connected to the serial port of your
  /// device, the first obtain a valid serial port. There is no golden
  /// rule to find the proper GPS serial port on the PDA. It may be
  /// the software device served by the GPS intermediate driver, you
  /// may take a look at the registry entries Drivers/Active or just
  /// try all of the approx. 10 serial ports available on PDA. Enter
  /// the parameters of the serial port to the controls m_ports,
  /// m_baud, m_databits, m_parities, m_stopbits. And select OFF in
  /// the control m_type.
  /// </item>
  /// <item>
  /// If the NMEA stream is coming through IP, for example served by
  /// the program s2ip.exe, then select the IP protocol type (TCP, UDP
  /// or BCAST) in the control m_type, anter the address of the
  /// server in m_server (doesn't matter for BCAST) and the TCP/UDP port
  /// of the server or the port on which the client is listening on for the
  /// type BCAST in m_ipport.
  /// </item>
  /// <item>
  /// Start reading GPS by selecting menu Port/Open. It may take several
  /// minutes until the GPS chip finds satelites and the first coordinates
  /// in the coordinate labels (m_lblutctim, m_lbllat, m_lblon, m_lblnutm,
  /// m_lbleutm, m_lblalt). The GPS chips usually need a good contact to
  /// the satelite, they may refuse to work indoor or on bad weather
  /// conditions. It his case you will never see any coordinates in the
  /// labels.
  /// </item>
  /// <item>
  /// With the menu Port/Log creating of GPX files may be switched on or off.
  /// Each time switched on, the gpx with the name
  /// <c>sgps&lt;yyyyMMddHHmmss&gt;.gpx</c> is created, where yyyMMddHHmmss is
  /// the current timestamp.
  /// </item>
  /// <item>
  /// Refer to the documentation on BndBmpForm for how to define maps.
  /// The map definitions according to BndBmpForm should be prepared in the
  /// file <c>sgps.map</c> located in the same directory like the executable
  /// <c>sgps.exe</c>. With the menu Port/Map the application showing maps
  /// map be switched on or off. A map is shown firs then, when the current
  /// coordinate is withing the bounds for the defined maps. On slow PDAs it
  /// may take some time, until the map bitmap will be entirelly shown on
  /// the screen. Currently the map application doesn't scroll automatically
  /// to the point on the map corresponding to the current position, so that
  /// you'll have to scroll in there by yourself. For example the following
  /// file <c>sgps.map</c>:
  /// <code>
  /// 11d59m05.60sE 49d33m16.05sN 12d00m36.03sE 49d32m21.97sN sbach.bmp
  /// 6d37m44.00sE 43d16m26.95sN 6d39m10.66sE 43d15m29.04sN sttropez.jpg
  /// </code>
  /// defines two maps, one for Schnaittenbach and one for St. Tropez.
  /// Assuming the existence of correct bitmap files sbach.bmp and sttropez.jpg
  /// you won't see the maps until you come to one of these charming cities.
  /// </item>
  /// <item>
  /// Menu Port/Display off switches the battery-hungry display off. The
  /// display may be switched on again by pressing any of the PDA's buttons.
  /// </item>
  /// <item>
  /// Menu Port/Exit closes the applications. Menu Port/Close closes reading
  /// from the GPS port. The application consumes currently a big amount of
  /// resources and it may be necessary to close reading GPS in order to
  /// do some other resource consuming tasks on the PDA, like watching MPGs
  /// or listening MP3s. Sorry, but on the cheaper divices you won't be
  /// currently able to use sgps.exe and do other things parallelly.
  /// </item>
  /// </list>
  /// </example>
  /// <seealso cref="BndBmpForm">BndBmpForm</seealso>
  /// <seealso cref="Gpsx">Gpsx</seealso>
  /// <seealso cref="IPClient">IPClient</seealso>
  /// <seealso cref="IPClientGPS">IPClientGPS</seealso>
  /// <seealso cref="Serial">Serial</seealso>
  /// <seealso cref="SerialGPS">SerialGPS</seealso>
  /// <seealso cref="Utils.toUTM">Utils.toUTM()</seealso>
  /// <seealso cref="Utils.displayOff">Utils.displayOff()</seealso>
  public class SerialGPSForm : Form {

    /// <summary>Name of the config file</summary>
    private static string CONFIGFILE {
      get {
        return Utils.getExtFileName(".cfg");
      }
    } // CONFIGFILE 

    /// <summary>Name of the map file for Maps</summary>
    private static string MAPFILE {
      get {
        return Utils.getExtFileName(".map");
      }
    } // MAPFILE

    /// <summary>
    /// Names of COM ports available in the system.
    /// </summary>
    private ComboBox m_ports;

    /// <summary>
    /// Control for the baud rate of the port.
    /// </summary>
    private ComboBox m_baud;

    /// <summary>
    /// Control for the number of data bits.
    /// </summary>
    private ComboBox m_databits;

    /// <summary>
    /// Available COM port parities
    /// </summary>
    private ComboBox m_parities;

    /// <summary>
    /// Available COM port parities
    /// </summary>
    private ComboBox m_stopbits;

    /// <summary>UTC time from GPS</summary>
    private Label m_lblutctim;

    /// <summary>Latitude</summary>
    private Label m_lbllat;

    /// <summary>Longitude</summary>
    private Label m_lbllon;

    /// <summary>UTM northing</summary>
    private Label m_lblnutm;

    /// <summary>UTM easting</summary>
    private Label m_lbleutm;

    /// <summary>Altitude</summary>
    private Label m_lblalt;

    /// <summary>
    /// Available systems for UTM coordinates
    /// </summary>
    private ComboBox m_system;

    /// <summary>Type of the input signal</summary>
    /// <remarks>
    /// OFF for the serial port defined in m_ports, m_baud,
    /// m_databits, m_parities and m_stopbits. TCP, UDB or BCAST
    /// for an IP based sorce like this one produced by s2ip.exe.
    /// </remarks>
    private ComboBox m_type;

    /// <summary>Name or IP of the server, if IP used</summary>
    private TextBox m_server;

    /// <summary>TCP or UDP port</summary>
    /// <remarks>
    /// Used only, if the m_type contains TCP, UDP or BCAST.
    /// </remarks>
    private TextBox m_ipport;

    /// <summary>Menu item port/open</summary>
    private MenuItem m_menuopen;

    /// <summary>Menu item config/log</summary>
    private MenuItem m_menulog;

    /// <summary>Menu for showing and hinding the map form</summary>
    private MenuItem m_menumap;

    /// <summary>Stream used for logging</summary>
    private Gpsx m_gpsx;

    /// <summary>The GPS serial port</summary>
    private SerialGPS m_serial;

    /// <summary>Used for the IP communication</summary>
    private IPClientGPS m_ipclient;

    /// <summary>Delegate for showing GPS coordinates</summary>
    /// <remarks>
    /// The serial port thread or event procedure cannot write driectly
    /// into the forms controls, therefore we define a delegate, which is
    /// executed using Form.Invoke().
    /// See also
    /// http://www.ppc-welt.info/community/archive/index.php?t-118493.html
    /// </remarks>
    private EventHandler m_gpscoordcb;

    /// <summary>UTC time obtained from the GPS</summary>
    private double m_utctim;

    /// <summary>Latitude north from the GPS, negative for south</summary>
    private double m_nlat;

    /// <summary>Longitude east from the GPS, negative for west</summary>
    private double m_elon;

    /// <summary>UTM northing from GPS</summary>
    private double m_nutm;

    /// <summary>UTM easting from GPS</summary>
    private double m_eutm;

    /// <summary>Altitude in meters</summary>
    private double m_alt;

    /// <summary>For for showing maps</summary>
    private BndBmpForm m_mapform;

    /// <summary>
    /// Common method for showing exception errors occured in this form.
    /// </summary>
    /// <param name="aTitle">The title of the exception</param>
    /// <param name="anExcp">The excpetion itself</param>
    private void printExcp(string aTitle, Exception anExcp) {
      StringBuilder sb;

      sb = new StringBuilder(aTitle);
      sb.Append('\n');
      sb.Append(anExcp.ToString());
      MessageBox.Show(sb.ToString(), "Error");
      this.dolog(aTitle, anExcp);
    } // printExcp()

    /// <summary>Switches logging on or off</summary>
    private bool Log {
      set {
        if (value) {
          this.m_gpsx.open(Utils.getDateFileName(".gpx"));
        }
        else {
          this.m_gpsx.close();
        }
      } // set
      get {
        return this.m_gpsx.isOpen;
      } // get
    } // Log

    /// <summary>Closes or opens the BndBmpForm</summary>
    /// <remarks>
    /// If set to true, then it opens the map form for the current position.
    /// If the form was already opened, then it closes it and opens again.
    /// Else it closes the map form. While querying, returns true, if the
    /// map form is opened or false else.
    /// </remarks>
    public bool MapForm {
      set {
        if (value) {
          this.MapForm = false;
          this.m_mapform = new BndBmpForm(SerialGPSForm.MAPFILE);
          this.m_mapform.Disposed += new EventHandler(this.onMapFormDispose);
          this.m_mapform.Show();
        }
        else {
          if (this.m_mapform != null) {
            try { this.m_mapform.Close(); } catch(Exception) {}
            this.m_mapform = null;
          }
        }
      }
      get {
        if (this.m_mapform != null && this.m_mapform.Visible) return true;
        return false;
      }
    } // MapForm

    /// <summary>Log method</summary>
    /// <param name="descr">Log entry description</param>
    /// <param name="o">The log content</param>
    private void dolog(string descr, object o) {
      StringBuilder sb;

      try {
        if (this.Log) {
          sb = new StringBuilder();
          sb.Append(DateTime.Now.ToString("yyyyMMdd HHmmss"))
            .Append(' ')
            .Append(descr)
            .Append(' ')
            .Append(o.ToString());
          this.m_gpsx.addComment(sb.ToString());
        }
      }
      catch(Exception) {}
    } // dolog()

    /// <summary>Event method called if the form is closed</summary>
    /// <remarks>
    /// Writes current config to the file and closes serial port.
    /// </remarks>
    private void onClose(object sender, EventArgs e) {
      Exception ee;

      ee = null;
      try { this.m_serial.Close(); } catch(Exception excp) { ee = excp; }
      try {
        this.m_ipclient.oper(ServerType.OFF, null, 0);
      }
      catch(Exception excp) {
        ee = excp;
      }
      try { this.m_gpsx.close(); } catch(Exception excp) { ee = excp; }
      try { this.MapForm = false; } catch(Exception excp) { ee = excp; }
      try { this.writeCfg(); } catch(Exception excp) { ee = excp; }
      if (ee != null) this.printExcp("exit", ee);
    } // onClose()

    /// <summary>Event if the map form is disposed</summary>
    private void onMapFormDispose(object sender, EventArgs e) {
      this.MapForm = false;
      this.m_menumap.Checked = false;
    } // onMapFormDispose()

    /// <summary>
    /// Event handler for the menu item close.
    /// </summary>
    /// <remarks>
    /// Closes the serial port.
    /// </remarks>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuClose_Click(object sender, EventArgs e) {
      try {
        this.m_serial.Close();
        this.m_ipclient.oper(ServerType.OFF, null, 0);
        this.m_menuopen.Checked = false;
      }
      catch(Exception excp) {
        this.printExcp("close", excp);
      }
    } // menuClose_Click()

    /// <summary>Switches logging on or off</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuLog_Click(object sender, EventArgs e) {
      try {
        if (this.Log) {
          this.Log = false;
          this.m_menulog.Checked = false;
        }
        else {
          this.Log = true;
          this.m_menulog.Checked = true;
        }
      }
      catch(Exception excp) {
        this.printExcp("log", excp);
      }
    } // menuLog_Click()

    /// <summary>Shows or hides the map form</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuMap_Click(object sender, EventArgs e) {
      try {
        if (this.MapForm) this.MapForm = false;
        else this.MapForm = true;
        this.m_menumap.Checked = this.MapForm;
      }
      catch(Exception excp) {
        this.printExcp("map", excp);
      }
    } // menuMap_Click()

    /// <summary>
    /// Event handler for the menu item exit.
    /// </summary>
    /// <remarks>
    /// Closes the application.
    /// </remarks>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuExit_Click(object sender, EventArgs e) {
      try {
        this.m_serial.Close();
        this.m_ipclient.oper(ServerType.OFF, null, 0);
        this.Close();
      }
      catch(Exception excp) {
        this.printExcp("exit", excp);
      }
    } // menuExit_Click()

    /// <summary>
    /// Opens the serial port.
    /// </summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuOpen_Click(object sender, EventArgs e) {
      ServerType typ;
      int        port;

      try {
        this.m_menuopen.Checked = false;
        this.m_serial.Close();
        this.m_ipclient.oper(ServerType.OFF, null, 0);
        typ = (ServerType) Enum.Parse(typeof(ServerType),
                                      this.m_type.Text,
                                      false);
        if (typ == ServerType.OFF) {
          this.m_serial.Open(this.m_ports.Text,
                             int.Parse(this.m_baud.Text),
                             int.Parse(this.m_databits.Text),
                             (Parity) Enum.Parse(typeof(Parity),
                                                 this.m_parities.Text,
                                                 false),
                             (StopBits) Enum.Parse(typeof(StopBits),
                                                   this.m_stopbits.Text,
                                                   false),
                             (Datum) Enum.Parse(typeof(Datum),
                                                this.m_system.Text,
                                                false));
        }
        else {
          port = int.Parse(this.m_ipport.Text);
          this.m_ipclient.oper(typ, this.m_server.Text, port);
        }
        this.m_menuopen.Checked = true;
      }
      catch(Exception excp) {
        this.printExcp("open", excp);
      }
    } // menuOpenClick()

    /// <summary>Display off</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void menuDoff_Click(object sender, EventArgs e) {
      Utils.displayOff();
    } // menuDoff_Click()

    /// <summary>Called if the type in m_type changes</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void typeChanged(object sender, EventArgs e) {
      this.deactivateControls();
    } // typeChanged()

    /// <summary>Event procedure called after each new GPS coordinates</summary>
    /// <param name="sender">ignored</param>
    /// <param name="e">ignored</param>
    private void updGPSControls(object sender, EventArgs e) {
      MapPoint point;
      double   utctim, nlat, elon, nutm, eutm, alt;

      try {
        utctim = this.m_utctim;
        nlat = this.m_nlat;
        elon = this.m_elon;
        nutm = this.m_nutm;
        eutm = this.m_eutm;
        alt = this.m_alt;
        this.m_lblutctim.Text = utctim.ToString("00:00:00.000");
        this.m_lbllat.Text = nlat < 0?  (-nlat).ToString("00.0000S"):
                                        nlat.ToString("00.0000N");
        this.m_lbllon.Text = elon < 0?  (-elon).ToString("00.0000W"):
                                        elon.ToString("00.0000E");
        this.m_lblnutm.Text = nutm.ToString("0.000km");
        this.m_lbleutm.Text = eutm.ToString("0.000km");
        this.m_lblalt.Text = alt.ToString("0.0m");
        if (this.MapForm) {
          point = new MapPoint(elon, nlat);
          this.m_mapform.setPoint(point);
        }
      }
      catch(Exception excp) {
        this.dolog("updGPSControls", excp);
      }
    } // updGPSControls()

    /// <summary>Reads configuration into combos</summary>
    /// <remarks>
    /// If the configuration can'b read for some reason, then defaults
    /// are assumed.
    /// </remarks>
    private void readCfg() {
      StreamReader stream;
      string       comport, baudrate, databits, parity, stopbits, system,
                   type, server, ipport;

      stream = null;
      comport = null;
      baudrate = null;
      databits = null;
      parity = null;
      stopbits = null;
      system = null;
      type = null;
      server = null;
      ipport = null;
      try {
        stream = new StreamReader(SerialGPSForm.CONFIGFILE);
        comport = stream.ReadLine();
        baudrate = stream.ReadLine();
        databits = stream.ReadLine();
        parity = stream.ReadLine();
        stopbits = stream.ReadLine();
        system = stream.ReadLine();
        type = stream.ReadLine();
        server = stream.ReadLine();
        ipport = stream.ReadLine();
      }
      catch(Exception excp) {
        this.dolog("read config", excp);
      }
      if (stream != null) stream.Close();
      if (comport == null) comport = Serial.DEFAULT_COMPORT;
      if (baudrate == null) baudrate = Serial.DEFAULT_BAUDRATE.ToString();
      if (databits == null) databits = Serial.DEFAULT_DATABITS.ToString();
      if (parity == null) parity = Serial.DEFAULT_PARITY.ToString();
      if (stopbits == null) stopbits = Serial.DEFAULT_STOPBITS.ToString();
      if (system == null) system = SerialGPS.DEFAULT_SYSTEM.ToString();
      if (type == null) type = ServerType.OFF.ToString();
      if (server == null) server = "169.254.2.1";
      if (ipport == null) ipport = "5000";
      this.m_ports.Text = comport;
      this.m_baud.Text = baudrate;
      this.m_databits.Text = databits;
      this.m_parities.Text = parity;
      this.m_stopbits.Text = stopbits;
      this.m_system.Text = system;
      this.m_type.Text = type;
      this.m_server.Text = server;
      this.m_ipport.Text = ipport;
    } // readCfg()

    /// <summary>Writes setting to the configuration file</summary>
    private void writeCfg() {
      StreamWriter stream;

      stream = null;
      try {
        stream = new StreamWriter(SerialGPSForm.CONFIGFILE);
        stream.WriteLine(this.m_ports.Text);
        stream.WriteLine(this.m_baud.Text);
        stream.WriteLine(this.m_databits.Text);
        stream.WriteLine(this.m_parities.Text);
        stream.WriteLine(this.m_stopbits.Text);
        stream.WriteLine(this.m_system.Text);
        stream.WriteLine(this.m_type.Text);
        stream.WriteLine(this.m_server.Text);
        stream.WriteLine(this.m_ipport.Text);
      }
      catch(Exception excp) {
        this.dolog("write config", excp);
      }
      if (stream != null) stream.Close();
    } // writeCfg()

    /// <summary>Callback method for coordinates received from GPS</summary>
    /// <param name="utcTime">UTC time from GPS</param>
    /// <param name="n_lat">North latitude, negative for south</param>
    /// <param name="e_long">East longitude, negative for west</param>
    /// <param name="utmnorthing">n_lat recomputed to UTM</param>
    /// <param name="utmeasting">e_long recomputed to UTM</param>
    /// <param name="altitude">altitude (elevation) in meters</param>
    private void gpscoord(double utcTime,
                          double n_lat,
                          double e_long,
                          double utmnorthing,
                          double utmeasting,
                          double altitude) {
      try {
        this.m_utctim = utcTime;
        this.m_nlat = n_lat;
        this.m_elon = e_long;
        this.m_nutm = utmnorthing;
        this.m_eutm = utmeasting;
        this.m_alt = altitude;
        if (this.Log) {
          this.m_gpsx.addLocation(n_lat, e_long, altitude);
        }
        this.Invoke(this.m_gpscoordcb);
      }
      catch(Exception excp) {
        this.dolog("coord", excp);
      }
    } // coordinates()

    /// <summary>Activates or deactivates controls depending on type</summary>
    /// <remarks>
    /// A lot of controls must not be used in the IP mode. In the serial
    /// port mode there is no sense for example to use the IP port control.
    /// This method marks controls as grayed, for which there is no sense to
    /// use in the certain mode.
    /// </remarks>
    private void deactivateControls() {
      ServerType typ;

      try {
        typ = (ServerType) Enum.Parse(typeof(ServerType),
                                      this.m_type.Text,
                                      false);
        if (typ == ServerType.OFF) {
          this.m_server.Enabled = false;
          this.m_ipport.Enabled = false;
          this.m_ports.Enabled = true;
          this.m_baud.Enabled = true;
          this.m_databits.Enabled = true;
          this.m_parities.Enabled = true;
          this.m_stopbits.Enabled = true;
        }
        else {
          this.m_server.Enabled = true;
          this.m_ipport.Enabled = true;
          this.m_ports.Enabled = false;
          this.m_baud.Enabled = false;
          this.m_databits.Enabled = false;
          this.m_parities.Enabled = false;
          this.m_stopbits.Enabled = false;
        }
      }
      catch(Exception excp) {
        this.dolog("deactivateControls", excp);
      }
    } // deactivateControls()

    /// <summary>Layout of the form</summary>
    private void InitializeComponent() {
      MainMenu   menu;
      MenuItem[] mit_port;
      MenuItem   menuPort;
      Label      titutc, titcoord, titutm, titalt;
      Size       size1, size2, size3, size4, size5;

      size1 = new Size(60, 30);
      size2 = new Size(50, 30);
      size3 = new Size(40, 30);
      size4 = new Size(80, 15);
      size5 = new Size(40, 15);
      this.SuspendLayout();
      mit_port = new MenuItem[6];
      mit_port[0] = this.m_menuopen;
      mit_port[1] = new MenuItem();
      mit_port[1].Text = "Close";
      mit_port[1].Click += menuClose_Click;
      mit_port[2] = this.m_menulog;
      mit_port[3] = this.m_menumap;
      mit_port[4] = new MenuItem();
      mit_port[4].Text = "Display off";
      mit_port[4].Click += menuDoff_Click;
      mit_port[4].Enabled = Utils.isMobile();
      mit_port[5] = new MenuItem();
      mit_port[5].Text = "Exit";
      mit_port[5].Click += menuExit_Click;
      menuPort = new MenuItem();
      menuPort.Text = "Port";
      menuPort.MenuItems.Add(mit_port[0]);
      menuPort.MenuItems.Add(mit_port[1]);
      menuPort.MenuItems.Add(mit_port[2]);
      menuPort.MenuItems.Add(mit_port[3]);
      menuPort.MenuItems.Add(mit_port[4]);
      menuPort.MenuItems.Add(mit_port[5]);
      menu = new MainMenu();
      menu.MenuItems.Add(menuPort);
      this.ClientSize = new Size(210, 155);
      this.m_ports.Location = new Point(5, 5);
      this.m_ports.Size = size1;
      this.m_baud.Location = new Point(65, 5);
      this.m_baud.Size = size2;
      this.m_databits.Location = new Point(115, 5);
      this.m_databits.Size = size3;
      this.m_parities.Location = new Point(155, 5);
      this.m_parities.Size = size2;
      this.m_stopbits.Location = new Point(5, 25);
      this.m_stopbits.Size = size1;
      this.m_system.Location = new Point(65, 25);
      this.m_system.Size = new Size(140, 30);
      this.m_type.Location = new Point(5, 45);
      this.m_type.Size = size1;
      this.m_server.Location = new Point(65, 45);
      this.m_server.Size = new Size(100, 15);
      this.m_ipport.Location = new Point(170, 45);
      this.m_ipport.Size = new Size(35, 15);
      titutc = new Label();
      titutc.Text = "UTC:";
      titutc.Location = new Point(5, 70);
      titutc.Size = size5;
      this.m_lblutctim.Location = new Point(45, 70);
      this.m_lblutctim.Size = size4;
      titcoord = new Label();
      titcoord.Location = new Point(5, 90);
      titcoord.Text = "Coord:";
      titcoord.Size = size5;
      this.m_lbllat.Location = new Point( 45, 90);
      this.m_lbllat.Size = size4;
      this.m_lbllon.Location = new Point(125, 90);
      this.m_lbllon.Size = size4;
      titutm = new Label();
      titutm.Text = "UTM:";
      titutm.Size = size5;
      titutm.Location = new Point(5, 110);
      this.m_lblnutm.Location = new Point( 45, 110);
      this.m_lblnutm.Size = size4;
      this.m_lbleutm.Location = new Point(125, 110);
      this.m_lbleutm.Size = size4;
      titalt = new Label();
      titalt.Text = "Alt:";
      titalt.Size = size5;
      titalt.Location = new Point(5, 130);
      this.m_lblalt.Location = new Point(45, 130);
      this.m_lblalt.Size = size4;
      this.AutoScroll = true;
      this.Controls.Add(this.m_ports);
      this.Controls.Add(this.m_baud);
      this.Controls.Add(this.m_databits);
      this.Controls.Add(this.m_parities);
      this.Controls.Add(this.m_stopbits);
      this.Controls.Add(this.m_system);
      this.Controls.Add(this.m_type);
      this.Controls.Add(this.m_server);
      this.Controls.Add(this.m_ipport);
      this.Controls.Add(titutc);
      this.Controls.Add(this.m_lblutctim);
      this.Controls.Add(titcoord);
      this.Controls.Add(this.m_lbllat);
      this.Controls.Add(this.m_lbllon);
      this.Controls.Add(titutm);
      this.Controls.Add(this.m_lblnutm);
      this.Controls.Add(this.m_lbleutm);
      this.Controls.Add(titalt);
      this.Controls.Add(this.m_lblalt);
      this.Name = "SerialGPSForm";
      this.Text = "GPS";
      this.Menu = menu;
      this.ResumeLayout(false);
      this.Closing += onClose;
    } // InitializeComponent()

    /// <summary>
    /// Creates the form
    /// </summary>
    public SerialGPSForm() {
      CoordDelegate    cb;
      Stream           str;
      Icon             ico;

      this.m_utctim = 0.0;
      this.m_nlat = 0.0;
      this.m_elon = 0.0;
      this.m_nutm = 0.0;
      this.m_eutm = 0.0;
      this.m_alt = 0.0;
      this.m_mapform = null;
      this.m_ports = new ComboBox();
      this.m_baud = new ComboBox();
      this.m_databits = new ComboBox();
      this.m_parities = new ComboBox();
      this.m_stopbits = new ComboBox();
      this.m_system = new ComboBox();
      this.m_type = new ComboBox();
      this.m_server = new TextBox();
      this.m_ipport = new TextBox();
      this.m_lblutctim = new Label();
      this.m_lbllat = new Label();
      this.m_lbllon = new Label();
      this.m_lblnutm = new Label();
      this.m_lbleutm = new Label();
      this.m_lblalt = new Label();
      this.m_menuopen = new MenuItem();
      this.m_menulog = new MenuItem();
      this.m_menumap = new MenuItem();
      this.m_menuopen.Text = "Open";
      this.m_menuopen.Click += menuOpen_Click;
      this.m_menulog.Text = "Log";
      this.m_menulog.Click += menuLog_Click;
      this.m_menumap.Text = "Map";
      this.m_menumap.Click += menuMap_Click;
      this.m_type.TextChanged += this.typeChanged;
      this.m_gpsx = new Gpsx();
      this.m_serial = new SerialGPS();
      this.m_ipclient = new IPClientGPS();
      this.m_gpscoordcb = new EventHandler(this.updGPSControls);
      cb = new CoordDelegate(this.gpscoord);
      this.m_serial.CoordCb = cb;
      this.m_ipclient.CoordCb = cb;
      this.InitializeComponent();
      // fill combos with all available values
      foreach(string s in Serial.COMPORTS) this.m_ports.Items.Add(s);
      foreach(int i in Serial.BAUDRATES) this.m_baud.Items.Add(i);
      foreach(int i in Serial.DATABITS) this.m_databits.Items.Add(i);
      this.m_parities.Items.Add(Parity.Even.ToString());
      this.m_parities.Items.Add(Parity.Mark.ToString());
      this.m_parities.Items.Add(Parity.None.ToString());
      this.m_parities.Items.Add(Parity.Odd.ToString());
      this.m_parities.Items.Add(Parity.Space.ToString());
      this.m_stopbits.Items.Add(StopBits.None.ToString());
      this.m_stopbits.Items.Add(StopBits.One.ToString());
      this.m_stopbits.Items.Add(StopBits.OnePointFive.ToString());
      this.m_stopbits.Items.Add(StopBits.Two.ToString());
      this.m_system.Items.Add(Datum.WGS84.ToString());
      this.m_system.Items.Add(Datum.GRS80.ToString());
      this.m_system.Items.Add(Datum.WGS72.ToString());
      this.m_system.Items.Add(Datum.AUSTRALIAN1965.ToString());
      this.m_system.Items.Add(Datum.KRASOVSKY1940.ToString());
      this.m_system.Items.Add(Datum.INTERNATIONAL1924.ToString());
      this.m_system.Items.Add(Datum.CLAKE1880.ToString());
      this.m_system.Items.Add(Datum.CLARKE1866.ToString());
      this.m_system.Items.Add(Datum.AIRY1830.ToString());
      this.m_system.Items.Add(Datum.BESSEL1841.ToString());
      this.m_system.Items.Add(Datum.EVEREST1830.ToString());
      this.m_type.Items.Add(ServerType.OFF.ToString());
      this.m_type.Items.Add(ServerType.TCP.ToString());
      this.m_type.Items.Add(ServerType.UDP.ToString());
      this.m_type.Items.Add(ServerType.BCAST.ToString());
      this.readCfg();
      this.deactivateControls();
      try {
        str = Assembly.GetExecutingAssembly().GetManifestResourceStream("Marta1.ico");
        ico = new Icon(str);
        str.Close();
        this.Icon = ico;
      }
      catch(Exception excp) {
        this.printExcp("ctor", excp);
      }
    } // SerialGPSForm()

  } // class SerialGPSForm
} // namespace serial

