using System;

namespace serial {

  /// <summary>Command line test program for UTM translations</summary>
  /// <remarks>
  /// There are two cases of usage. Either two geographical coordinates
  /// having format degree minutes. partial minutes, first the
  /// the latitude North then the longitude East or the whole NMEA GGA sequence
  /// may be passed to the command line, completed by the name of the UTM system
  /// to use.
  /// </remarks>
  /// <example>
  /// <code>
  /// utm.exe 4932,8682 1200,0293 WGS84
  /// 4932,8682N 1200,0293E system WGS84, UTM 547978,303650611 597,637501493447
  /// utm.exe $GPGGA,061254.000,4932.8682,N,01200.0293,E,1,03,4.1,424.4,M,47.4,M,,0000*5C WGS84
  /// utc 06:12:54,000 49,547803N 12,000488E alt 424,4 system WGS84 UTM 5492,679km 283,040km
  /// </code>
  /// In the first case the coordinates 49 deg. 32.8602 min. North and 
  /// 12 deg. 0.0293 min East are recomputed to UTM, ellipsoide WGS84. In the second
  /// case the same coordinates are computed from a NMEA GGA sequence produced by
  /// a GPS chip. Possible values for the ellisoide types:
  /// WGS84, GRS80, WGS72, AUSTRALIAN1965, KRASOVSKY1940, INTERNATIONAL1924,
  /// CLAKE1880, CLARKE1866, IRY1830, BESSEL1841, EVEREST1830
  /// </example>
  /// <seealso cref="Utils.toUTM">Utils.toUTM()</seealso>
  public class UTMTest {

    /// <summary>The main method</summary>
    /// <param name="args">
    /// command line args.
    ///
    /// Case 1: latitude north, longitude east, system.
    /// Latitude and longitude in decimal degress, west and south may be entered
    /// as negative values. system is one of the enum Datum: WGS84, GRS80,
    /// WGS72 etc.
    ///
    /// Case 2: GGA, system.
    /// GGA is a NMEA GGA string according to Utils.fromGGA(), system is one of
    /// enum Utils.Datum. 
    /// </param>
    public static void Main(string[] args) {
      Datum  d;
      double n_lat, e_long, utmnorthing, utmeasting, utctim, altitude;
      
      try {
        if (args.Length >= 3) {
          n_lat = double.Parse(args[0]);
          e_long = double.Parse(args[1]);
          d = (Datum) Datum.Parse(typeof(Datum), args[2], false);
          Utils.toUTM(out utmnorthing, out utmeasting, n_lat, e_long, d);
          Console.WriteLine("{0}N {1}E system {2}, UTM {3} {4}",
                            n_lat, e_long, d, utmnorthing, utmeasting);
        }
        else if (args.Length >= 2) {
          d = (Datum) Datum.Parse(typeof(Datum), args[1], false);
          if (Utils.fromGGA(out utctim,
                            out n_lat,
                            out e_long,
                            out utmnorthing,
                            out utmeasting,
                            out altitude,
                            args[0],
                            d))
            Console.WriteLine("utc {0} {1} {2} alt {3} system {4} UTM {5} {6}",
                              utctim.ToString("00:00:00.000"),
                              n_lat > 0? n_lat.ToString("00.000000N") :
                                         (-n_lat).ToString("00.000000S"),
                              e_long > 0? e_long.ToString("00.000000E") :
                                          (-e_long).ToString("00.000000W"),
                              altitude,
                              d,
                              utmnorthing.ToString("0.000km"),
                              utmeasting.ToString("0.000km"));
          else
            Console.WriteLine("invalid or empty GGA string");
        }
        else {
          Console.WriteLine("Usage:");
          Console.WriteLine("{0} latitude longitude system", Utils.Exepath);
          Console.WriteLine("{0} GGA system", Utils.Exepath);
          Console.WriteLine("latitude is the north latitude in decimal degrees");
          Console.WriteLine("longitude is the east longitude in decimal degrees");
          Console.WriteLine("GGA is an NMEA GGA string");
          Console.WriteLine("system may be one of WGS84, GRS80, WGS72, AUSTRALIAN1965,");
          Console.WriteLine("KRASOVSKY1940, INTERNATIONAL1924, CLAKE1880, CLARKE1866,");
          Console.WriteLine("AIRY1830, BESSEL1841, EVEREST1830");
        }
      }
      catch(Exception excp) {
        Console.WriteLine(excp.ToString());
      }
    } // main()

  } // class UTMTest

} // namespace serial

