using System;
using System.Collections;
using System.IO;
using System.Globalization;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;
using System.Text.RegularExpressions;

namespace serial {

  /// <summary>Datum types used for UTM</summary>
  public enum Datum : int {
    /// <summary>NAD83/WGS84 use Global</summary>
    WGS84 = 0,
    /// <summary>GRS  use US</summary>
    GRS80 = 1,
    /// <summary>WGS72 use NASA, DOD</summary>
    WGS72 = 2,
    /// <summary>Australian  use Australia</summary>
    AUSTRALIAN1965 = 3,
    /// <summary>Krasovsky  use Soviet Union</summary>
    KRASOVSKY1940 = 4,
    /// <summary>International (1924) -Hayford (1909) use Global</summary>
    INTERNATIONAL1924 = 5,
    /// <summary>Clake use France, Africa</summary>
    CLAKE1880 = 6,
    /// <summary>Clarke use North America</summary>
    CLARKE1866 = 7,
    /// <summary>Airy use Great Britain</summary>
    AIRY1830 = 8,
    /// <summary>Bessel use Central Europe, Chile, Indonesia</summary>
    BESSEL1841 = 9,
    /// <summary>Everest use South Asia</summary>
    EVEREST1830 = 10
  } // enum Datum

  /// <summary>Message types for Utils.psrf103()</summary>
  public enum Psrf103msg : byte {
    /// <summary>GGA message</summary>
    GGA = 0,
    /// <summary>GLL message</summary>
    GLL = 1,
    /// <summary>GSA message</summary>
    GSA = 2,
    /// <summary>GSV message</summary>
    GSV = 3,
    /// <summary>RMC message</summary>
    RMC = 4,
    /// <summary>VTG message</summary>
    VTG = 5
  } // enum Psrf103msg

  /// <summary>The IP server type</summary>
  public enum ServerType : short {
    /// <summary>Server is witched off</summary>
    OFF,
    /// <summary>server is on and works as TCP</summary>
    TCP,
    /// <summary>server is on and works as UDP</summary>
    UDP,
    /// <summary>server is on and broadcasts on UDP</summary>
    BCAST
  } // enum ServerType

#region delegates
  /// <summary>
  /// Callback for line strings read from the serial port or IP.
  /// </summary>
  /// <param name="arr">
  /// Array of line strings received from the serial port
  /// </param>
  public delegate void ReadLineDelegate(ArrayList arr);

  /// <summary>Callback for single byzes read from the serial port</summary>
  /// <param name="b">The byte array read from the serial port</param>
  /// <param name="len">Number of valid bytes in b</param>
  public delegate void ReadByteDelegate(byte[] b, int len);

  /// <summary>Callback for position informations obtained from GGA</summary>
  /// <param name="utcTime">UTC time from GPS</param>
  /// <param name="n_lat">North latitude, negative for south</param>
  /// <param name="e_long">East longitude, negative for west</param>
  /// <param name="utmnorthing">n_lat recomputed to UTM</param>
  /// <param name="utmeasting">e_long recomputed to UTM</param>
  /// <param name="altitude">altitude (alevation) in meters</param>
  public delegate void CoordDelegate(double utcTime,
                                     double n_lat,
                                     double e_long,
                                     double utmnorthing,
                                     double utmeasting,
                                     double altitude);
#endregion

  /// <summary>Common purpose utilities</summary>
  public class Utils {

#region PInvoke
    /// <summary>Resets idle time and prevents suspending</summary>
    /// <remarks>Only on windows ce/mobile</remarks>
    [DllImport("coredll.dll",
               EntryPoint="SystemIdleTimerReset",
               SetLastError=false)]
    private static extern void SystemIdleTimerReset();

    /// <summary>Retreives device context</summary>
    /// <remarks>Used internally for switching off display</remarks>
    /// <param name="hwnd">ignored, the code passes always NULL here</param>
    [DllImport("coredll.dll",
               EntryPoint="GetDC",
               SetLastError=false)]
    private static extern IntPtr GetDC(IntPtr hwnd);

    /// <summary>Used in order to switch off display</summary>
    /// <remarks>
    /// http://www.pinvoke.net/default.aspx/coredll/ExtEscape.html
    /// </remarks>
    /// <param name="hdc">device context, GetDC(IntPtr.Zero)</param>
    /// <param name="nEscape">SETPOWERMANAGEMENT, 6147</param>
    /// <param name="cbInput">size of lpszInData, 12</param>
    /// <param name="lpszInData">size 12, dpms ver. 1, videoPowerOff 4</param>
    /// <param name="cbOutput">ignored, 0</param>
    /// <param name="lpszOutData">ignored, IntPtr.Zero</param>
    /// <returns>ignored</returns>
    [DllImport("coredll.dll",
               EntryPoint="ExtEscape",
               SetLastError=false)]
    private static extern int ExtEscape(IntPtr hdc,
                                        int    nEscape,
                                        int    cbInput,
                                        byte[] lpszInData,
                                        int    cbOutput,
                                        IntPtr lpszOutData);
#endregion

    /// <summary>Equatorial Radius in meters, UTM</summary>
    /// <remarks>For systems like in the enumeration Datum</remarks>
    private static readonly double[] A = new double[] {
      6378137.0,6378137.0,6378135.0,6378160.0,6378245.0,6378388.0,
      6378249.1,6378206.4,6377563.4,6377397.2,6377276.3
     };

    /// <summary>Polar Radius in meters, UTM</summary>
    /// <remarks>For systems like in the enumeration Datum</remarks>
    private static readonly double[] B = new double[] {
      6356752.3142,6356752.3141,6356750.5000,6356774.7000,
      6356863.0000,6356911.9000,6356514.9000,6356583.8000,
      6356256.9000,6356079.0000,6356075.4000
     };

    /// <summary>Separators for spliting NMEA GGA strings</summary>
    private static readonly char[] GGASEPS = new char[] { ',' };

    /// <summary>SIRF NMEA PSRF103 message pattern</summary>
    /// <seealso cref="Utils.psrf103">Utils.psrf103</seealso>
    private static readonly byte[] PSRF103PAT = 
      Encoding.ASCII.GetBytes("$PSRF103,00,00,00,00*00\r\n");

    /// <summary>Regular expression used for parsing coordinates</summary>
    /// <seealso cref="Utils.coord2dec">Utils.coord2dec()</seealso>
    private static Regex RECOORD = new Regex(
      @"^\s*(?<deg>\d+)(|d)\s*(?<min>\d+)('|m)\s*(?<sec>\d*(\.\d+|\d+))(''|s)\s*(?<dir>(E|e|W|w|N|n|S|s))\s*"
     );

    /// <summary>Checks if running under MS Windows</summary>
    /// <returns>true for windows</returns>
    public static bool isWindows() {
      bool retval;

      switch(Environment.OSVersion.Platform) {
        case PlatformID.Win32NT :
        case PlatformID.Win32S :
        case PlatformID.Win32Windows :
        case PlatformID.WinCE :
          retval = true;
          break;
        default :
          retval = false;
          break;
      } // switch
      return retval;
    } // isWindows()

    /// <summary>Checks if the platform is windows ce</summary>
    /// <returns>True for windows ce</returns>
    public static bool isMobile() {
      if (Environment.OSVersion.Platform == PlatformID.WinCE) return true;
      return false;
    } // isMobile()

    /// <summary>Causes that the system won't suspend</summary>
    /// <remarks>
    /// Only on windows ce/mobile. For other platforms the method does nothing.
    /// </remarks>
    public static void dontSuspend() {
      if (Utils.isMobile()) {
        try {
          Utils.SystemIdleTimerReset();
        }
        catch(Exception) {}
      }
    } // dontSuspend()

    /// <summary>Switches display of the mobile device off</summary>
    /// <remarks>
    /// Only on windows ce/mobile. For other platforms the method does nothing.
    /// http://www.pinvoke.net/default.aspx/coredll/ExtEscape.html
    /// </remarks>
    public static void displayOff() {
      const int SETPOWERMANAGEMENT = 6147;
      const int VIDEOPOWEROFF = 4;
      const int SIZ = 12;
      const int DPMSVER = 1;
      IntPtr    hdc;
      byte[]    vpm;

      if (Utils.isMobile()) {
        hdc = Utils.GetDC(IntPtr.Zero);
        vpm = new byte[SIZ];
        BitConverter.GetBytes(SIZ).CopyTo(vpm, 0);
        BitConverter.GetBytes(DPMSVER).CopyTo(vpm, 4);
        BitConverter.GetBytes(VIDEOPOWEROFF).CopyTo(vpm, 8);
        Utils.ExtEscape(hdc, SETPOWERMANAGEMENT, SIZ, vpm, 0, IntPtr.Zero);
      }
    } // displayOff()

    /// <summary>The full path to the executing assembly</summary>
    public static string Exepath {
      get {
        Uri    u;
        string s;

        s = Assembly.GetExecutingAssembly().GetName().CodeBase;
        u = new Uri(s);
        if (u.IsFile) s = u.LocalPath;
        return s;
      }
    } // Exepath

    /// <summary>The directory where the executing assembly is located</summary>
    public static string Exedir {
      get {
        return Path.GetDirectoryName(Utils.Exepath);
      }
    } // Exedir

    /// <summary>
    /// Creates a file name consisting of the exe name and the given extension
    /// </summary>
    /// <param name="ext">the desired extension</param>
    /// <returns>
    /// File name consiting of the exe name with the extension replaced by the
    /// given extension. Eg. Exepath is /usr/bin/test.exe, ext is cfg, the
    /// resulting file name will be /usr/bin/test.cfg.
    /// </returns>
    public static string getExtFileName(string ext) {
      return Path.ChangeExtension(Utils.Exepath, ext);
    } // getExtFileName()

    /// <summary>
    /// Creates a file name with the given extentsion and a timestamp in the name.
    /// </summary>
    /// <remarks>
    /// Usefull for creating unique file names etc.
    /// </remarks>
    /// <param name="ext">the desired extension</param>
    /// <returns>
    /// Filename consisting of exe dir, exe file name, present date yyyyMMddHHmmss,
    /// the desired extension.
    /// </returns>
    public static string getDateFileName(string ext) {
      StringBuilder sb;
      string        exe, dir;

      exe = Utils.Exepath;
      dir = Path.GetDirectoryName(exe);
      sb = new StringBuilder(Path.GetFileNameWithoutExtension(exe));
      sb.Append(DateTime.Now.ToString("yyyyMMddHHmmss"));
      sb.Append(ext);
      return Path.Combine(dir, sb.ToString());
    } // getUniqFileName()

    /// <summary>Converts latitude to UTM northing</summary>
    /// <remarks>
    /// Based on the theoretical description and an Excel sheet from
    /// Prof. Steve Dutch, see:
    /// <A HREF="http://www.uwgb.edu/DutchS/UsefulData/UTMFormulas.HTM">http://www.uwgb.edu/DutchS/UsefulData/UTMFormulas.HTM</A> and
    /// <A HREF="http://www.uwgb.edu/DutchS/UsefulData/UTMConversions1.xls">http://www.uwgb.edu/DutchS/UsefulData/UTMConversions1.xls</A>.
    /// </remarks>
    /// <param name="utmnorthing">The resulting UTM northing in km</param>
    /// <param name="utmeasting">The resulting UTM easting in km</param>
    /// <param name="n_lat">
    /// north latitude in decimal degrees, negative for south
    /// </param>
    /// <param name="e_long">
    /// east longitude in decimal degress, negative for west
    /// </param>
    /// <param name="system">
    /// the system used for recomputing, WGS84, GRS80 etc.
    /// </param>
    /// <returns>UTM nortihng, negative for south</returns>
    public static void toUTM(out double utmnorthing, out double utmeasting,
                             double n_lat, double e_long, Datum system) {
      const double k0 = 0.9996;
      const double sin1sec1 = 4.8481368110763677e-006; // sin(1")
      const double sin1sec2 = 2.3504430538913731e-011; // sin(1")^2
      const double sin1sec3 = 1.1395269491909521e-016; // sin(1")^3
      const double sin1sec4 = 5.5245825495862045e-022; // sin(1")^4
      double       a, b, a2, b2, n, n_1, n2, n3, n4, n5, n2_3, n4_5,
                   aprim, bprim, cprim, dprim, eprim, s,
                   lat, lat2, lat4, lat6, lat8, k1, k2, k3, k4, k5,
                   sinlat, coslat, coslat2, coslat3, coslat4, tanlat, tanlat2,
                   nu, e, e2, e4, p, p2, p3, p4;
      int          idx, lonzone, lonzonecm;

      idx = (int) system;
      a = Utils.A[idx];
      b = Utils.B[idx];
      a2 = a * a; b2 = b * b;
      e = Math.Sqrt((a2 - b2) / a2);
      e2 = e * e; e4 = e2 * e2;
      n = (a - b) / (a + b);
      n2 = n * n; n3 = n2 * n; n4 = n3 * n; n5 = n4 * n;
      n2_3 = n2 - n3; n4_5 = n4 - n5;
      n_1 = 1.0 - n;
      aprim = a * (n_1 + 1.25 * n2_3 + 1.265625 * n4_5);
      bprim = 1.5 * a * n * (n_1 + 0.875 * n2_3 + 0.859375 * n4_5);
      cprim = 0.9375 * a * n2 * (n_1 + 0.75 * n2_3);
      dprim = 35.0 * a * n3 * (n_1 + 11.0 * n2_3 / 16.0) / 48.0;
      eprim = 315.0 * a * n4 * n_1 / 51.0;
      lat = n_lat * Math.PI / 180.0;
      lat2 = lat + lat;
      lat4 = lat2 + lat2;
      lat6 = lat4 + lat2;
      lat8 = lat4 + lat4;
      sinlat = Math.Sin(lat);
      coslat = Math.Cos(lat);
      coslat2 = coslat * coslat;
      coslat3 = coslat2 * coslat;
      coslat4 = coslat2 * coslat2;
      tanlat = Math.Tan(lat);
      tanlat2 = tanlat * tanlat;
      lonzone = 31 + (int) Math.Floor(e_long / 6);
      lonzonecm = 6 * lonzone - 183;
      p = 0.36 * (e_long - lonzonecm); // (lon - lonzonecm) * 3600 / 10000
      p2 = p * p; p3 = p2 * p; p4 = p2 * p2;
      s = aprim * lat - bprim * Math.Sin(lat2) + cprim * Math.Sin(lat4) -
          dprim * Math.Sin(lat6) + eprim * Math.Sin(lat8);
      nu = a / Math.Sqrt(1.0 - e2 * sinlat * sinlat);
      k1 = s * k0;
      k2 = nu * sinlat * coslat * sin1sec2 * k0 * 50000000;
      k3 = k0 * sin1sec4 * nu * sinlat * coslat3 *
           (5.0 - tanlat * tanlat + 9.0 * e2 * coslat2 + 4.0 * e4 * coslat4) *
           1e16/ 24.0;
      k4 = nu * coslat * sin1sec1 * k0 * 10000;
      k5 = 1000000000000.0 * k0 * sin1sec3 * nu * coslat3 * 
           (1.0 - tanlat2 + e2 * coslat2) / 6.0;
      utmnorthing = 0.001 * (k1 + k2 * p2 + k3 * p4);
      utmeasting = 0.001 * (500000.0 + k4 * p + k5 * p3);
    } // toUTM()

    /// <summary>Converts GGA numbers deg, min, sec to decimal degrees</summary>
    /// <param name="dms">Double number in format degdegminmin.part</param>
    /// <returns>decimal degress</returns>
    private static double dms2dec(double dms) {
      double deg, min, deg100, part;
   
      deg = Math.Floor(0.01 * dms);
      deg100 = 100 * deg;
      min = Math.Floor(dms - deg100);
      part = dms - deg100 - min;
      return deg + (min + part)/ 60;
    } // dms2dec()

    /// <summary>Parses coordinates from an NMEA GGA string</summary>
    /// <remarks>
    /// Parses NMEA GGA strings according to http://gpsd.berlios.de/NMEA.txt
    /// <code>
    /// GGA - Global Positioning System Fix Data
    /// Time, Position and fix related data for a GPS receiver.
    /// 
    ///  0      1         2       3 4        5 6 7  8   9  10 |  12 13  14   15
    ///  |      |         |       | |        | | |  |   |   | |   | |   |    |
    ///  $--GGA,hhmmss.ss,llll.ll,a,yyyyy.yy,a,x,xx,x.x,x.x,M,x.x,M,x.x,xxxx*hh
    /// 
    ///  Field Number:
    ///   1) Universal Time Coordinated (UTC)
    ///   2) Latitude
    ///   3) N or S (North or South)
    ///   4) Longitude
    ///   5) E or W (East or West)
    ///   6) GPS Quality Indicator,
    ///      0 - fix not available,
    ///      1 - GPS fix,
    ///      2 - Differential GPS fix
    ///      (values above 2 are 2.3 features)
    ///      3 - PPS fix
    ///      4 - Real Time Kinematic
    ///      5 - Float RTK
    ///      6 - estimated (dead reckoning)
    ///      7 - Manual input mode
    ///      8 - Simulation mode
    ///   7) Number of satellites in view, 00 - 12
    ///   8) Horizontal Dilution of precision (meters)
    ///   9) Antenna Altitude above/below mean-sea-level (geoid) (in meters)
    ///  10) Units of antenna altitude, meters
    ///  11) Geoidal separation, the difference between the WGS-84 earth
    ///      ellipsoid and mean-sea-level (geoid), "-" means mean-sea-level
    ///      below ellipsoid
    ///  12) Units of geoidal separation, meters
    ///  13) Age of differential GPS data, time in seconds since last SC104
    ///      type 1 or 9 update, null field when DGPS is not used
    ///  14) Differential reference station ID, 0000-1023
    ///  15) Checksum
    ///
    /// Example:
    /// $GPGGA,200329.000,4932.7248,N,01159.2531,E,1,06,1.5,401.0,M,47.4,M,,0000*5A
    /// </code>
    /// This method neighter uses nor validates the entire GGA string.
    /// For this method it is enough, that the fields 1, 2, 4 and 9 are valid
    /// floting point numbers with a dot as decimal separator, the fields
    /// 3 and 5 contain respectively N/S and E/W and that the field 0
    /// has the format $..GGA meaning: char[0] is '$', char[3] and char[4] are
    /// both 'G' and char[5] is 'A'.
    /// </remarks>
    /// <param name="utcTime">
    /// Universal Time Coordinated, 0 for invalid GGA strings
    /// </param>
    /// <param name="n_lat">
    /// latitude north, negative for south, 0 for invalid GGA strings
    /// </param>
    /// <param name="e_long">
    /// longitude east, negative for west, 0 for invalid GGA strings
    /// </param>
    /// <param name="utmnorthing">
    /// utm northing coordinate resulting from toUTM, 0 for invalid GGA strings
    /// </param>
    /// <param name="utmeasting">
    /// utm easting coordinate resulting from toUTM, 0 for invalid GGA strings
    /// </param>
    /// <param name="altitude">
    /// the altitude (elevation) in meters
    /// </param>
    /// <param name="gga">
    /// the NMEA GGA string to parse
    /// </param>
    /// <param name="system">
    /// the system to use for translating into UTM coordinates, eg. WGS84
    /// </param>
    /// <returns>
    /// true for success, false for invalid GGA strings.
    /// </returns>
    public static bool fromGGA(out double utcTime,
                               out double n_lat,
                               out double e_long,
                               out double utmnorthing,
                               out double utmeasting,
                               out double altitude,
                               string     gga,
                               Datum      system) {
      const NumberStyles ST = NumberStyles.AllowDecimalPoint |
                              NumberStyles.AllowLeadingSign |
                              NumberStyles.AllowLeadingWhite |
                              NumberStyles.AllowTrailingWhite;
      string[] sa;
      double   lat, lon, tim, alt;
      int      i;
      bool     retval;

      retval = false;
      utcTime = 0.0;
      n_lat = 0.0;
      e_long = 0.0;
      utmnorthing = 0.0;
      utmeasting = 0.0;
      altitude = 0.0;
      if (string.IsNullOrEmpty(gga) || gga.Length <= 6 ||
          gga[0] != '$' || gga[3] != 'G' || gga[4] != 'G' || gga[5] != 'A') {
        goto end;
      }
      sa = gga.Split(Utils.GGASEPS);
      if (sa.Length < 10) goto end;
      for(i = 0; i < 10; i++) if (string.IsNullOrEmpty(sa[i])) goto end;
      try {
        tim = Double.Parse(sa[1], ST, CultureInfo.InvariantCulture);
        lat = Double.Parse(sa[2], ST, CultureInfo.InvariantCulture);
        lon = Double.Parse(sa[4], ST, CultureInfo.InvariantCulture);
        alt = Double.Parse(sa[9], ST, CultureInfo.InvariantCulture);
      }
      catch(Exception) {
        goto end;
      }
      lat = Utils.dms2dec(lat);
      lon = Utils.dms2dec(lon);
      switch(sa[3].Trim()) {
        case "N" :
        case "n" :
          break;
        case "S" :
        case "s" :
          lat = -lat; break;
        default  :
          goto end;
      }
      switch(sa[5].Trim()) {
        case "E" :
        case "e" :
          break;
        case "W" :
        case "w" :
          lon = -lon; break;
        default  :
          goto end;
      }
      Utils.toUTM(out utmnorthing, out utmeasting, lat, lon, system);
      utcTime = tim;
      n_lat = lat;
      e_long = lon;
      altitude = alt;
      retval = true;
      end:
      return retval;
    } // fromGGA()

    /// <summary>Computes nme checksum for the given NMEA string</summary>
    /// <remarks>
    /// The checksum for an NMEA string is computed as an XOR value for all
    /// bytes between (but not including) the dollar sign and the asterisk
    /// character, according to
    /// http://www.codepedia.com/1/Taking+Out+the+Garbage:+NMEA+Checksums
    /// This method computes the checksum from the first dollar character
    /// and the first asterisk after it or up to the end of the array, if
    /// asterisk has not been found. If the array has at least two bytes
    /// more after the asterisk, the the checksum is additionally put
    /// hexadecimally in there. Exammple, for the followig byte array
    /// <code>$PSRF103,00,01,00,01*xx\r\n</code>
    /// the method computes checksum 37 (hex 25) and puts it into the
    /// two bytes xx after the asterisk character. The resulting byte
    /// array will be then:
    /// <code>$PSRF103,00,01,00,01*25\r\n</code>
    /// </remarks>
    /// <param name="nmea">The NMEA message for the checksum</param>
    /// <returns>The computed checksum</returns>
    public static byte nmeaChksum(byte[] nmea) {
      string s;
      int    i;
      byte   retval, dollar, asterisk, b;

      retval = 0;
      if (nmea != null) {
        dollar = Convert.ToByte('$');
        asterisk = Convert.ToByte('*');
        // find the first occurence of dollar
        for(i = 0; i < nmea.Length && nmea[i] != dollar; i++);
        i++;
        // compute the checksum
        for(; i < nmea.Length; i++) {
          b = nmea[i];
          if (b == asterisk) break;
          retval ^= b;
        }
        // put the checksum after the asterisk, if possible
        if (i + 2 < nmea.Length) {
          s = retval.ToString("X2");
          nmea[i+1] = Convert.ToByte(s[0]);
          nmea[i+2] = Convert.ToByte(s[1]);
        }
      } // for
      return retval;
    } // nmeaChksum

    /// <summary>Creates SIRF propertiary NMEA input message PSRF103</summary>
    /// <remarks>
    /// The message PSRF103 sent to the serial port of the SIRF GPS chip
    /// controls if and how frequent the NMEA messages GGA, GLL, GSA,
    /// GSV, RMC and VTG have to be sent. The format of the message:
    /// <code>$PSRF103,msg,mode,rate,cksumEnable*Checksum\r\n</code>
    /// Msg is 0 for GGA, 1 for GLL, 2 for GSA, 3 for GSV, 4 for RMC and
    /// 5 for VTG. Mode is 0 for set and 1 for query. Rate the freuqnecy
    /// of the message generation in Hz. CksumEnable is 0 fr disabling
    /// the checksum generation and 1 for enabling it. Checksum is the
    /// nmea checkus as computed in Utils.nmeaChksum(). Examples of
    /// PSRF103 messages:
    /// <code>$PSRF103,00,01,00,01*25\r\n</code>
    /// query GGA message with checksum enabled.
    /// <code>$PSRF103,05,00,01,01*20\r\n</code>
    /// enable VTG message for 1Hz with checksum enabled.
    /// Based on "GPS Protocol Reference Manual", P/N: 980-0330-A, Rev. 1.30,
    /// Sectron, http://gpsd.berlios.de/vendor-docs/sirf/
    /// </remarks>
    /// <param name="msg">type of NMEA message, GGA, GLL etc.</param>
    /// <param name="mode">true for query, false for set rate</param>
    /// <param name="rate">rate in Hz, 0 for disable, max 255</param>
    /// <param name="chksumEnable">true for enable checksum</param>
    /// <returns>The PSRF103 message incl. checksum and CR, LF</returns>
    public static byte[] psrf103(Psrf103msg msg,
                                 bool       mode,
                                 byte       rate,
                                 bool       chksumEnable) {
      string s;
      byte[] retval;
      int    len;

      len = Utils.PSRF103PAT.Length;
      retval = new byte[len];
      Array.Copy(Utils.PSRF103PAT, retval, len);
      if (len > 10) retval[10] += (byte) msg;
      if (len > 13 && mode) retval[13]++;
      s = rate.ToString("X2");
      if (len > 15) retval[15] = Convert.ToByte(s[0]);
      if (len > 16) retval[16] = Convert.ToByte(s[1]);
      if (len > 19 && chksumEnable) retval[19]++;
      Utils.nmeaChksum(retval);
      return retval;
    } // psrf103()

    /// <summary>
    /// Converts the given coordinate string to decimal degrees.
    /// </summary>
    /// <remarks>
    /// The coordinates may be given either as decimal degrees or as
    /// degrees, minutes, seconds, decimal part of seconds.
    /// Example: the left or right coordinate in decimal format
    /// <c>12.58243888889</c> corresponds
    /// to <c>1234'56.78''E</c> in degrees, minutes and seconds.
    /// The negative left or right coordinate <c>-12.58243888889</c>
    /// means <c>1234'56.78''E</c>. Similar for top and bottom
    /// coordinates, where positive numbers mean North (N) and negative
    /// numbers South (S). Please consider using always dot for decimal
    /// points, rather than the locales dependend comma. Alternatively to
    /// to the character  for degrees, ' for minutes and '' for seconds the
    /// small letters d, m and s may be used, thus you can get rid of
    /// encoding specific problems especially for the degrees, for which
    /// the character is differently encoded in different code pages.
    /// So the strings <c>1234'56.78''E</c> and <c>12d34m56.78sE</c>
    /// mean exactly the same, the first one is more pretty, the later one
    /// more reliable.
    /// </remarks>
    /// <param name="coord">the coordinate string</param>
    /// <param name="isNorthSouth">true vertical, false horizontal</param>
    /// <returns>the resulting decimal degrees</returns>
    /// <exception cref="ArgumentException">
    /// For not well formated input strings, also for commas instead of
    /// dots or for Weast/East for isNorthSouth true and vice versa.
    /// </exception>
    public static double coord2dec(string coord, bool isNorthSouth) {
      const NumberStyles ST = NumberStyles.AllowDecimalPoint |
                              NumberStyles.AllowLeadingSign |
                              NumberStyles.AllowLeadingWhite |
                              NumberStyles.AllowTrailingWhite;
      CultureInfo        ci;
      MatchCollection    matches;
      Match              m;
      GroupCollection    g;
      Group              gdeg, gmin, gsec, gdir;
      string             sdir;
      double             retval, deg, min, sec, mult;

      retval = 0.0;
      if (string.IsNullOrEmpty(coord))
        throw new ArgumentException("empty string is not allowed");
      ci = CultureInfo.InvariantCulture;
      try {
        deg = double.Parse(coord, ST, ci);
        // successfully parsed decimal coordinate
        retval = deg;
      }
      catch(Exception) {
        // try to parse degrees, minutes, seconds
        matches = RECOORD.Matches(coord);
        if (matches.Count != 1)
          throw new ArgumentException(string.Format("{0} matches in {1}",
                                                    matches.Count, coord));
        m = matches[0];
        g = m.Groups;
        gdeg = g["deg"];
        gmin = g["min"];
        gsec = g["sec"];
        gdir = g["dir"];
        try {
          sdir = gdir.ToString().ToUpper().Trim();
          deg = double.Parse(gdeg.ToString(), ST, ci);
          min = double.Parse(gmin.ToString(), ST, ci);
          sec = double.Parse(gsec.ToString(), ST, ci);
        }
        catch(Exception excp) {
          throw new ArgumentException(string.Format("wrong format {0}", coord),
                                      excp);
        }
        if (isNorthSouth) {
          if (sdir == "N") mult = 1.0;
          else if (sdir == "S") mult = -1.0;
          else
            throw new ArgumentException(string.Format("wrong dir. {0} in {1}",
                                                      sdir, coord));
        }
        else {
          if (sdir == "E") mult = 1.0;
          else if (sdir == "W") mult = -1.0;
          else
            throw new ArgumentException(string.Format("wrong dir. {0} in {1}",
                                                      sdir, coord));
        }
        retval = mult * (deg + (60.0 * min + sec) / 3600.0);
      } // catch(Exception)
      return retval;
    } // coord2dec()

  } // Utils()

} // namespace serial

